@@ -1,5 +1,5 @@
 /*
-    $Id: addressobj.c 1857 2019-01-31 18:49:49Z soci $
+    $Id: addressobj.c 2122 2019-12-21 06:27:50Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -237,18 +237,16 @@ static inline bool check_addr2(atype_t type) {
     return false;
 }
 
-static FAST_CALL Obj *address(Obj *o1, uint32_t *am) {
+static FAST_CALL uint32_t address(const Obj *o1) {
     const Address *v1 = (Address *)o1;
-    atype_t type;
     Obj *v = v1->val;
-    v = v->obj->address(v, am);
-    type = v1->type;
+    uint32_t am = v->obj->address(v);
+    atype_t type = v1->type;
     while (type != A_NONE) {
-        *am <<= 4;
+        am <<= 4;
         type >>= 4;
     }
-    *am |= v1->type;
-    return v;
+    return am | v1->type;
 }
 
 static MUST_CHECK Error *ival(Obj *o1, ival_t *iv, unsigned int bits, linepos_t epoint) {
@@ -281,6 +279,18 @@ static MUST_CHECK Error *uval2(Obj *o1, uval_t *uv, unsigned int bits, linepos_t
     return v->obj->uval2(v, uv, bits, epoint);
 }
 
+static MUST_CHECK Error *iaddress(Obj *o1, ival_t *iv, unsigned int bits, linepos_t epoint) {
+    Address *v1 = (Address *)o1;
+    Obj *v = v1->val;
+    return v->obj->iaddress(v, iv, bits, epoint);
+}
+
+static MUST_CHECK Error *uaddress(Obj *o1, uval_t *uv, unsigned int bits, linepos_t epoint) {
+    Address *v1 = (Address *)o1;
+    Obj *v = v1->val;
+    return v->obj->uaddress(v, uv, bits, epoint);
+}
+
 MUST_CHECK Obj *float_from_address(Address *v1, linepos_t epoint) {
     if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
         return (Obj *)new_error_conv(&v1->v, FLOAT_OBJ, epoint);
@@ -319,14 +329,15 @@ static MUST_CHECK Obj *sign(Obj *o1, linepos_t epoint) {
     return v->obj->sign(v, epoint);
 }
 
-static MUST_CHECK Obj *function(Obj *o1, Func_types f, linepos_t epoint) {
-    Address *v1 = (Address *)o1;
+static MUST_CHECK Obj *function(oper_t op) {
+    Address *v1 = (Address *)op->v2;
     Obj *v;
     if (v1->type != A_NONE && v1->val != &none_value->v && v1->val->obj != ERROR_OBJ) {
-        return DEFAULT_OBJ->function(o1, f, epoint);
+        return DEFAULT_OBJ->function(op);
     }
-    v = v1->val;
-    return v->obj->function(v, f, epoint);
+    op->v2 = v = v1->val;
+    op->inplace = op->inplace == &v1->v && v->refcount == 1 ? v : NULL;
+    return v->obj->function(op);
 }
 
 static MUST_CHECK Obj *calc1(oper_t op) {
@@ -361,12 +372,12 @@ static MUST_CHECK Obj *calc1(oper_t op) {
     return obj_oper_error(op);
 }
 
-static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
-    Obj *val = ((Address *)o1)->val;
+static MUST_CHECK Obj *slice(oper_t op, size_t indx) {
+    Obj *val = ((Address *)op->v1)->val;
     if (val == &none_value->v || val->obj == ERROR_OBJ) {
         return val_reference(val);
     }
-    return DEFAULT_OBJ->slice(o1, op, indx);
+    return DEFAULT_OBJ->slice(op, indx);
 }
 
 static MUST_CHECK Obj *calc2(oper_t op) {
@@ -529,27 +540,17 @@ static MUST_CHECK Obj *rcalc2(oper_t op) {
     case T_GAP:
         am = v2->type;
         switch (op->op->op) {
-        case O_CMP:
-        case O_EQ:
-        case O_NE:
-        case O_MIN:
-        case O_LT:
-        case O_LE:
-        case O_MAX:
-        case O_GT:
-        case O_GE:
-            if (am == A_NONE) {
-                op->v2 = v2->val;
-                op->inplace = NULL;
-                return t1->calc2(op);
-            }
-            break;
         default:
             if (am == A_NONE) {
                 op->v2 = v2->val;
                 op->inplace = NULL;
                 return t1->calc2(op);
             }
+            break;
+        case O_MUL:
+        case O_OR:
+        case O_XOR:
+        case O_AND:
             if (check_addr2(am)) break;
             goto ok;
         case O_ADD:
@@ -585,6 +586,8 @@ void addressobj_init(void) {
     obj.ival = ival;
     obj.uval = uval;
     obj.uval2 = uval2;
+    obj.iaddress = iaddress;
+    obj.uaddress = uaddress;
     obj.sign = sign;
     obj.function = function;
     obj.calc1 = calc1;
