@@ -1,5 +1,5 @@
 /*
-    $Id: arguments.c 1687 2018-12-09 17:44:10Z soci $
+    $Id: arguments.c 2200 2020-04-07 19:18:23Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -30,7 +30,7 @@
 
 struct arguments_s arguments = {
     true,        /* warning */
-    true,        /* caret */
+    CARET_ALWAYS,/* caret */
     true,        /* quiet */
     false,       /* to_ascii */
     true,        /* monitor */
@@ -39,13 +39,10 @@ struct arguments_s arguments = {
     false,       /* longbranch */
     false,       /* tasmcomp */
     false,       /* verbose */
+    false,       /* make_phony */
     0x20,        /* caseinsensitive */
-    {            /* output */
-        "a.out",     /* name */
-        NULL,        /* section */
-        OUTPUT_CBM,  /* mode */
-        false        /* longaddr */
-    },
+    NULL,        /* output */
+    0,           /* output_len */
     &c6502,      /* cpumode */
     NULL,        /* symbol_output */
     0,           /* symbol_output_len */
@@ -61,8 +58,14 @@ struct diagnostics_s diagnostics = {
     false,       /* optimize */
     false,       /* implied_reg */
     true,        /* jmp_bug */
-    true,        /* pc_wrap */
-    true,        /* mem_wrap */
+    {
+        true,    /* wrap-pc */
+        true,    /* wrap-mem */
+        true,    /* wrap-addr */
+        true,    /* wrap-dpage */
+        true,    /* wrap-bank0 */
+        true,    /* wrap-pbank */
+    },
     true,        /* label_left */
     false,       /* branch_page */
     true,        /* deprecated */
@@ -86,7 +89,8 @@ struct diagnostics_s diagnostics = {
     false,       /* long_branch */
     false,       /* altmode */
     true,        /* page */
-    true         /* type_mixing */
+    false,       /* macro_prefix */
+    false        /* float_round */
 };
 
 struct diagnostics_s diagnostic_errors = {
@@ -95,8 +99,14 @@ struct diagnostics_s diagnostic_errors = {
     false,       /* optimize */
     false,       /* implied_reg */
     false,       /* jmp_bug */
-    false,       /* pc_wrap */
-    false,       /* mem_wrap */
+    {
+        false,   /* wrap-pc */
+        false,   /* wrap-mem */
+        false,   /* wrap-addr */
+        false,   /* wrap-dpage */
+        false,   /* wrap-bank0 */
+        false,   /* wrap-pbank */
+    },
     false,       /* label_left */
     false,       /* branch_page */
     false,       /* deprecated */
@@ -120,7 +130,8 @@ struct diagnostics_s diagnostic_errors = {
     false,       /* long_branch */
     false,       /* altmode */
     true,        /* page */
-    true         /* type_mixing */
+    false,       /* macro_prefix */
+    false        /* float_round */
 };
 
 static struct diagnostics_s diagnostic_no_all;
@@ -130,8 +141,14 @@ static struct diagnostics_s diagnostic_all = {
     false,       /* optimize */
     true,        /* implied_reg */
     true,        /* jmp_bug */
-    true,        /* pc_wrap */
-    true,        /* mem_wrap */
+    {
+        true,    /* wrap-pc */
+        true,    /* wrap-mem */
+        true,    /* wrap-addr */
+        true,    /* wrap-dpage */
+        true,    /* wrap-bank0 */
+        true,    /* wrap-pbank */
+    },
     true,        /* label_left */
     false,       /* branch_page */
     true,        /* deprecated */
@@ -155,7 +172,8 @@ static struct diagnostics_s diagnostic_all = {
     false,       /* long_branch */
     false,       /* altmode */
     true,        /* page */
-    true         /* type_mixing */
+    false,       /* macro_prefix */
+    true         /* float_round */
 };
 
 static struct diagnostics_s diagnostic_no_error_all;
@@ -165,8 +183,14 @@ static struct diagnostics_s diagnostic_error_all = {
     true,        /* optimize */
     true,        /* implied_reg */
     true,        /* jmp_bug */
-    true,        /* pc_wrap */
-    true,        /* mem_wrap */
+    {
+        true,    /* wrap-pc */
+        true,    /* wrap-mem */
+        true,    /* wrap-addr */
+        true,    /* wrap-dpage */
+        true,    /* wrap-bank0 */
+        true,    /* wrap-pbank */
+    },
     true,        /* label_left */
     true,        /* branch_page */
     true,        /* deprecated */
@@ -190,7 +214,8 @@ static struct diagnostics_s diagnostic_error_all = {
     true,        /* long_branch */
     true,        /* altmode */
     true,        /* page */
-    true         /* type_mixing */
+    true,        /* macro_prefix */
+    true         /* float_round */
 };
 
 struct w_options_s {
@@ -204,8 +229,14 @@ static const struct w_options_s w_options[] = {
     {"strict-bool",     &diagnostics.strict_bool},
     {"implied-reg",     &diagnostics.implied_reg},
     {"jmp-bug",         &diagnostics.jmp_bug},
-    {"pc-wrap",         &diagnostics.pc_wrap},
-    {"mem-wrap",        &diagnostics.mem_wrap},
+    {"pc-wrap",         &diagnostics.wrap.pc},
+    {"mem-wrap",        &diagnostics.wrap.mem},
+    {"wrap-pc",         &diagnostics.wrap.pc},
+    {"wrap-mem",        &diagnostics.wrap.mem},
+    {"wrap-addr",       &diagnostics.wrap.addr},
+    {"wrap-dpage",      &diagnostics.wrap.dpage},
+    {"wrap-bank0",      &diagnostics.wrap.bank0},
+    {"wrap-pbank",      &diagnostics.wrap.pbank},
     {"label-left",      &diagnostics.label_left},
     {"branch-page",     &diagnostics.branch_page},
     {"deprecated",      &diagnostics.deprecated},
@@ -227,7 +258,8 @@ static const struct w_options_s w_options[] = {
     {"long-branch",     &diagnostics.long_branch},
     {"altmode",         &diagnostics.altmode},
     {"page",            &diagnostics.page},
-    {"type-mixing",     &diagnostics.type_mixing},
+    {"macro-prefix",    &diagnostics.macro_prefix},
+    {"float-round",     &diagnostics.float_round},
     {NULL,              NULL}
 };
 
@@ -289,19 +321,26 @@ static const char *short_options = "wqnbfXaTCBicxtel:L:I:M:msV?o:D:E:W:";
 
 static const struct my_option long_options[] = {
     {"no-warn"          , my_no_argument      , NULL, 'w'},
+    {"warn"             , my_no_argument      , NULL,  0x11f},
+    {"no-quiet"         , my_no_argument      , NULL,  0x120},
     {"quiet"            , my_no_argument      , NULL, 'q'},
     {"nonlinear"        , my_no_argument      , NULL, 'n'},
     {"nostart"          , my_no_argument      , NULL, 'b'},
     {"flat"             , my_no_argument      , NULL, 'f'},
+    {"no-long-address"  , my_no_argument      , NULL,  0x121},
     {"long-address"     , my_no_argument      , NULL, 'X'},
     {"atari-xex"        , my_no_argument      , NULL,  0x107},
     {"apple-ii"         , my_no_argument      , NULL,  0x108},
     {"intel-hex"        , my_no_argument      , NULL,  0x10e},
     {"s-record"         , my_no_argument      , NULL,  0x10f},
     {"cbm-prg"          , my_no_argument      , NULL,  0x10c},
+    {"no-ascii"         , my_no_argument      , NULL,  0x11e},
     {"ascii"            , my_no_argument      , NULL, 'a'},
+    {"no-tasm-compatible",my_no_argument      , NULL,  0x11d},
     {"tasm-compatible"  , my_no_argument      , NULL, 'T'},
+    {"no-case-sensitive", my_no_argument      , NULL,  0x11c},
     {"case-sensitive"   , my_no_argument      , NULL, 'C'},
+    {"no-long-branch"   , my_no_argument      , NULL,  0x11b},
     {"long-branch"      , my_no_argument      , NULL, 'B'},
     {"m65xx"            , my_no_argument      , NULL,  0x101},
     {"m6502"            , my_no_argument      , NULL, 'i'},
@@ -317,15 +356,27 @@ static const struct my_option long_options[] = {
     {"output"           , my_required_argument, NULL, 'o'},
     {"output-section"   , my_required_argument, NULL,  0x114},
     {"error"            , my_required_argument, NULL, 'E'},
+    {"normal-labels"    , my_no_argument      , NULL,  0x124},
+    {"export-labels"    , my_no_argument      , NULL,  0x115},
     {"vice-labels"      , my_no_argument      , NULL,  0x10b},
+    {"vice-labels-numeric",my_no_argument     , NULL,  0x123},
     {"dump-labels"      , my_no_argument      , NULL,  0x10d},
     {"labels-root"      , my_required_argument, NULL,  0x113},
     {"list"             , my_required_argument, NULL, 'L'},
+    {"dependencies"     , my_required_argument, NULL, 'M'},
+    {"no-make-phony"    , my_no_argument      , NULL,  0x126},
+    {"make-phony"       , my_no_argument      , NULL,  0x125},
+    {"no-verbose-list"  , my_no_argument      , NULL,  0x11a},
     {"verbose-list"     , my_no_argument      , NULL,  0x110},
     {"no-monitor"       , my_no_argument      , NULL, 'm'},
+    {"monitor"          , my_no_argument      , NULL,  0x119},
     {"no-source"        , my_no_argument      , NULL, 's'},
+    {"source"           , my_no_argument      , NULL,  0x118},
+    {"no-line-numbers"  , my_no_argument      , NULL,  0x117},
     {"line-numbers"     , my_no_argument      , NULL,  0x112},
     {"no-caret-diag"    , my_no_argument      , NULL,  0x10a},
+    {"macro-caret-diag" , my_no_argument      , NULL,  0x122},
+    {"caret-diag"       , my_no_argument      , NULL,  0x116},
     {"tab-size"         , my_required_argument, NULL,  0x109},
     {"version"          , my_no_argument      , NULL, 'V'},
     {"usage"            , my_no_argument      , NULL,  0x102},
@@ -405,6 +456,29 @@ static MUST_CHECK char *read_one(FILE *f) {
     return (char *)data;
 }
 
+static address_t get_all_mem2(void) {
+    size_t i;
+    bool tostdout = false;
+    address_t min = 0xffffffff;
+    for (i = 0; i < arguments.output_len; i++) {
+        const struct output_s *output = &arguments.output[i];
+        switch (output->mode) {
+        case OUTPUT_RAW:
+        case OUTPUT_NONLINEAR:
+        case OUTPUT_CBM: min &= output->longaddr ? 0xffffff : 0xffff; break;
+        case OUTPUT_IHEX:
+        case OUTPUT_SREC:
+        case OUTPUT_FLAT: min &= 0xffffffff; break;
+        case OUTPUT_APPLE:
+        case OUTPUT_XEX: min &= 0xffff; break;
+        }
+        if (dash_name(output->name)) tostdout = true;
+    }
+    if (tostdout) arguments.quiet = false;
+    else setvbuf(stdout, NULL, _IOLBF, 1024);
+    return min;
+}
+
 int testarg(int *argc2, char **argv2[], struct file_s *fin) {
     int argc = *argc2;
     char **argv = *argv2;
@@ -413,6 +487,7 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
     int max = 10;
     bool again;
     struct symbol_output_s symbol_output = { NULL, LABEL_64TASS, NULL };
+    struct output_s output = { "a.out", NULL, OUTPUT_CBM, false };
 
     do {
         int i;
@@ -425,43 +500,59 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
                 if (woption(my_optarg)) goto exit;
                 break;
             case 'w':arguments.warning = false;break;
+            case 0x11f:arguments.warning = true;break;
             case 'q':arguments.quiet = false;break;
-            case 'X':arguments.output.longaddr = true;break;
-            case 'n':arguments.output.mode = OUTPUT_NONLINEAR;break;
-            case 0x107:arguments.output.mode = OUTPUT_XEX;break;
-            case 0x108:arguments.output.mode = OUTPUT_APPLE;break;
-            case 0x10e:arguments.output.mode = OUTPUT_IHEX;break;
-            case 0x10f:arguments.output.mode = OUTPUT_SREC;break;
-            case 0x10c:arguments.output.mode = OUTPUT_CBM;break;
-            case 'b':arguments.output.mode = OUTPUT_RAW;break;
-            case 'f':arguments.output.mode = OUTPUT_FLAT;break;
+            case 0x120:arguments.quiet = true;break;
+            case 'X':output.longaddr = true;break;
+            case 0x121:output.longaddr = false;break;
+            case 'n':output.mode = OUTPUT_NONLINEAR;break;
+            case 0x107:output.mode = OUTPUT_XEX;break;
+            case 0x108:output.mode = OUTPUT_APPLE;break;
+            case 0x10e:output.mode = OUTPUT_IHEX;break;
+            case 0x10f:output.mode = OUTPUT_SREC;break;
+            case 0x10c:output.mode = OUTPUT_CBM;break;
+            case 'b':output.mode = OUTPUT_RAW;break;
+            case 'f':output.mode = OUTPUT_FLAT;break;
             case 'a':arguments.to_ascii = true;break;
+            case 0x11e:arguments.to_ascii = false;break;
             case 'T':arguments.tasmcomp = true;break;
-            case 'o':arguments.output.name = my_optarg;break;
-            case 0x114: arguments.output.section = my_optarg; break;
-            case 0x10a:arguments.caret = false;break;
+            case 0x11d:arguments.tasmcomp = false;break;
+            case 'o': output.name = my_optarg;
+                      arguments.output_len++;
+                      arguments.output = (struct output_s *)realloc(arguments.output, arguments.output_len * sizeof *arguments.output);
+                      if (arguments.output == NULL) err_msg_out_of_memory2();
+                      arguments.output[arguments.output_len - 1] = output;
+                      output.section = NULL;
+                      break;
+            case 0x114:output.section = my_optarg; break;
+            case 0x116:arguments.caret = CARET_ALWAYS;break;
+            case 0x122:arguments.caret = CARET_MACRO;break;
+            case 0x10a:arguments.caret = CARET_NEVER;break;
             case 'D':
                 {
                     size_t len = strlen(my_optarg) + 1;
 
                     if (fin->lines >= max_lines) {
                         max_lines += 1024;
-                        if (/*max_lines < 1024 ||*/ max_lines > SIZE_MAX / sizeof *fin->line) err_msg_out_of_memory(); /* overflow */
-                        fin->line = (size_t *)reallocx(fin->line, max_lines * sizeof *fin->line);
+                        if (/*max_lines < 1024 ||*/ max_lines > SIZE_MAX / sizeof *fin->line) err_msg_out_of_memory2(); /* overflow */
+                        fin->line = (size_t *)realloc(fin->line, max_lines * sizeof *fin->line);
+                        if (fin->line == NULL) err_msg_out_of_memory2();
                     }
                     fin->line[fin->lines++] = fp;
 
-                    if (len < 1 || fp + len < len) err_msg_out_of_memory();
+                    if (len < 1 || fp + len < len) err_msg_out_of_memory2();
                     if (fp + len > fin->len) {
                         fin->len = fp + len + 1024;
-                        if (fin->len < 1024) err_msg_out_of_memory();
-                        fin->data = (uint8_t*)reallocx(fin->data, fin->len);
+                        if (fin->len < 1024) err_msg_out_of_memory2();
+                        fin->data = (uint8_t*)realloc(fin->data, fin->len);
+                        if (fin->data == NULL) err_msg_out_of_memory2();
                     }
                     memcpy(fin->data + fp, my_optarg, len);
                     fp += len;
                 }
                 break;
             case 'B': arguments.longbranch = true;break;
+            case 0x11b: arguments.longbranch = false;break;
             case 0x101: arguments.cpumode = &c6502;break;
             case 'i': arguments.cpumode = &c6502i;break;
             case 'c': arguments.cpumode = &c65c02;break;
@@ -474,12 +565,15 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
             case 0x111: arguments.cpumode = &c4510;break;
             case 'l': symbol_output.name = my_optarg;
                       arguments.symbol_output_len++;
-                      arguments.symbol_output = (struct symbol_output_s *)reallocx(arguments.symbol_output, arguments.symbol_output_len * sizeof *arguments.symbol_output);
+                      arguments.symbol_output = (struct symbol_output_s *)realloc(arguments.symbol_output, arguments.symbol_output_len * sizeof *arguments.symbol_output);
+                      if (arguments.symbol_output == NULL) err_msg_out_of_memory2();
                       arguments.symbol_output[arguments.symbol_output_len - 1] = symbol_output;
-                      symbol_output.mode = LABEL_64TASS;
                       symbol_output.space = NULL;
                       break;
+            case 0x124: symbol_output.mode = LABEL_64TASS; break;
+            case 0x115: symbol_output.mode = LABEL_EXPORT; break;
             case 0x10b: symbol_output.mode = LABEL_VICE; break;
+            case 0x123: symbol_output.mode = LABEL_VICE_NUMERIC; break;
             case 0x10d: symbol_output.mode = LABEL_DUMP; break;
             case 0x113: symbol_output.space = my_optarg; break;
             case 'E': arguments.error = my_optarg;break;
@@ -487,10 +581,17 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
             case 'M': arguments.make = my_optarg;break;
             case 'I': include_list_add(my_optarg);break;
             case 'm': arguments.monitor = false;break;
+            case 0x119: arguments.monitor = true;break;
             case 's': arguments.source = false;break;
+            case 0x118: arguments.source = true;break;
             case 0x112: arguments.linenum = true;break;
+            case 0x117: arguments.linenum = false;break;
             case 'C': arguments.caseinsensitive = 0;break;
+            case 0x11c: arguments.caseinsensitive = 0x20;break;
             case 0x110: arguments.verbose = true;break;
+            case 0x11a: arguments.verbose = false;break;
+            case 0x125: arguments.make_phony = true;break;
+            case 0x126: arguments.make_phony = false;break;
             case 0x109:
                 {
                     char *s;
@@ -506,11 +607,12 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
                "        [--atari-xex] [--apple-ii] [--intel-hex] [--s-record] [--nonlinear]\n"
                "        [--tasm-compatible] [--quiet] [--no-warn] [--long-address] [--m65c02]\n"
                "        [--m6502] [--m65xx] [--m65dtv02] [--m65816] [--m65el02] [--mr65c02]\n"
-               "        [--mw65c02] [--m65ce02] [--m4510] [--labels=<file>] [--vice-labels]\n"
+               "        [--mw65c02] [--m65ce02] [--m4510] [--labels=<file>] [--normal-labels]\n"
+               "        [--export-labels] [--vice-labels] [--vice-labels-numeric]\n"
                "        [--dump-labels] [--list=<file>] [--no-monitor] [--no-source]\n"
-               "        [--line-numbers] [--tab-size=<value>] [--verbose-list] [-W<option>]\n"
-               "        [--errors=<file>] [--output=<file>] [--help] [--usage]\n"
-               "        [--version] SOURCES");
+               "        [--line-numbers] [--tab-size=<value>] [--verbose-list] \n"
+               "        [--dependencies=<file>] [--make-phony] [-W<option>] [--errors=<file>]\n"
+               "        [--output=<file>] [--help] [--usage] [--version] SOURCES");
                    return 0;
 
             case 'V':puts("64tass Turbo Assembler Macro V" VERSION);
@@ -520,95 +622,106 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
                "Usage: 64tass [OPTIONS...] SOURCES\n"
                "64tass Turbo Assembler Macro V" VERSION "\n"
                "\n"
-               "  -a, --ascii           Source is not in PETASCII\n"
-               "  -B, --long-branch     Automatic bxx *+3 jmp $xxxx\n"
-               "  -C, --case-sensitive  Case sensitive labels\n"
-               "  -D <label>=<value>    Define <label> to <value>\n"
-               "  -E, --error=<file>    Place errors into <file>\n"
-               "  -I <path>             Include search path\n"
-               "  -M <file>             Makefile dependencies to <file>\n"
-               "  -q, --quiet           Do not output summary and header\n"
-               "  -T, --tasm-compatible Enable TASM compatible mode\n"
-               "  -w, --no-warn         Suppress warnings\n"
-               "      --no-caret-diag   Suppress source line display\n"
+               "  -a, --ascii            Source is not in PETASCII\n"
+               "  -B, --long-branch      Automatic bxx *+5 jmp $xxxx\n"
+               "  -C, --case-sensitive   Case sensitive labels\n"
+               "  -D <label>=<value>     Define <label> to <value>\n"
+               "  -E, --error=<file>     Place errors into <file>\n"
+               "  -I <path>              Include search path\n"
+               "  -M, --dependencies=<f> Makefile dependencies to <file>\n"
+               "  -q, --quiet            Do not output summary and header\n"
+               "  -T, --tasm-compatible  Enable TASM compatible mode\n"
+               "  -w, --no-warn          Suppress warnings\n"
+               "      --make-phony       Add phony target to dependencies\n"
+               "      --no-caret-diag    Suppress source line display\n"
+               "      --macro-caret-diag Source lines in macros only\n"
                "\n"
                " Diagnostic options:\n"
-               "  -Wall                 Enable most diagnostic warnings\n"
-               "  -Werror               Diagnostic warnings to errors\n"
-               "  -Werror=<name>        Make a diagnostic to an error\n"
-               "  -Wno-error=<name>     Make a diagnostic to a warning\n"
-               "  -Walias               Warn about instruction aliases\n"
-               "  -Waltmode             Warn about alternative addressing\n"
-               "  -Wbranch-page         Warn if a branch crosses a page\n"
-               "  -Wcase-symbol         Warn on mismatch of symbol case\n"
-               "  -Wimmediate           Suggest immediate addressing\n"
-               "  -Wimplied-reg         No implied register aliases\n"
-               "  -Wleading-zeros       Warn for ignored leading zeros\n"
-               "  -Wlong-branch         Warn when a long branch is used\n"
-               "  -Wno-deprecated       No deprecated feature warnings\n"
-               "  -Wno-float-compare    No approximate compare warnings\n"
-               "  -Wno-ignored          No directive ignored warnings\n"
-               "  -Wno-jmp-bug          No jmp ($xxff) bug warning\n"
-               "  -Wno-label-left       No warning about strange labels\n"
-               "  -Wno-mem-wrap         No offset overflow warning\n"
-               "  -Wno-page             No page crossing error\n"
-               "  -Wno-pc-wrap          No PC overflow warning\n"
-               "  -Wno-pitfalls         No common pitfall notes\n"
-               "  -Wno-portable         No portability warnings\n"
-               "  -Wno-star-assign      No label multiply warnings\n"
-               "  -Wold-equal           Warn about old equal operator\n"
-               "  -Woptimize            Optimization warnings\n"
-               "  -Wshadow              Check symbol shadowing\n"
-               "  -Wstrict-bool         No implicit bool conversions\n"
-               "  -Wswitch-case         Warn about ignored cases\n"
-               "  -Wunused              Warn about unused symbols\n"
-               "  -Wunused-macro        Warn about unused macros\n"
-               "  -Wunused-const        Warn about unused consts\n"
-               "  -Wunused-label        Warn about unused labels\n"
-               "  -Wunused-variable     Warn about unused variables\n"
-               "\n"
-               " Output selection:\n"
-               "  -o, --output=<file>   Place output into <file>\n"
-               "   --output-section=<n> Output this section only\n"
-               "  -b, --nostart         Strip starting address\n"
-               "  -f, --flat            Generate flat output file\n"
-               "  -n, --nonlinear       Generate nonlinear output file\n"
-               "  -X, --long-address    Use 3 byte start/len address\n"
-               "      --cbm-prg         Output CBM program file\n"
-               "      --atari-xex       Output Atari XEX file\n"
-               "      --apple-ii        Output Apple II file\n"
-               "      --intel-hex       Output Intel HEX file\n"
-               "      --s-record        Output Motorola S-record file\n"
+               "  -Wall                  Enable most diagnostic warnings\n"
+               "  -Werror                Diagnostic warnings to errors\n"
+               "  -Werror=<name>         Make a diagnostic to an error\n"
+               "  -Wno-error=<name>      Make a diagnostic to a warning\n"
+               "  -Walias                Warn about instruction aliases\n"
+               "  -Waltmode              Warn about alternative addressing\n"
+               "  -Wbranch-page          Warn if a branch crosses a page\n"
+               "  -Wcase-symbol          Warn on mismatch of symbol case\n"
+               "  -Wimmediate            Suggest immediate addressing\n"
+               "  -Wimplied-reg          No implied register aliases\n"
+               "  -Wleading-zeros        Warn for ignored leading zeros\n"
+               "  -Wlong-branch          Warn when a long branch is used\n"
+               "  -Wmacro-prefix         Warn about unprefixed macro calls\n"
+               "  -Wno-deprecated        No deprecated feature warnings\n"
+               "  -Wno-float-compare     No approximate compare warnings\n"
+               "  -Wno-float-round       No implicit rounding warnings\n"
+               "  -Wno-ignored           No directive ignored warnings\n"
+               "  -Wno-jmp-bug           No jmp ($xxff) bug warning\n"
+               "  -Wno-label-left        No warning about strange labels\n"
+               "  -Wno-page              No page crossing error\n"
+               "  -Wno-wrap-addr         No memory address overflow warning\n"
+               "  -Wno-wrap-bank0        No bank0 overflow warning\n"
+               "  -Wno-wrap-dpage        No direct page overflow warning\n"
+               "  -Wno-wrap-mem          No offset overflow warning\n"
+               "  -Wno-wrap-pbank        No program bank overflow warning\n"
+               "  -Wno-wrap-pc           No PC overflow warning\n"
+               "  -Wno-pitfalls          No common pitfall notes\n"
+               "  -Wno-portable          No portability warnings\n"
+               "  -Wno-star-assign       No label multiply warnings\n"
+               "  -Wold-equal            Warn about old equal operator\n"
+               "  -Woptimize             Optimization warnings\n"
+               "  -Wshadow               Check symbol shadowing\n"
+               "  -Wstrict-bool          No implicit bool conversions\n"
+               "  -Wswitch-case          Warn about ignored cases\n"
+               "  -Wunused               Warn about unused symbols\n"
+               "  -Wunused-macro         Warn about unused macros\n"
+               "  -Wunused-const         Warn about unused consts\n"
+               "  -Wunused-label         Warn about unused labels\n"
+               "  -Wunused-variable      Warn about unused variables\n");
+               puts(" Output selection:\n"
+               "  -o, --output=<file>    Place output into <file>\n"
+               "   --output-section=<n>  Output this section only\n"
+               "  -b, --nostart          Strip starting address\n"
+               "  -f, --flat             Generate flat output file\n"
+               "  -n, --nonlinear        Generate nonlinear output file\n"
+               "  -X, --long-address     Use 3 byte start/len address\n"
+               "      --cbm-prg          Output CBM program file\n"
+               "      --atari-xex        Output Atari XEX file\n"
+               "      --apple-ii         Output Apple II file\n"
+               "      --intel-hex        Output Intel HEX file\n"
+               "      --s-record         Output Motorola S-record file\n"
                "\n"
                " Target CPU selection:\n"
-               "      --m65xx           Standard 65xx (default)\n"
-               "  -c, --m65c02          CMOS 65C02\n"
-               "      --m65ce02         CSG 65CE02\n"
-               "  -e, --m65el02         65EL02\n"
-               "  -i, --m6502           NMOS 65xx\n"
-               "  -t, --m65dtv02        65DTV02\n"
-               "  -x, --m65816          W65C816\n"
-               "      --mr65c02         R65C02\n"
-               "      --mw65c02         W65C02\n"
-               "      --m4510           CSG 4510\n"
+               "      --m65xx            Standard 65xx (default)\n"
+               "  -c, --m65c02           CMOS 65C02\n"
+               "      --m65ce02          CSG 65CE02\n"
+               "  -e, --m65el02          65EL02\n"
+               "  -i, --m6502            NMOS 65xx\n"
+               "  -t, --m65dtv02         65DTV02\n"
+               "  -x, --m65816           W65C816\n"
+               "      --mr65c02          R65C02\n"
+               "      --mw65c02          W65C02\n"
+               "      --m4510            CSG 4510\n"
                "\n"
                " Source listing and labels:\n"
-               "  -l, --labels=<file>   List labels into <file>\n"
-               "      --vice-labels     Labels in VICE format\n"
-               "      --dump-labels     Dump for debugging\n"
-               "      --labels-root=<l> List from scope <l> only\n"
-               "  -L, --list=<file>     List into <file>\n"
-               "  -m, --no-monitor      Don't put monitor code into listing\n"
-               "  -s, --no-source       Don't put source code into listing\n"
-               "      --line-numbers    Put line numbers into listing\n"
-               "      --tab-size=<n>    Override the default tab size (8)\n"
-               "      --verbose-list    List unused lines as well\n"
+               "  -l, --labels=<file>    List labels into <file>\n"
+               "      --normal-labels    Labels in native format\n"
+               "      --export-labels    Export for other source\n"
+               "      --vice-labels      Labels in VICE format\n"
+               "      --vice-labels-numeric Labels for VICE with numeric constants\n"
+               "      --dump-labels      Dump for debugging\n"
+               "      --labels-root=<l>  List from scope <l> only\n"
+               "  -L, --list=<file>      List into <file>\n"
+               "  -m, --no-monitor       Don't put monitor code into listing\n"
+               "  -s, --no-source        Don't put source code into listing\n"
+               "      --line-numbers     Put line numbers into listing\n"
+               "      --tab-size=<n>     Override the default tab size (8)\n"
+               "      --verbose-list     List unused lines as well\n"
                "\n"
                " Misc:\n"
-               "  -?, --help            Give this help list\n"
-               "      --usage           Give a short usage message\n"
-               "  -V, --version         Print program version\n");
-               puts("Mandatory or optional arguments to long options are also mandatory or optional\n"
+               "  -?, --help             Give this help list\n"
+               "      --usage            Give a short usage message\n"
+               "  -V, --version          Print program version\n"
+               "\n"
+               "Mandatory or optional arguments to long options are also mandatory or optional\n"
                "for any corresponding short options.\n"
                "\n"
                "Report bugs to <soci" "\x40" "c64.rulez.org>.");
@@ -626,29 +739,29 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
         for (i = my_optind; i < argc; i++) {
             char *arg = argv[i];
             if (arg[0] == '@' && arg[1] != 0) {
-                FILE *f = fopen(arg+1, "rb");
-                if (f != NULL) {
-                    int j;
-                    free(arg);
-                    argc--;
-                    for (j = i; j < argc; j++) {
-                        argv[j] = argv[j + 1];
-                    }
-                    while (feof(f) == 0) {
-                        char *onepar = read_one(f);
-                        if (onepar == NULL) break;
-                        *argv2 = argv = (char **)reallocx(argv, ((size_t)argc + 1) * sizeof *argv);
+                FILE *f = file_open(arg + 1, "rb");
+                if (f == NULL) continue;
+                while (feof(f) == 0) {
+                    char *onepar = read_one(f);
+                    if (onepar == NULL) break;
+                    *argv2 = argv = (char **)realloc(argv, ((size_t)argc + 1) * sizeof *argv);
+                    if (argv == NULL) err_msg_out_of_memory2();
+                    if (arg != NULL) {
+                        free(arg);
+                        arg = NULL;
+                    } else {
+                        int j;
                         for (j = argc; j > i; j--) {
                             argv[j] = argv[j - 1];
                         }
                         argc++;
-                        argv[i++] = onepar;
-                        again = true;
                     }
-                    fclose(f);
-                    *argc2 = argc;
-                    break;
+                    argv[i++] = onepar;
+                    again = true;
                 }
+                fclose(f);
+                *argc2 = argc;
+                break;
             }
         }
         max--;
@@ -660,25 +773,25 @@ int testarg(int *argc2, char **argv2[], struct file_s *fin) {
         return -1;
     }
 
-    if (arguments.symbol_output_len > 0) {
-        if (symbol_output.mode != LABEL_64TASS) arguments.symbol_output[arguments.symbol_output_len - 1].mode = symbol_output.mode;
+    if (arguments.symbol_output_len != 0) {
+        arguments.symbol_output[arguments.symbol_output_len - 1].mode = symbol_output.mode;
         if (symbol_output.space != NULL) arguments.symbol_output[arguments.symbol_output_len - 1].space = symbol_output.space;
     }
 
-    switch (arguments.output.mode) {
-    case OUTPUT_RAW:
-    case OUTPUT_NONLINEAR:
-    case OUTPUT_CBM: all_mem2 = arguments.output.longaddr ? 0xffffff : 0xffff; break;
-    case OUTPUT_IHEX:
-    case OUTPUT_SREC:
-    case OUTPUT_FLAT: all_mem2 = 0xffffffff; break;
-    case OUTPUT_APPLE:
-    case OUTPUT_XEX: all_mem2 = 0xffff; break;
+    if (arguments.output == NULL) {
+        arguments.output = (struct output_s *)mallocx(sizeof *arguments.output);
+        arguments.output[0] = output;
+        arguments.output_len = 1;
+    } else {
+        arguments.output[arguments.output_len - 1].mode = output.mode;
+        if (output.section != NULL) arguments.output[arguments.output_len - 1].section = NULL;
+        arguments.output[arguments.output_len - 1].longaddr = output.longaddr;
     }
+
+    all_mem2 = get_all_mem2();
     if (arguments.caseinsensitive == 0) {
         diagnostics.case_symbol = false;
     }
-    if (dash_name(arguments.output.name)) arguments.quiet = false;
     if (fin->lines != max_lines) {
         size_t *d = (size_t *)realloc(fin->line, fin->lines * sizeof *fin->line);
         if (fin->lines == 0 || d != NULL) fin->line = d;
