@@ -1,5 +1,5 @@
 /*
-    $Id: bytesobj.c 1861 2019-02-03 19:36:52Z soci $
+    $Id: bytesobj.c 2142 2020-01-03 16:26:22Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -51,8 +51,8 @@ static inline size_t byteslen(const Bytes *v1) {
     return (len < 0) ? (size_t)~len : (size_t)len;
 }
 
-static MUST_CHECK Obj *bytes_from_bits(const Bits *, linepos_t);
 static MUST_CHECK Obj *bytes_from_int(const Int *, linepos_t);
+static MUST_CHECK Bytes *bytes_from_u8(unsigned int i);
 
 static MUST_CHECK Obj *create(Obj *v1, linepos_t epoint) {
     Obj *err, *ret;
@@ -77,9 +77,13 @@ static MUST_CHECK Obj *create(Obj *v1, linepos_t epoint) {
     return (Obj *)new_error_conv(v1, BYTES_OBJ, epoint);
 }
 
+static FAST_CALL NO_INLINE void bytes_destroy(Bytes *v1) {
+    free(v1->data);
+}
+
 static FAST_CALL void destroy(Obj *o1) {
     Bytes *v1 = (Bytes *)o1;
-    if (v1->u.val != v1->data) free(v1->data);
+    if (v1->u.val != v1->data) bytes_destroy(v1);
 }
 
 MALLOC Bytes *new_bytes(size_t ln) {
@@ -130,7 +134,7 @@ static uint8_t *extend_bytes(Bytes *v, size_t ln) {
     }
     tmp = (uint8_t *)malloc(ln);
     if (tmp != NULL) {
-        memcpy(tmp, v->u.val, v->len);
+        memcpy(tmp, v->u.val, byteslen(v));
         v->data = tmp;
         v->u.s.max = ln;
         v->u.s.hash = -1;
@@ -181,11 +185,20 @@ failed:
     return (Obj *)new_error_mem(epoint);
 }
 
+static FAST_CALL NO_INLINE bool bytes_same(const Bytes *v1, const Bytes *v2) {
+    return memcmp(v1->data, v2->data, byteslen(v2)) == 0;
+}
+
 static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
     const Bytes *v1 = (const Bytes *)o1, *v2 = (const Bytes *)o2;
-    return o2->obj == BYTES_OBJ && v1->len == v2->len && (
-            v1->data == v2->data ||
-            memcmp(v1->data, v2->data, byteslen(v2)) == 0);
+    if (o1->obj != o2->obj || v1->len != v2->len) return false;
+    switch (v1->len) {
+    case ~0:
+    case 0: return true;
+    case ~1:
+    case 1: return v1->data[0] == v2->data[0];
+    default: return bytes_same(v1, v2);
+    }
 }
 
 static bool to_bool(const Bytes *v1) {
@@ -233,9 +246,9 @@ static uint8_t *z85_encode(uint8_t *dest, const uint8_t *src, size_t len) {
         tmp = src2[3] | (src2[2] << 8) | (src2[1] << 16) | (src2[0] << 24);
 
         for (j = 4; j > 0; j--) {
-            uint32_t div = tmp / 85;
-            dest[j] = z85[tmp - div * 85];
-            tmp = div;
+            uint32_t divided = tmp / 85;
+            dest[j] = z85[tmp - divided * 85];
+            tmp = divided;
         }
         dest[j] = z85[tmp];
         dest += 5;
@@ -243,7 +256,7 @@ static uint8_t *z85_encode(uint8_t *dest, const uint8_t *src, size_t len) {
     return dest;
 }
 
-const uint8_t z85_dec[93] = {
+static const uint8_t z85_dec[93] = {
         68, 85, 84, 83, 82, 72, 85, 75, 76, 70, 65, 85, 63, 62, 69,
     0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  64, 85, 73, 66, 74, 71,
     81, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
@@ -390,8 +403,8 @@ MUST_CHECK Obj *bytes_from_hexstr(const uint8_t *s, size_t *ln, linepos_t epoint
     j /= 2;
     v = new_bytes2(j);
     if (v == NULL) return (Obj *)new_error_mem(epoint);
-    v->len = j;
-    for (i = 0; i < (size_t)v->len; i++) {
+    v->len = (ssize_t)j;
+    for (i = 0; i < j; i++) {
         uint8_t c1, c2;
         s++;
         c1 = s[i] ^ 0x30;
@@ -540,7 +553,7 @@ failed:
     return (Obj *)new_error_mem(epoint);
 }
 
-MUST_CHECK Bytes *bytes_from_u8(unsigned int i) {
+static MUST_CHECK Bytes *bytes_from_u8(unsigned int i) {
     Bytes *v;
     i &= 0xff;
     v = bytes_value[i];
@@ -553,7 +566,7 @@ MUST_CHECK Bytes *bytes_from_u8(unsigned int i) {
     return ref_bytes(v);
 }
 
-MUST_CHECK Bytes *bytes_from_u16(unsigned int i) {
+static MUST_CHECK Bytes *bytes_from_u16(unsigned int i) {
     Bytes *v = new_bytes(2);
     v->len = 2;
     v->data[0] = (uint8_t)i;
@@ -561,6 +574,18 @@ MUST_CHECK Bytes *bytes_from_u16(unsigned int i) {
     return v;
 }
 
+MUST_CHECK Obj *bytes_calc1(Oper_types op, unsigned int val) {
+    switch (op) {
+    case O_BANK: val >>= 8; /* fall through */
+    case O_HIGHER: val >>= 8; /* fall through */
+    case O_LOWER: 
+    default: return (Obj *)bytes_from_u8(val);
+    case O_HWORD: val >>= 8; /* fall through */
+    case O_WORD: return (Obj *)bytes_from_u16(val);
+    case O_BSWORD: return (Obj *)bytes_from_u16((uint8_t)(val >> 8) | (uint16_t)(val << 8));
+    }
+}
+
 MUST_CHECK Bytes *bytes_from_uval(uval_t i, unsigned int bytes) {
     Bytes *v = new_bytes(bytes);
     v->len = (ssize_t)bytes;
@@ -574,7 +599,7 @@ MUST_CHECK Bytes *bytes_from_uval(uval_t i, unsigned int bytes) {
     return v;
 }
 
-static MUST_CHECK Obj *bytes_from_bits(const Bits *v1, linepos_t epoint) {
+MUST_CHECK Obj *bytes_from_bits(const Bits *v1, linepos_t epoint) {
     size_t i, sz, len1;
     uint8_t *d;
     Bytes *v;
@@ -770,52 +795,46 @@ static MUST_CHECK Obj *sign(Obj *o1, linepos_t UNUSED(epoint)) {
     return (Obj *)ref_int(int_value[0]);
 }
 
-static MUST_CHECK Obj *function(Obj *o1, Func_types f, linepos_t epoint) {
-    Bytes *v1 = (Bytes *)o1;
-    Obj *tmp = int_from_bytes(v1, epoint);
-    Obj *ret = tmp->obj->function(tmp, f, epoint);
+static MUST_CHECK Obj *function(oper_t op) {
+    Bytes *v1 = (Bytes *)op->v2;
+    Obj *tmp, *ret;
+    op->v2 = tmp = int_from_bytes(v1, op->epoint2);
+    op->inplace = tmp->refcount == 1 ? tmp : NULL;
+    ret = tmp->obj->function(op);
     val_destroy(tmp);
     return ret;
 }
 
-static MUST_CHECK Obj *len(Obj *o1, linepos_t UNUSED(epoint)) {
-    Bytes *v1 = (Bytes *)o1;
+static MUST_CHECK Obj *len(oper_t op) {
+    Bytes *v1 = (Bytes *)op->v2;
     return (Obj *)int_from_size(byteslen(v1));
 }
 
-static size_t iter_len(Iter *v1) {
-    return ((Bytes *)v1->data)->len - v1->val;
-}
-
-static FAST_CALL MUST_CHECK Obj *next(Iter *v1) {
+static FAST_CALL MUST_CHECK Obj *next(struct iter_s *v1) {
     Bytes *iter;
     const Bytes *vv1 = (Bytes *)v1->data;
-    if (v1->val >= byteslen(vv1)) return NULL;
+    if (v1->val >= v1->len) return NULL;
     iter = (Bytes *)v1->iter;
-    if (iter == NULL) {
-    renew:
+    if (iter->v.refcount != 1) {
+        iter->v.refcount--;
         iter = new_bytes(1);
         v1->iter = &iter->v;
         iter->len = (vv1->len < 0) ? ~1 : 1;
-    } else if (iter->v.refcount != 1) {
-        iter->v.refcount--;
-        goto renew;
     }
     iter->data[0] = vv1->data[v1->val++];
     return &iter->v;
 }
 
-static MUST_CHECK Iter *getiter(Obj *v1) {
-    Iter *v = (Iter *)val_alloc(ITER_OBJ);
-    v->iter = NULL;
+static void getiter(struct iter_s *v) {
+    v->iter = val_reference(v->data);
     v->val = 0;
-    v->data = val_reference(v1);
+    v->data = val_reference(v->data);
     v->next = next;
-    v->len = iter_len;
-    return v;
+    v->len = byteslen((Bytes *)v->data);
 }
 
-static MUST_CHECK Obj *and_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint) {
+static inline MUST_CHECK Obj *and_(oper_t op) {
+    Bytes *vv1 = (Bytes *)op->v1, *vv2 = (Bytes *)op->v2;
     size_t i, len1, len2, sz;
     bool neg1, neg2;
     uint8_t *v1, *v2, *v;
@@ -823,15 +842,23 @@ static MUST_CHECK Obj *and_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint
     len1 = byteslen(vv1); len2 = byteslen(vv2);
 
     if (len1 < len2) {
-        const Bytes *tmp = vv1; vv1 = vv2; vv2 = tmp;
+        Bytes *tmp = vv1; vv1 = vv2; vv2 = tmp;
         i = len1; len1 = len2; len2 = i;
     }
     neg1 = vv1->len < 0; neg2 = vv2->len < 0;
 
     sz = neg2 ? len1 : len2;
     if (sz == 0) return val_reference((neg1 && neg2) ? &inv_bytes->v : &null_bytes->v);
-    vv = new_bytes2(sz);
-    if (vv == NULL) return (Obj *)new_error_mem(epoint);
+    if (op->inplace == &vv1->v) {
+        vv = ref_bytes(vv1);
+        if (vv->data != vv->u.val) vv->u.s.hash = -1;
+    } else if (op->inplace == &vv2->v && sz <= len2) {
+        vv = ref_bytes(vv2);
+        if (vv->data != vv->u.val) vv->u.s.hash = -1;
+    } else {
+        vv = new_bytes2(sz);
+        if (vv == NULL) return (Obj *)new_error_mem(op->epoint3);
+    }
     v = vv->data;
     v1 = vv1->data; v2 = vv2->data;
 
@@ -856,7 +883,8 @@ static MUST_CHECK Obj *and_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint
     return &vv->v;
 }
 
-static MUST_CHECK Obj *or_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint) {
+static inline MUST_CHECK Obj *or_(oper_t op) {
+    Bytes *vv1 = (Bytes *)op->v1, *vv2 = (Bytes *)op->v2;
     size_t i, len1, len2, sz;
     bool neg1, neg2;
     uint8_t *v1, *v2, *v;
@@ -864,15 +892,23 @@ static MUST_CHECK Obj *or_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint)
     len1 = byteslen(vv1); len2 = byteslen(vv2);
 
     if (len1 < len2) {
-        const Bytes *tmp = vv1; vv1 = vv2; vv2 = tmp;
+        Bytes *tmp = vv1; vv1 = vv2; vv2 = tmp;
         i = len1; len1 = len2; len2 = i;
     }
     neg1 = vv1->len < 0; neg2 = vv2->len < 0;
 
     sz = neg2 ? len2 : len1;
     if (sz == 0) return val_reference((neg1 || neg2) ? &inv_bytes->v : &null_bytes->v);
-    vv = new_bytes2(sz);
-    if (vv == NULL) return (Obj *)new_error_mem(epoint);
+    if (op->inplace == &vv1->v) {
+        vv = ref_bytes(vv1);
+        if (vv->data != vv->u.val) vv->u.s.hash = -1;
+    } else if (op->inplace == &vv2->v && sz <= len2) {
+        vv = ref_bytes(vv2);
+        if (vv->data != vv->u.val) vv->u.s.hash = -1;
+    } else {
+        vv = new_bytes2(sz);
+        if (vv == NULL) return (Obj *)new_error_mem(op->epoint3);
+    }
     v = vv->data;
     v1 = vv1->data; v2 = vv2->data;
 
@@ -898,7 +934,8 @@ static MUST_CHECK Obj *or_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint)
     return &vv->v;
 }
 
-static MUST_CHECK Obj *xor_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint) {
+static inline MUST_CHECK Obj *xor_(oper_t op) {
+    Bytes *vv1 = (Bytes *)op->v1, *vv2 = (Bytes *)op->v2;
     size_t i, len1, len2, sz;
     bool neg1, neg2;
     uint8_t *v1, *v2, *v;
@@ -906,15 +943,23 @@ static MUST_CHECK Obj *xor_(const Bytes *vv1, const Bytes *vv2, linepos_t epoint
     len1 = byteslen(vv1); len2 = byteslen(vv2);
 
     if (len1 < len2) {
-        const Bytes *tmp = vv1; vv1 = vv2; vv2 = tmp;
+        Bytes *tmp = vv1; vv1 = vv2; vv2 = tmp;
         i = len1; len1 = len2; len2 = i;
     }
     neg1 = vv1->len < 0; neg2 = vv2->len < 0;
 
     sz = len1;
     if (sz == 0) return val_reference((neg1 != neg2) ? &inv_bytes->v : &null_bytes->v);
-    vv = new_bytes2(sz);
-    if (vv == NULL) return (Obj *)new_error_mem(epoint);
+    if (op->inplace == &vv1->v) {
+        vv = ref_bytes(vv1);
+        if (vv->data != vv->u.val) vv->u.s.hash = -1;
+    } else if (op->inplace == &vv2->v && sz <= len2) {
+        vv = ref_bytes(vv2);
+        if (vv->data != vv->u.val) vv->u.s.hash = -1;
+    } else {
+        vv = new_bytes2(sz);
+        if (vv == NULL) return (Obj *)new_error_mem(op->epoint3);
+    }
     v = vv->data;
     v1 = vv1->data; v2 = vv2->data;
 
@@ -942,11 +987,11 @@ static MUST_CHECK Obj *concat(oper_t op) {
     len2 = byteslen(v2);
     ln = len1 + len2;
     if (ln < len2 || ln > SSIZE_MAX) goto failed; /* overflow */
-
     if (op->inplace == &v1->v) {
         s = extend_bytes(v1, ln);
         if (s == NULL) goto failed;
         v = ref_bytes(v1);
+        if (v->data != v->u.val) v->u.s.hash = -1;
     } else {
         v = new_bytes2(ln);
         if (v == NULL) goto failed;
@@ -962,7 +1007,8 @@ failed:
     return (Obj *)new_error_mem(op->epoint3);
 }
 
-static int icmp(Bytes *v1, Bytes *v2) {
+static int icmp(oper_t op) {
+    const Bytes *v1 = (Bytes *)op->v1, *v2 = (Bytes *)op->v2;
     size_t len1 = byteslen(v1), len2 = byteslen(v2);
     int h = memcmp(v1->data, v2->data, (len1 < len2) ? len1 : len2);
     if (h != 0) return h;
@@ -1018,7 +1064,7 @@ static MUST_CHECK Obj *calc1(oper_t op) {
     case O_STRING:
         tmp = int_from_bytes(v1, op->epoint);
         op->v1 = tmp;
-        op->inplace = NULL;
+        op->inplace = (tmp->refcount == 1) ? tmp : NULL;
         v = tmp->obj->calc1(op);
         val_destroy(tmp);
         return v;
@@ -1046,38 +1092,38 @@ static MUST_CHECK Obj *calc2_bytes(oper_t op) {
             tmp2 = int_from_bytes(v2, op->epoint2);
             op->v1 = tmp;
             op->v2 = tmp2;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp2);
             val_destroy(tmp);
             return result;
         }
-    case O_AND: return and_(v1, v2, op->epoint3);
-    case O_OR: return or_(v1, v2, op->epoint3);
-    case O_XOR: return xor_(v1, v2, op->epoint3);
+    case O_AND: return and_(op);
+    case O_OR: return or_(op);
+    case O_XOR: return xor_(op);
     case O_LSHIFT:
     case O_RSHIFT:
         {
             Obj *tmp, *result;
             tmp = bits_from_bytes(v1, op->epoint);
             op->v1 = tmp;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp);
             return result;
         }
     case O_CMP:
-        val = icmp(v1, v2);
+        val = icmp(op);
         if (val < 0) return (Obj *)ref_int(minus1_value);
         return (Obj *)ref_int(int_value[(val > 0) ? 1 : 0]);
-    case O_EQ: return truth_reference(icmp(v1, v2) == 0);
-    case O_NE: return truth_reference(icmp(v1, v2) != 0);
+    case O_EQ: return truth_reference(icmp(op) == 0);
+    case O_NE: return truth_reference(icmp(op) != 0);
     case O_MIN:
-    case O_LT: return truth_reference(icmp(v1, v2) < 0);
-    case O_LE: return truth_reference(icmp(v1, v2) <= 0);
+    case O_LT: return truth_reference(icmp(op) < 0);
+    case O_LE: return truth_reference(icmp(op) <= 0);
     case O_MAX:
-    case O_GT: return truth_reference(icmp(v1, v2) > 0);
-    case O_GE: return truth_reference(icmp(v1, v2) >= 0);
+    case O_GT: return truth_reference(icmp(op) > 0);
+    case O_GE: return truth_reference(icmp(op) >= 0);
     case O_CONCAT: return concat(op);
     case O_IN:
         {
@@ -1141,11 +1187,11 @@ static inline MUST_CHECK Obj *repeat(oper_t op) {
     return (Obj *)new_error_mem(op->epoint3);
 }
 
-static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
+static MUST_CHECK Obj *slice(oper_t op, size_t indx) {
     uint8_t *p2;
     size_t offs2, len1;
     size_t i;
-    Bytes *v, *v1 = (Bytes *)o1;
+    Bytes *v, *v1 = (Bytes *)op->v1;
     Obj *o2 = op->v2;
     Error *err;
     uint8_t inv = (v1->len < 0) ? (uint8_t)~0 : 0;
@@ -1153,40 +1199,37 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
     linepos_t epoint2;
 
     if (args->len < 1 || args->len > indx + 1) {
-        err_msg_argnum(args->len, 1, indx + 1, op->epoint2);
-        return (Obj *)ref_none();
+        return (Obj *)new_error_argnum(args->len, 1, indx + 1, op->epoint2);
     }
     o2 = args->val[indx].val;
     epoint2 = &args->val[indx].epoint;
 
     len1 = byteslen(v1);
 
-    if (o2->obj == LIST_OBJ) {
-        iter_next_t iter_next;
-        Iter *iter = o2->obj->getiter(o2);
-        size_t len2 = iter->len(iter);
+    if (o2->obj->iterable) {
+        struct iter_s iter;
+        iter.data = o2; o2->obj->getiter(&iter);
 
-        if (len2 == 0) {
-            val_destroy(&iter->v);
+        if (iter.len == 0) {
+            iter_destroy(&iter);
             return (Obj *)ref_bytes(null_bytes);
         }
-        v = new_bytes2(len2);
+        v = new_bytes2(iter.len);
         if (v == NULL) {
-            val_destroy(&iter->v);
+            iter_destroy(&iter);
             goto failed;
         }
         p2 = v->data;
-        iter_next = iter->next;
-        for (i = 0; i < len2 && (o2 = iter_next(iter)) != NULL; i++) {
+        for (i = 0; i < iter.len && (o2 = iter.next(&iter)) != NULL; i++) {
             err = indexoffs(o2, len1, &offs2, epoint2);
             if (err != NULL) {
                 val_destroy(&v->v);
-                val_destroy(&iter->v);
+                iter_destroy(&iter);
                 return &err->v;
             }
             p2[i] = v1->data[offs2] ^ inv;
         }
-        val_destroy(&iter->v);
+        iter_destroy(&iter);
         if (i > SSIZE_MAX) goto failed2; /* overflow */
         v->len = (ssize_t)i;
         return &v->v;
@@ -1198,26 +1241,55 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
         err = (Error *)sliceparams((Colonlist *)o2, len1, &length, &offs, &end, &step, epoint2);
         if (err != NULL) return &err->v;
 
-        if (length == 0) {
+        switch (length) {
+        case 0:
             return (Obj *)ref_bytes(null_bytes);
+        case 1:
+            return (Obj *)bytes_from_u8(v1->data[offs] ^ inv);
         }
         if (step == 1 && inv == 0) {
             if (length == byteslen(v1)) {
                 return (Obj *)ref_bytes(v1); /* original bytes */
             }
-            if (length == 1) return (Obj *)bytes_from_u8(v1->data[offs]);
-            v = new_bytes2(length);
-            if (v == NULL) goto failed;
-            memcpy(v->data, v1->data + offs, length);
+            if (op->inplace == &v1->v) {
+                v = ref_bytes(v1);
+                if (v->data != v->u.val && length <= sizeof v->u.val) {
+                    p2 = v->u.val;
+                    memcpy(p2, v1->data + offs, length);
+                } else {
+                    p2 = v->data;
+                    if (offs != 0) memmove(p2, v1->data + offs, length);
+                    if (v->data != v->u.val) v->u.s.hash = -1;
+                }
+            } else {
+                v = new_bytes2(length);
+                if (v == NULL) goto failed;
+                p2 = v->data;
+                memcpy(p2, v1->data + offs, length);
+            }
         } else {
-            v = new_bytes2(length);
-            if (v == NULL) goto failed;
-            p2 = v->data;
+            if (step > 0 && op->inplace == &v1->v) {
+                v = ref_bytes(v1);
+                if (v->data != v->u.val && length <= sizeof v->u.val) {
+                    p2 = v->u.val;
+                } else {
+                    p2 = v->data;
+                    if (v->data != v->u.val) v->u.s.hash = -1;
+                }
+            } else {
+                v = new_bytes2(length);
+                if (v == NULL) goto failed;
+                p2 = v->data;
+            }
             for (i = 0; i < length; i++) {
                 p2[i] = v1->data[offs] ^ inv;
                 offs += step;
             }
         }
+        if (p2 != v->data) {
+            free(v->data);
+            v->data = p2;
+        }
         v->len = (ssize_t)length;
         return &v->v;
     }
@@ -1247,6 +1319,11 @@ static MUST_CHECK Obj *calc2(oper_t op) {
         if (diagnostics.strict_bool) err_msg_bool_oper(op);
         return val_reference(to_bool(v1) ? &v1->v : o2);
     }
+    if (o2->obj->iterable) {
+        if (op->op != &o_MEMBER) {
+            return o2->obj->rcalc2(op);
+        }
+    }
     switch (o2->obj->type) {
     case T_BYTES: return calc2_bytes(op);
     case T_BOOL:
@@ -1269,16 +1346,13 @@ static MUST_CHECK Obj *calc2(oper_t op) {
             default: tmp = int_from_bytes(v1, op->epoint);
             }
             op->v1 = tmp;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp);
             return result;
         }
-    case T_TUPLE:
-    case T_LIST:
     case T_STR:
     case T_GAP:
-    case T_REGISTER:
     case T_DICT:
         if (op->op != &o_MEMBER) {
             return o2->obj->rcalc2(op);
@@ -1347,6 +1421,8 @@ void bytesobj_init(void) {
     obj.ival = ival;
     obj.uval = uval;
     obj.uval2 = uval2;
+    obj.iaddress = ival;
+    obj.uaddress = uval;
     obj.sign = sign;
     obj.function = function;
     obj.len = len;
