@@ -1,5 +1,5 @@
 /*
-    $Id: eval.c 1897 2019-02-17 20:50:04Z soci $
+    $Id: eval.c 2112 2019-12-11 17:56:46Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -52,66 +52,42 @@
 #include "errorobj.h"
 #include "identobj.h"
 #include "foldobj.h"
+#include "memblocksobj.h"
 
-size_t get_label(void) {
-    uchar_t ch;
+static FAST_CALL NO_INLINE unsigned int get_label_start(const uint8_t *s) {
     unsigned int l;
-    const struct properties_s *prop;
-    const uint8_t *s = pline + lpoint.pos;
-    const uint8_t *e;
-    static const uint8_t typ[256] = {
-        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 0 */
-        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 1 */
-        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, /* 2 */
-        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, /* 3 */
-        0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 4 */
-        2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 2, /* 5 */
-        0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, /* 6 */
-        2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, /* 7 */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 8 */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* 9 */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* a */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* b */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* c */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* d */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, /* e */
-        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3  /* f */
-    };
+    uchar_t ch;
+    if (!arguments.to_ascii) return 0;
+    l = utf8in(s, &ch);
+    return ((uget_property(ch)->property & id_Start) != 0) ? l : 0;
+}
 
-    switch (typ[*s]) {
-    default:
-    case 1:
-    case 0: return 0;
-    case 2:
-        e = s;
-        s++;
-        break;
-    case 3:
-        if (!arguments.to_ascii) return 0;
-        l = utf8in(s, &ch);
-        prop = uget_property(ch);
-        if ((prop->property & id_Start) == 0) return 0;
-        e = s;
-        s += l;
-    }
+static FAST_CALL NO_INLINE unsigned int get_label_continue(const uint8_t *s) {
+    unsigned int l;
+    uchar_t ch;
+    if (!arguments.to_ascii) return 0;
+    l = utf8in(s, &ch);
+    return ((uget_property(ch)->property & (id_Continue | id_Start)) != 0) ? l : 0;
+}
+
+FAST_CALL size_t get_label(const uint8_t *s) {
+    size_t i;
+    if (((uint8_t)((*s | 0x20) - 'a')) > 'z' - 'a' && *s != '_') {
+        if (*s < 0x80) return 0;
+        i = get_label_start(s);
+        if (i == 0) return 0;
+    } else i = 1;
     for (;;) {
-        switch (typ[*s]) {
-        default:
-        case 0: break;
-        case 1:
-        case 2: s++; continue;
-        case 3:
-            if (!arguments.to_ascii) break;
-            l = utf8in(s, &ch);
-            prop = uget_property(ch);
-            if ((prop->property & (id_Continue | id_Start)) == 0) break;
-            s += l;
+        unsigned int l;
+        if (((uint8_t)((s[i] | 0x20) - 'a')) <= 'z' - 'a' || (s[i] ^ 0x30) < 10 || s[i] == '_') {
+            i++;
             continue;
         }
-        break;
+        if (s[i] < 0x80) return i;
+        l = get_label_continue(s + i);
+        if (l == 0) return i;
+        i += l;
     }
-    lpoint.pos = (linecpos_t)(s - pline);
-    return (size_t)(s - e);
 }
 
 static MUST_CHECK Obj *get_dec(linepos_t epoint) {
@@ -129,100 +105,111 @@ static double ldexp10(double d, unsigned int expo, bool neg) {
     return neg ? d / scal : d * scal;
 }
 
-static MUST_CHECK Obj *get_exponent(double real, linepos_t epoint) {
+static MUST_CHECK Obj *get_exponent(Obj *v1, Obj *v2, size_t len, linepos_t epoint) {
+    Obj *v;
+    double real;
+    uval_t expo = 0;
+    bool neg = false;
     uint8_t base = here() | 0x20;
 
     if (base == 'p' || base == 'e') {
-        bool neg = false;
         neg = (pline[lpoint.pos + 1] == '-');
         if (neg || pline[lpoint.pos + 1] == '+') {
             if ((pline[lpoint.pos + 2] ^ 0x30) < 10) lpoint.pos++;
         }
         if ((pline[lpoint.pos + 1] ^ 0x30) < 10) {
-            uval_t expo;
             Error *err;
-            size_t len, len2;
-            Obj *v;
+            size_t len1, len2;
             lpoint.pos++;
 
-            v = int_from_decstr(pline + lpoint.pos, &len, &len2, epoint);
+            v = int_from_decstr(pline + lpoint.pos, &len1, &len2, epoint);
             err = v->obj->uval(v, &expo, 8 * (sizeof expo < sizeof(int) ? sizeof expo : sizeof(int)) - 1, &lpoint);
             val_destroy(v);
-            lpoint.pos += len;
-            if (err != NULL) return &err->v;
-
-            if (expo != 0) real = (base == 'e') ? ldexp10(real, expo, neg) : ldexp(real, neg ? -(ival_t)expo : (ival_t)expo);
+            lpoint.pos += len1;
+            if (err != NULL) {
+                if (v2 != NULL) val_destroy(v2);
+                val_destroy(v1);
+                return &err->v;
+            }
         }
     }
-    return float_from_double(real, epoint);
-}
-
-static double toreal_destroy(Obj *v, linepos_t epoint) {
-    Obj *err = FLOAT_OBJ->create(v, epoint);
-    double real;
-    if (err->obj != FLOAT_OBJ) {
-        if (err == &none_value->v) err_msg_still_none(NULL, epoint);
-        else if (err->obj == ERROR_OBJ) err_msg_output((Error *)err);
-        real = 0;
+    if (v2 == NULL) {
+        real = 0.0;
     } else {
-        real = ((Float *)err)->real;
+        bool bits = v2->obj == BITS_OBJ;
+        if (bits) {
+            len = (size_t)((Bits *)v2)->bits;
+        }
+        v = FLOAT_OBJ->create(v2, epoint);
+        val_destroy(v2);
+        if (v->obj != FLOAT_OBJ) {
+            val_destroy(v1);
+            return v;
+        }
+        real = ((Float *)v)->real;
+        if (len != 0 && real != 0.0) real = bits ? ldexp(real, -(int)len) : ldexp10(real, (unsigned int)len, true);
+        val_destroy(v);
+    }
+    v = FLOAT_OBJ->create(v1, epoint);
+    val_destroy(v1);
+    if (v->obj != FLOAT_OBJ) {
+        return v;
+    }
+    real += ((Float *)v)->real;
+    if (expo != 0) {
+        real = (base == 'p') ? ldexp(real, neg ? -(ival_t)expo : (ival_t)expo) : ldexp10(real, expo, neg);
+    }
+    if (real == HUGE_VAL || real == -HUGE_VAL) {
+        val_destroy(v);
+        return (Obj *)new_error(ERROR_NUMERIC_OVERF, epoint);
+    }
+    if (v->refcount == 1) {
+        ((Float *)v)->real = real;
+        return v;
     }
-    val_destroy(err);
     val_destroy(v);
-    return real;
+    return (Obj *)new_float(real);
 }
 
 static MUST_CHECK Obj *get_exponent2(Obj *v, linepos_t epoint) {
     if (pline[lpoint.pos + 1] == '-' || pline[lpoint.pos + 1] == '+') {
         if ((pline[lpoint.pos + 2] ^ 0x30) < 10) {
-            return get_exponent(toreal_destroy(v, &lpoint), epoint);
+            return get_exponent(v, NULL, 0, epoint);
         }
     } else if ((pline[lpoint.pos + 1] ^ 0x30) < 10) {
-        return get_exponent(toreal_destroy(v, &lpoint), epoint);
+        return get_exponent(v, NULL, 0, epoint);
     }
     return v;
 }
 
 static MUST_CHECK Obj *get_hex(linepos_t epoint) {
-    size_t len, len2;
-    Obj *v;
+    size_t len;
+    Obj *v, *v2;
 
     v = bits_from_hexstr(pline + lpoint.pos + 1, &len, epoint);
     lpoint.pos += len + 1;
     if (here() == '.' && pline[lpoint.pos + 1] != '.') {
-        double real, real2;
         lpoint.pos++;
-        real = toreal_destroy(v, epoint);
 
-        v = bits_from_hexstr(pline + lpoint.pos, &len, epoint);
-        len2 = v->obj == BITS_OBJ ? ((Bits *)v)->bits : 0;
-        real2 = toreal_destroy(v, &lpoint);
+        v2 = bits_from_hexstr(pline + lpoint.pos, &len, epoint);
         lpoint.pos += len;
-
-        if (real2 != 0.0) real += ldexp(real2, -(int)len2);
-        return get_exponent(real, epoint);
+        return get_exponent(v, v2, 0, epoint);
     }
     return (here() | 0x20) == 'p' ? get_exponent2(v, epoint) : v;
 }
 
 static MUST_CHECK Obj *get_bin(linepos_t epoint) {
-    size_t len, len2;
-    Obj *v;
+    size_t len;
+    Obj *v, *v2;
 
     v = bits_from_binstr(pline + lpoint.pos + 1, &len, epoint);
     lpoint.pos += len + 1;
     if (here() == '.' && pline[lpoint.pos + 1] != '.') {
-        double real, real2;
         lpoint.pos++;
-        real = toreal_destroy(v, epoint);
 
-        v = bits_from_binstr(pline + lpoint.pos, &len, epoint);
-        len2 = v->obj == BITS_OBJ ? ((Bits *)v)->bits : 0;
-        real2 = toreal_destroy(v, &lpoint);
+        v2 = bits_from_binstr(pline + lpoint.pos, &len, epoint);
         lpoint.pos += len;
-
-        if (real2 != 0.0) real += ldexp(real2, -(int)len2);
-        return get_exponent(real, epoint);
+        return get_exponent(v, v2, 0, epoint);
     }
     switch (here() | 0x20) {
     case 'e':
@@ -235,21 +222,17 @@ static MUST_CHECK Obj *get_bin(linepos_t epoint) {
 
 static MUST_CHECK Obj *get_float(linepos_t epoint) {
     size_t len, len2;
-    Obj *v;
+    Obj *v, *v2;
 
     v = int_from_decstr(pline + lpoint.pos, &len, &len2, epoint);
     lpoint.pos += len;
     if (here() == '.' && pline[lpoint.pos + 1] != '.') {
-        double real, real2;
         lpoint.pos++;
-        real = toreal_destroy(v, epoint);
 
-        v = int_from_decstr(pline + lpoint.pos, &len, &len2, epoint);
-        real2 = toreal_destroy(v, &lpoint);
+        v2 = int_from_decstr(pline + lpoint.pos, &len, &len2, epoint);
         lpoint.pos += len;
 
-        if (real2 != 0.0) real += ldexp10(real2, (unsigned int)len2, true);
-        return get_exponent(real, epoint);
+        return get_exponent(v, v2, len2, epoint);
     }
     switch (here() | 0x20) {
     case 'e':
@@ -303,41 +286,56 @@ void touch_label(Label *tmp) {
     tmp->usepass = pass;
 }
 
-static uval_t bitscalc(Bits *val) {
+static uval_t bitscalc(address_t addr, Bits *val) {
     size_t b = val->bits;
-    if (b >= 8 * sizeof(star)) return (uval_t)b;
-    if ((star >> b) == 0) return (uval_t)b;
-    if (star <= 0xff) return 8;
-    if (star <= 0xffff) return 16;
+    if (b >= 8 * sizeof(addr)) return (uval_t)b;
+    if ((addr >> b) == 0) return (uval_t)b;
+    if (addr <= 0xff) return 8;
+    if (addr <= 0xffff) return 16;
     return all_mem_bits;
 }
 
-static uval_t bytescalc(Bytes *val) {
+static uval_t bytescalc(address_t addr, Bytes *val) {
     size_t b = val->len < 0 ? (size_t)~val->len : (size_t)val->len;
-    if (b >= 8 * sizeof(star)) return (uval_t)b;
-    if ((star >> (b << 3)) == 0) return (uval_t)b;
-    if (star <= 0xff) return 1;
-    if (star <= 0xffff) return 2;
+    if (b >= 8 * sizeof(addr)) return (uval_t)b;
+    if ((addr >> (b << 3)) == 0) return (uval_t)b;
+    if (addr <= 0xff) return 1;
+    if (addr <= 0xffff) return 2;
     return all_mem_bits >> 3;
 }
 
-MUST_CHECK Obj *get_star_value(Obj *val) {
+MUST_CHECK Obj *get_star_value(address_t addr, Obj *val) {
     switch (val->obj->type) {
-    case T_BITS: return (Obj *)bits_from_uval(star, bitscalc((Bits *)val));
-    case T_CODE: return get_star_value(((Code *)val)->addr);
+    case T_BITS: return (Obj *)bits_from_uval(addr, bitscalc(addr, (Bits *)val));
+    case T_CODE: return get_star_value(addr, ((Code *)val)->typ);
     default:
     case T_BOOL:
-    case T_INT: return (Obj *)int_from_uval(star);
-    case T_FLOAT: return (Obj *)new_float(star + (((Float *)val)->real - trunc(((Float *)val)->real)));
-    case T_STR: return (Obj *)bytes_from_uval(star, all_mem_bits >> 3);
-    case T_BYTES: return (Obj *)bytes_from_uval(star, bytescalc((Bytes *)val));
-    case T_ADDRESS: return (Obj *)new_address(get_star_value(((Address *)val)->val), ((Address *)val)->type);
+    case T_INT: return (Obj *)int_from_uval(addr);
+    case T_FLOAT: return (Obj *)new_float(addr + (((Float *)val)->real - trunc(((Float *)val)->real)));
+    case T_STR: return (Obj *)bytes_from_uval(addr, all_mem_bits >> 3);
+    case T_BYTES: return (Obj *)bytes_from_uval(addr, bytescalc(addr, (Bytes *)val));
+    case T_ADDRESS: return (Obj *)new_address(get_star_value(addr, ((Address *)val)->val), ((Address *)val)->type);
     }
 }
 
-static MUST_CHECK Obj *get_star(void) {
+MUST_CHECK Obj *get_star(void) {
+    Code *code;
     if (diagnostics.optimize) cpu_opt_invalidate();
-    return get_star_value(current_address->l_address_val);
+    code = new_code();
+    code->addr = star;
+    code->typ = val_reference(current_address->l_address_val);
+    code->size = 0;
+    code->offs = 0;
+    code->dtype = D_NONE;
+    code->pass = pass;
+    code->apass = pass;
+    code->memblocks = ref_memblocks(current_address->mem);
+    code->names = ref_namespace(current_context);
+    code->requires = current_section->requires;
+    code->conflicts = current_section->conflicts;
+    code->memaddr = current_address->address;
+    code->membp = 0;
+    return &code->v;
 }
 
 static size_t evxnum, evx_p;
@@ -431,6 +429,7 @@ rest:
     for (;;) {
         Oper *op;
         Obj *val;
+        size_t ln;
         ignore();ch = here(); epoint = lpoint;
 
         switch (ch) {
@@ -445,11 +444,13 @@ rest:
         case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
             push_oper(get_dec(&epoint), &epoint);goto other;
         default:
-            if (get_label() == 0) {
+            ln = get_label(pline + lpoint.pos);
+            if (ln == 0) {
                 if (opr.p != 0) epoint = opr.data[opr.p - 1].epoint;
                 err_msg2(ERROR______EXPECTED, "an expression is", &lpoint);
                 goto error;
             }
+            lpoint.pos += ln;
             break;
         }
     as_ident:
@@ -481,7 +482,8 @@ rest:
         switch (ch) {
         case ',':
             lpoint.pos++;
-            llen = get_label();
+            llen = get_label(pline + lpoint.pos);
+            lpoint.pos += llen;
             if (llen == 1) {
                 switch (pline[epoint.pos + 1] | arguments.caseinsensitive) {
                 case 'x':
@@ -762,8 +764,7 @@ MUST_CHECK Obj *sliceparams(const struct List *v2, size_t len2, uval_t *olen, iv
     if (len2 >= (1u << (8 * sizeof(ival_t) - 1))) return (Obj *)new_error_mem(epoint); /* overflow */
     len = (ival_t)len2;
     if (v2->len > 3 || v2->len < 1) {
-        err_msg_argnum(v2->len, 1, 3, epoint);
-        return (Obj *)ref_none();
+        return (Obj *)new_error_argnum(v2->len, 1, 3, epoint);
     }
     end = len;
     if (v2->len > 2) {
@@ -816,41 +817,40 @@ MUST_CHECK Obj *sliceparams(const struct List *v2, size_t len2, uval_t *olen, iv
     return NULL;
 }
 
-static MUST_CHECK Obj *apply_addressing(Obj *o1, Address_types am) {
-    switch (o1->obj->type) {
-    case T_ADDRESS:
-        {
-            Address *v1 = (Address *)o1;
-            return (Obj *)new_address(val_reference(v1->val), am | (v1->type << 4));
-        }
-    case T_LIST:
-    case T_TUPLE:
-        {
-            iter_next_t iter_next;
-            Iter *iter = o1->obj->getiter(o1);
-            size_t i, len = iter->len(iter);
-            List *v;
-            Obj **vals;
+static MUST_CHECK Obj *apply_addressing(Obj *o1, Address_types am, bool inplace) {
+    if (o1->obj->iterable) {
+        struct iter_s iter;
+        size_t i;
+        List *v;
+        Obj **vals;
 
-            if (len == 0) {
-                val_destroy(&iter->v);
-                return val_reference(o1->obj == TUPLE_OBJ ? &null_tuple->v : &null_list->v);
-            }
+        if (o1->refcount != 1) inplace = false;
 
-            v = (List *)val_alloc(o1->obj == TUPLE_OBJ ? TUPLE_OBJ : LIST_OBJ);
-            vals = list_create_elements(v, len);
-            iter_next = iter->next;
-            for (i = 0; i < len && (o1 = iter_next(iter)) != NULL; i++) {
-                vals[i] = apply_addressing(o1, am);
-            }
-            val_destroy(&iter->v);
-            v->len = i;
-            v->data = vals;
-            return &v->v;
+        iter.data = o1; o1->obj->getiter(&iter);
+        if (iter.len == 0) {
+            iter_destroy(&iter);
+            return val_reference(o1->obj == TUPLE_OBJ ? &null_tuple->v : &null_list->v);
         }
-    default:
-        return (Obj *)new_address(val_reference(o1), am);
+
+        v = (List *)val_alloc(o1->obj == TUPLE_OBJ ? TUPLE_OBJ : LIST_OBJ);
+        vals = list_create_elements(v, iter.len);
+        for (i = 0; i < iter.len && (o1 = iter.next(&iter)) != NULL; i++) {
+            vals[i] = apply_addressing(o1, am, inplace);
+        }
+        iter_destroy(&iter);
+        v->len = i;
+        v->data = vals;
+        return &v->v;
     }
+    if (o1->obj == ADDRESS_OBJ) {
+        Address *v1 = (Address *)o1;
+        if (inplace && o1->refcount == 1) {
+            v1->type = am | (v1->type << 4);
+            return val_reference(o1);
+        }
+        return (Obj *)new_address(val_reference(v1->val), am | (v1->type << 4));
+    }
+    return (Obj *)new_address(val_reference(o1), am);
 }
 
 static bool get_val2(struct eval_context_s *ev) {
@@ -888,7 +888,7 @@ static bool get_val2(struct eval_context_s *ev) {
         case O_FUNC:
         case O_INDEX:
             {
-                unsigned int args = 0;
+                size_t args = 0;
                 Funcargs tmp;
                 op = (op == O_FUNC) ? O_PARENT : O_BRACKET;
                 while (v1->val->obj != OPER_OBJ || ((Oper *)v1->val)->op != op) {
@@ -908,12 +908,17 @@ static bool get_val2(struct eval_context_s *ev) {
                 oper.epoint = &v1->epoint;
                 oper.epoint2 = (args != 0) ? &tmp.val->epoint : &o_out->epoint;
                 oper.epoint3 = &o_out->epoint;
-                oper.inplace = NULL;
                 if (op == O_BRACKET) {
-                    v1->val = oper.v1->obj->slice(oper.v1, &oper, 0);
+                    oper.inplace = (oper.v1->refcount == 1) ? oper.v1 : NULL;
+                    v1->val = oper.v1->obj->slice(&oper, 0);
                 } else {
+                    oper.inplace = NULL;
                     v1->val = oper.v1->obj->calc2(&oper);
                 }
+                while ((args--) != 0) {
+                    val_destroy(tmp.val[args].val);
+                    tmp.val[args].val = NULL;
+                }
                 continue;
             }
         case O_RBRACKET:
@@ -950,7 +955,9 @@ static bool get_val2(struct eval_context_s *ev) {
                             }
                         }
                         am = (op == O_BRACKET) ? A_LI : A_I;
-                        v1->val = apply_addressing(values[vsp].val, am);
+                        v1->val = apply_addressing(values[vsp].val, am, true);
+                        val_destroy(values[vsp].val);
+                        values[vsp].val = NULL;
                         continue;
                     }
                     if (tup) {
@@ -977,7 +984,7 @@ static bool get_val2(struct eval_context_s *ev) {
         case O_RBRACE:
         case O_DICT:
             {
-                unsigned int args = 0;
+                size_t args = 0;
                 while (v1->val->obj != OPER_OBJ || ((Oper *)v1->val)->op != O_BRACE) {
                     args++;
                     if (vsp <= args) goto syntaxe;
@@ -1088,7 +1095,7 @@ static bool get_val2(struct eval_context_s *ev) {
         case O_POS:     /* +  */
         case O_LNOT:    /* !  */
             oper.v1 = v1->val;
-            oper.v2 = NULL;
+            oper.v2 = &none_value->v;
             oper.epoint = &v1->epoint;
             oper.epoint3 = &o_out->epoint;
             oper.inplace = (oper.v1->refcount == 1) ? oper.v1 : NULL;
@@ -1107,8 +1114,12 @@ static bool get_val2(struct eval_context_s *ev) {
         case O_HASH_SIGNED: am = A_IMMEDIATE_SIGNED; goto addr; /* #+ */
         case O_HASH: am = A_IMMEDIATE;                          /* #  */
         addr:
-            val = apply_addressing(v1->val, am);
-            val_destroy(v1->val); v1->val = val;
+            if (v1->val->obj != ADDRESS_OBJ && !v1->val->obj->iterable) {
+                v1->val = (Obj *)new_address(v1->val, am);
+            } else {
+                val = apply_addressing(v1->val, am, true);
+                val_destroy(v1->val); v1->val = val;
+            }
             if (op == O_HASH || op == O_HASH_SIGNED) v1->epoint = o_out->epoint;
             continue;
         case O_SPLAT:   /* *  */
@@ -1120,61 +1131,40 @@ static bool get_val2(struct eval_context_s *ev) {
                     continue;
                 }
             }
-            if (v1->val->obj == TUPLE_OBJ || v1->val->obj == LIST_OBJ || v1->val->obj == ADDRLIST_OBJ) {
-                iter_next_t iter_next;
-                Iter *iter = v1->val->obj->getiter(v1->val);
-                size_t k, len = iter->len(iter);
-                size_t len2 = vsp + len;
-                Obj *tmp;
+            if (v1->val->obj->iterable || v1->val->obj == ADDRLIST_OBJ) {
+                struct iter_s iter;
+                size_t k, len;
+                size_t len2;
+                Obj *tmp, *def;
+                iter.data = v1->val; v1->val->obj->getiter(&iter);
+                len = iter.len;
 
+                if (v1->val->obj == DICT_OBJ && ((Dict *)v1->val)->def != NULL) {
+                    Colonlist *list = new_colonlist();
+                    list->len = 2;
+                    list->data = list->u.val;
+                    list->data[0] = (Obj *)ref_default();
+                    list->data[1] = val_reference(((Dict *)v1->val)->def);
+                    def = &list->v;
+                    len++;
+                    if (len < 1) err_msg_out_of_memory(); /* overflow */
+                } else def = NULL;
+
+                len2 = vsp + len;
                 if (len2 < len) err_msg_out_of_memory(); /* overflow */
                 vsp--;
                 if (len2 >= ev->values_size) values = extend_values(ev, len);
-                iter_next = iter->next;
-                for (k = 0; k < len && (tmp = iter_next(iter)) != NULL; k++) {
+                for (k = 0; k < len && (tmp = iter.next(&iter)) != NULL; k++) {
                     if (values[vsp].val != NULL) val_destroy(values[vsp].val);
                     values[vsp].val = val_reference(tmp);
                     values[vsp++].epoint = o_out->epoint;
                 }
-                val_destroy(&iter->v);
-                continue;
-            }
-            if (v1->val->obj == DICT_OBJ) {
-                Dict *tmp = (Dict *)v1->val;
-                const struct avltree_node *n;
-                size_t len = (tmp->def == NULL) ? tmp->len : tmp->len + 1;
-                size_t len2 = vsp + len;
-                if (len < tmp->len || len2 < len) err_msg_out_of_memory(); /* overflow */
-                v1->val = NULL;
-                vsp--;
-                if (len2 >= ev->values_size) values = extend_values(ev, len);
-                for (n = avltree_first(&tmp->members); n != NULL; n = avltree_next(n)) {
-                    const struct pair_s *p = cavltree_container_of(n, struct pair_s, node);
-                    if (p->data == NULL) values[vsp].val = val_reference(p->key);
-                    else {
-                        Colonlist *list = new_colonlist();
-                        list->len = 2;
-                        list->data = list_create_elements(list, 2);
-                        list->data[0] = val_reference(p->key);
-                        list->data[1] = val_reference(p->data);
-
-                        if (values[vsp].val != NULL) val_destroy(values[vsp].val);
-                        values[vsp].val = (Obj *)list;
-                    }
-                    values[vsp++].epoint = o_out->epoint;
-                }
-                if (tmp->def != NULL) {
-                    Colonlist *list = new_colonlist();
-                    list->len = 2;
-                    list->data = list_create_elements(list, 2);
-                    list->data[0] = (Obj *)ref_default();
-                    list->data[1] = val_reference(tmp->def);
-
+                iter_destroy(&iter);
+                if (def != NULL) {
                     if (values[vsp].val != NULL) val_destroy(values[vsp].val);
-                    values[vsp].val = (Obj *)list;
+                    values[vsp].val = def;
                     values[vsp++].epoint = o_out->epoint;
                 }
-                val_destroy(&tmp->v);
                 continue;
             }
             v1->epoint = o_out->epoint;
@@ -1392,9 +1382,10 @@ static bool get_exp2(int stop) {
                 }
                 goto tryanon;
             }
-            /* fall through */
+            push_oper(get_float(&epoint), &epoint);
+            goto other;
         case '0':
-            if (diagnostics.leading_zeros && pline[lpoint.pos + 1] >= '0' && pline[lpoint.pos + 1] <= '9') err_msg2(ERROR_LEADING_ZEROS, NULL, &lpoint);
+            if (diagnostics.leading_zeros && (pline[lpoint.pos + 1] ^ 0x30) < 10) err_msg2(ERROR_LEADING_ZEROS, NULL, &lpoint);
             /* fall through */
         case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
             push_oper(get_float(&epoint), &epoint);
@@ -1409,11 +1400,13 @@ static bool get_exp2(int stop) {
             }
             goto tryanon;
         default:
-            if (get_label() != 0) {
+            llen = get_label(pline + lpoint.pos);
+            if (llen != 0) {
                 bool down;
                 Label *l;
                 Obj *val;
                 str_t ident;
+                lpoint.pos += llen;
             as_ident:
                 if (pline[epoint.pos + 1] == '"' || pline[epoint.pos + 1] == '\'') {
                     Textconv_types mode;
@@ -1560,43 +1553,45 @@ static bool get_exp2(int stop) {
         switch (ch) {
         case ',':
             lpoint.pos++;
-            llen = get_label();
-            if (llen == 1 && pline[epoint.pos + 2] != '"' && pline[epoint.pos + 2] != '\'') {
-                switch (pline[epoint.pos + 1] | arguments.caseinsensitive) {
-                case 'x': op = &o_COMMAX; break;
-                case 'y': op = &o_COMMAY; break;
-                case 'z': op = &o_COMMAZ; break;
-                case 'r': op = &o_COMMAR; break;
-                case 's': op = &o_COMMAS; break;
-                case 'd': op = &o_COMMAD; break;
-                case 'b': op = &o_COMMAB; break;
-                case 'k': op = &o_COMMAK; break;
-                default: op = &o_COMMA; break;
+            if (pline[lpoint.pos] >= 'A') {
+                llen = get_label(pline + lpoint.pos);
+                lpoint.pos += llen;
+                if (llen == 1 && pline[epoint.pos + 2] != '"' && pline[epoint.pos + 2] != '\'') {
+                    switch (pline[epoint.pos + 1] | arguments.caseinsensitive) {
+                    case 'x': op = &o_COMMAX; break;
+                    case 'y': op = &o_COMMAY; break;
+                    case 'z': op = &o_COMMAZ; break;
+                    case 'r': op = &o_COMMAR; break;
+                    case 's': op = &o_COMMAS; break;
+                    case 'd': op = &o_COMMAD; break;
+                    case 'b': op = &o_COMMAB; break;
+                    case 'k': op = &o_COMMAK; break;
+                    default: op = NULL; break;
+                    }
+                    if (op != NULL) {
+                        prec = o_HASH.prio;
+                        while (opr.p != 0 && prec <= opr.data[opr.p - 1].val->prio && opr.data[opr.p - 1].val != &o_COLON2 && opr.data[opr.p - 1].val != &o_COND) {
+                            opr.p--;
+                            push_oper((Obj *)opr.data[opr.p].val, &opr.data[opr.p].epoint);
+                        }
+                        opr.data[opr.p].epoint = epoint;
+                        opr.data[opr.p++].val = op;
+                        if (opr.p >= opr.l) extend_opr(&opr);
+                        goto other;
+                    } 
                 }
-            } else op = &o_COMMA;
-            prec = op->prio;
-            if (op == &o_COMMA) {
-                while (opr.p != 0 && prec <= opr.data[opr.p - 1].val->prio) {
-                    opr.p--;
-                    push_oper((Obj *)opr.data[opr.p].val, &opr.data[opr.p].epoint);
-                }
-            } else {
-                prec = o_HASH.prio;
-                while (opr.p != 0 && prec <= opr.data[opr.p - 1].val->prio && opr.data[opr.p - 1].val != &o_COLON2 && opr.data[opr.p - 1].val != &o_COND) {
-                    opr.p--;
-                    push_oper((Obj *)opr.data[opr.p].val, &opr.data[opr.p].epoint);
-                }
-                opr.data[opr.p].epoint = epoint;
-                opr.data[opr.p++].val = op;
-                if (opr.p >= opr.l) extend_opr(&opr);
-                goto other;
+            } else llen = 0;
+            prec = o_COMMA.prio;
+            while (opr.p != 0 && prec <= opr.data[opr.p - 1].val->prio) {
+                opr.p--;
+                push_oper((Obj *)opr.data[opr.p].val, &opr.data[opr.p].epoint);
             }
             if (opr.p == 0) {
                 if (stop == 1) {lpoint = epoint;break;}
             }
             push_oper(&o_COMMA.v, &epoint);
             opr.data[opr.p].epoint = epoint;
-            opr.data[opr.p++].val = op;
+            opr.data[opr.p++].val = &o_COMMA;
             if (opr.p >= opr.l) extend_opr(&opr);
             if (llen != 0) {
                 epoint.pos++;
@@ -1769,7 +1764,9 @@ static bool get_exp2(int stop) {
         case '\t':
         case ' ': break;
         default:
-            switch (get_label()) {
+            llen = get_label(pline + lpoint.pos);
+            lpoint.pos += llen;
+            switch (llen) {
             case 1: if ((pline[epoint.pos] | arguments.caseinsensitive) == 'x') {if (pline[lpoint.pos] == '=') {lpoint.pos++; op = &o_X_ASSIGN;} else op = &o_X;goto push2a;} break;
             case 2: if ((pline[epoint.pos] | arguments.caseinsensitive) == 'i' &&
                         (pline[epoint.pos + 1] | arguments.caseinsensitive) == 'n') {op = &o_IN;goto push2a;} break;
@@ -1861,7 +1858,7 @@ Obj *get_vals_addrlist(struct linepos_s *epoints) {
             }
             if (am != A_NONE) {
                 val_destroy(val2);
-                val2 = apply_addressing(list->data[i - 1], am);
+                val2 = apply_addressing(list->data[i - 1], am, true);
                 val_destroy(list->data[i - 1]);
                 list->data[i - 1] = val2;
                 continue;
