@@ -1,5 +1,5 @@
 /*
-    $Id: file.c 1873 2019-02-09 14:34:38Z soci $
+    $Id: file.c 2200 2020-04-07 19:18:23Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -24,6 +24,9 @@
 #include <locale.h>
 #include <windows.h>
 #endif
+#if defined _POSIX_C_SOURCE || defined __unix__ || defined __MINGW32__
+#include <sys/stat.h>
+#endif
 #include "64tass.h"
 #include "unicode.h"
 #include "error.h"
@@ -44,9 +47,53 @@ struct include_list_s {
 };
 
 static struct include_list_s include_list;
-struct include_list_s *include_list_last = &include_list;
+static struct include_list_s *include_list_last = &include_list;
 
-static struct avltree file_tree;
+static struct {
+    size_t len, mask;
+    struct file_s **data;
+} file_table;
+
+static struct file_s *file_table_update(struct file_s *p) {
+    size_t mask, hash, offs;
+    if (file_table.len * 3 / 2 >= file_table.mask) {
+        size_t i, max = (file_table.data == NULL) ? 8 : (file_table.mask + 1) << 1;
+        struct file_s **n = (struct file_s **)calloc(max, sizeof *n);
+        if (n == NULL) err_msg_out_of_memory();
+        mask = max - 1;
+        if (file_table.data != NULL) {
+            for (i = 0; i <= file_table.mask; i++) if (file_table.data[i] != NULL) {
+                hash = (size_t)file_table.data[i]->hash;
+                offs = hash & mask;
+                while (n[offs] != NULL) {
+                    hash >>= 5;
+                    offs = (5 * offs + hash + 1) & mask;
+                }
+                n[offs] = file_table.data[i];
+            }
+            free(file_table.data);
+        }
+        file_table.mask = mask;
+        file_table.data = n;
+    }
+    mask = file_table.mask;
+    hash = (size_t)p->hash;
+    offs = hash & mask;
+    while (file_table.data[offs] != NULL) {
+        struct file_s *d = file_table.data[offs];
+        if (p->hash == d->hash && p->type == d->type && strcmp(p->name, d->name) == 0) {
+            if (p->base.data == d->base.data) return d;
+            if (p->base.len == d->base.len && memcmp(p->base.data, d->base.data, p->base.len) == 0) {
+                return d;
+            }
+        }
+        hash >>= 5;
+        offs = (5 * offs + hash + 1) & mask;
+    } 
+    file_table.data[offs] = p;
+    file_table.len++;
+    return NULL;
+}
 
 void include_list_add(const char *path)
 {
@@ -59,7 +106,7 @@ void include_list_add(const char *path)
     if (path[i - 1] != '/') j++;
 #endif
     len = j + 1 + sizeof(struct include_list_s);
-    if (len < sizeof(struct include_list_s)) err_msg_out_of_memory();
+    if (len <= sizeof(struct include_list_s)) err_msg_out_of_memory();
     include_list_last->next = (struct include_list_s *)mallocx(len);
     include_list_last = include_list_last->next;
     include_list_last->next = NULL;
@@ -96,10 +143,8 @@ char *get_path(const str_t *v, const char *base) {
 #else
     i = (v->len != 0 && v->data[0] == '/') ? 0 : get_base(base);
 #endif
-    len = i + v->len;
-    if (len < i) err_msg_out_of_memory(); /* overflow */
-    len += 1;
-    if (len < 1) err_msg_out_of_memory(); /* overflow */
+    len = i + 1 + v->len;
+    if (len <= i) err_msg_out_of_memory(); /* overflow */
     path = (char *)mallocx(len);
     memcpy(path, base, i);
     memcpy(path + i, v->data, v->len);
@@ -177,17 +222,17 @@ static bool portability(const str_t *name, linepos_t epoint) {
         return false;
     }
 #else
-    size_t i;
-    const uint8_t *c = name->data;
-    for (i = 0; i < name->len; i++) {
-        if (strchr("\\:*?\"<>|", c[i]) != NULL) {
+    const char *c;
+    if (name->len > 0) {
+        for (c = "\\:*?\"<>|"; *c != '\0'; c++) {
+            if (memchr(name->data, *c, name->len) == NULL) continue;
             err_msg2(ERROR__RESERVED_CHR, name, epoint);
             return false;
         }
-    }
-    if (name->len > 0 && name->data[0] == '/') {
-        err_msg2(ERROR_ABSOLUTE_PATH, name, epoint);
-        return false;
+        if (name->data[0] == '/') {
+            err_msg2(ERROR_ABSOLUTE_PATH, name, epoint);
+            return false;
+        }
     }
 #endif
     return true;
@@ -239,14 +284,16 @@ FILE *file_open(const char *name, const char *mode) {
         }
         memcpy(newname + len - l, temp, (size_t)l);
     } while (ch != 0);
+    errno = 0;
     f = fopen(newname, mode);
+    if (f == NULL && errno == 0) errno = (mode[0] == 'r') ? ENOENT : EINVAL;
 failed:
     free(newname);
 #endif
     return f;
 }
 
-static int star_compare(const struct avltree_node *aa, const struct avltree_node *bb)
+static FAST_CALL int star_compare(const struct avltree_node *aa, const struct avltree_node *bb)
 {
     const struct star_s *a = cavltree_container_of(aa, struct star_s, node);
     const struct star_s *b = cavltree_container_of(bb, struct star_s, node);
@@ -255,21 +302,8 @@ static int star_compare(const struct avltree_node *aa, const struct avltree_node
     return 0;
 }
 
-static int file_compare(const struct avltree_node *aa, const struct avltree_node *bb)
+static void file_free(struct file_s *a)
 {
-    const struct file_s *a = cavltree_container_of(aa, struct file_s, node);
-    const struct file_s *b = cavltree_container_of(bb, struct file_s, node);
-    int c = strcmp(a->name, b->name);
-    if (c != 0) return c;
-    c = str_cmp(&a->base, &b->base);
-    if (c != 0) return c;
-    return a->type - b->type;
-}
-
-static void file_free(struct avltree_node *aa)
-{
-    struct file_s *a = avltree_container_of(aa, struct file_s, node);
-
     free(a->data);
     free(a->line);
     free(a->nomacro);
@@ -326,12 +360,31 @@ static inline uchar_t fromiso(uchar_t c) {
     return conv[c];
 }
 
+static size_t fsize(FILE *f) {
+#if defined _POSIX_C_SOURCE || defined __unix__ || defined __MINGW32__
+    struct stat st;
+    if (fstat(fileno(f), &st) == 0) {
+        if (S_ISREG(st.st_mode) && st.st_size > 0) {
+            return (size_t)st.st_size;
+        }
+    }
+#else
+    if (fseek(f, 0, SEEK_END) == 0) {
+        long len = ftell(f);
+        rewind(f);
+        if (len > 0) {
+            return (size_t)len;
+        }
+    }
+#endif
+    return 0;
+}
+
 static struct file_s *command_line = NULL;
 static struct file_s *lastfi = NULL;
 static struct ubuff_s last_ubuff;
 static uint16_t curfnum;
 struct file_s *openfile(const char *name, const char *base, int ftype, const str_t *val, linepos_t epoint) {
-    struct avltree_node *b;
     struct file_s *tmp;
     char *s;
     if (lastfi == NULL) {
@@ -341,15 +394,17 @@ struct file_s *openfile(const char *name, const char *base, int ftype, const str
     lastfi->base.len = get_base(base);
     lastfi->type = ftype;
     if (name != NULL) {
+        str_t n;
         lastfi->name = name;
-        b = avltree_insert(&lastfi->node, &file_tree, file_compare);
+        n.data = (const uint8_t *)name;
+        n.len = strlen(name);
+        lastfi->hash = (str_hash(&n) + str_hash(&lastfi->base) + lastfi->type) & ((~0U) >> 1);
+        tmp = file_table_update(lastfi);
     } else {
-        b = (command_line != NULL) ? &command_line->node : NULL;
+        tmp = (command_line != NULL) ? command_line : NULL;
         if (command_line == NULL) command_line = lastfi;
     }
-    if (b != NULL) {
-        tmp = avltree_container_of(b, struct file_s, node);
-    } else { /* new file */
+    if (tmp == NULL) { /* new file */
         Encoding_types encoding = E_UNKNOWN;
         FILE *f;
         uchar_t c = 0;
@@ -365,6 +420,7 @@ struct file_s *openfile(const char *name, const char *base, int ftype, const str
         lastfi->read_error = false;
         lastfi->portable = false;
         lastfi->pass = 0;
+        lastfi->entercount = 0;
         tmp = lastfi;
         lastfi = NULL;
         if (name != NULL) {
@@ -388,23 +444,20 @@ struct file_s *openfile(const char *name, const char *base, int ftype, const str
             if (path == NULL) path = s;
             tmp->realname = path;
             if (arguments.quiet) {
-                printf((ftype == 1) ? "Reading file:      " : "Assembling file:   ");
+                fputs((ftype == 1) ? "Reading file:      " : "Assembling file:   ", stdout);
                 argv_print(path, stdout);
                 putchar('\n');
             }
             if (f == NULL) goto openerr;
             tmp->read_error = true;
             if (ftype == 1) {
-                bool check = true;
-                if (fseek(f, 0, SEEK_END) == 0) {
-                    long len = ftell(f);
-                    if (len > 0) {
-                        size_t len2 = (size_t)len;
-                        tmp->data = (uint8_t *)malloc(len2);
-                        if (tmp->data != NULL) tmp->len = len2;
-                    }
-                    rewind(f);
+                bool check;
+                size_t fs = fsize(f);
+                if (fs > 0) {
+                    tmp->data = (uint8_t *)malloc(fs);
+                    if (tmp->data != NULL) tmp->len = fs;
                 }
+                check = (tmp->data != NULL);
                 clearerr(f); errno = 0;
                 if (tmp->len != 0 || !extendfile(tmp)) {
                     for (;;) {
@@ -433,19 +486,16 @@ struct file_s *openfile(const char *name, const char *base, int ftype, const str
                 line_t lines = 0;
                 uint8_t buffer[BUFSIZ * 2];
                 size_t bp = 0, bl, qr = 1;
-                if (fseek(f, 0, SEEK_END) == 0) {
-                    long len = ftell(f);
-                    if (len > 0) {
-                        size_t len2 = (size_t)len + 4096;
-                        if (len2 < 4096) len2 = SIZE_MAX; /* overflow */
-                        tmp->data = (uint8_t *)malloc(len2);
-                        if (tmp->data != NULL) tmp->len = len2;
-                        max_lines = (len2 / 20 + 1024) & ~(size_t)1023;
-                        if (max_lines > SIZE_MAX / sizeof *tmp->line) max_lines = SIZE_MAX / sizeof *tmp->line; /* overflow */
-                        tmp->line = (size_t *)malloc(max_lines * sizeof *tmp->line);
-                        if (tmp->line == NULL) max_lines = 0;
-                    }
-                    rewind(f);
+                size_t fs = fsize(f);
+                if (fs > 0) {
+                    size_t len2 = fs + 4096;
+                    if (len2 < 4096) len2 = SIZE_MAX; /* overflow */
+                    tmp->data = (uint8_t *)malloc(len2);
+                    if (tmp->data != NULL) tmp->len = len2;
+                    max_lines = (len2 / 20 + 1024) & ~(size_t)1023;
+                    if (max_lines > SIZE_MAX / sizeof *tmp->line) max_lines = SIZE_MAX / sizeof *tmp->line; /* overflow */
+                    tmp->line = (size_t *)malloc(max_lines * sizeof *tmp->line);
+                    if (tmp->line == NULL) max_lines = 0;
                 }
                 clearerr(f); errno = 0;
                 bl = fread(buffer, 1, BUFSIZ, f);
@@ -758,10 +808,17 @@ struct star_s *new_star(line_t line, bool *exists) {
 void destroy_file(void) {
     struct stars_s *old;
 
-    avltree_destroy(&file_tree, file_free);
+    if (file_table.data != NULL) {
+        size_t i;
+        for (i = 0; i <= file_table.mask; i++) {
+            struct file_s *p = file_table.data[i];
+            if (p != NULL) file_free(p);
+        }
+        free(file_table.data);
+    }
     free(lastfi);
     free(last_ubuff.data);
-    if (command_line != NULL) file_free(&command_line->node);
+    if (command_line != NULL) file_free(command_line);
 
     include_list_last = include_list.next;
     while (include_list_last != NULL) {
@@ -779,7 +836,9 @@ void destroy_file(void) {
 
 void init_file(void) {
     curfnum = 1;
-    avltree_init(&file_tree);
+    file_table.len = 0;
+    file_table.mask = 0;
+    file_table.data = NULL;
     stars = (struct stars_s *)mallocx(sizeof *stars);
     stars->next = NULL;
     starsp = 0;
@@ -788,11 +847,22 @@ void init_file(void) {
     last_ubuff.len = 16;
 }
 
-void makefile(int argc, char *argv[]) {
+static size_t wrap_print(const char *txt, FILE *f, size_t len) {
+    if (len != 0) {
+        if (len > 64) {
+            fputs(" \\\n", f);
+            len = 0;
+        }
+        len++;
+        putc(' ', f);
+    }
+    return len + makefile_print(txt, f);
+}
+
+void makefile(int argc, char *argv[], bool make_phony) {
     FILE *f;
     struct linepos_s nopoint = {0, 0};
-    struct avltree_node *n;
-    size_t len = 0;
+    size_t len = 0, j;
     int i, err;
 
     f = dash_name(arguments.make) ? stdout : file_open(arguments.make, "wt");
@@ -801,35 +871,50 @@ void makefile(int argc, char *argv[]) {
         return;
     }
     clearerr(f); errno = 0;
-    if (!dash_name(arguments.output.name)) {
-        len += argv_print(arguments.output.name + get_base(arguments.output.name), f);
+    for (j = 0; j < arguments.output_len; j++) {
+        const struct output_s *output = &arguments.output[j];
+        if (dash_name(output->name)) continue;
+        len = wrap_print(output->name, f, len);
     }
-    if (len > 0) {
+    if (arguments.list != NULL) {
+        if (!dash_name(arguments.list)) {
+            len = wrap_print(arguments.list, f, len);
+        }
+    }
+    for (j = 0; j < arguments.symbol_output_len; j++) {
+        const struct symbol_output_s *output = &arguments.symbol_output[j];
+        if (dash_name(output->name)) continue;
+        len = wrap_print(output->name, f, len);
+    }
+    if (len != 0) {
         len++;
         putc(':', f);
 
         for (i = 0; i < argc; i++) {
             if (dash_name(argv[i])) continue;
-            if (len > 64) {
-                fputs(" \\\n", f);
-                len = 0;
-            }
-            putc(' ', f);
-            len += argv_print(argv[i], f) + 1;
+            len = wrap_print(argv[i], f, len);
         }
 
-        for (n = avltree_first(&file_tree); n != NULL; n = avltree_next(n)) {
-            const struct file_s *a = cavltree_container_of(n, struct file_s, node);
-            if (a->type != 0) {
-                if (len > 64) {
-                    fputs(" \\\n", f);
-                    len = 0;
-                }
-                putc(' ', f);
-                len += argv_print(a->realname, f) + 1;
+        if (file_table.data != NULL) {
+            for (j = 0; j <= file_table.mask; j++) {
+                const struct file_s *a = file_table.data[j];
+                if (a == NULL) continue;
+                if (a->type == 0) continue;
+                len = wrap_print(a->realname, f, len);
             }
         }
         putc('\n', f);
+
+        if (file_table.data != NULL && make_phony) {
+            len = 0;
+            for (j = 0; j <= file_table.mask; j++) {
+                const struct file_s *a = file_table.data[j];
+                if (a == NULL) continue;
+                if (a->type == 0) continue;
+                len = wrap_print(a->realname, f, len);
+            }
+            if (len != 0) fputs(":\n", f);
+        }
     }
 
     err = ferror(f);
