@@ -1,5 +1,5 @@
 /*
-    $Id: foldobj.h 1600 2018-08-25 16:23:27Z soci $
+    $Id: foldobj.h 1925 2019-08-25 11:09:57Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
 #define FOLDOBJ_H
 #include "obj.h"
 
-extern struct Type *FOLD_OBJ;
+extern struct Type *const FOLD_OBJ;
 
 typedef struct Fold {
     Obj v;
