@@ -1,5 +1,5 @@
 /*
-    $Id: functionobj.c 1850 2019-01-29 13:42:17Z soci $
+    $Id: functionobj.c 2173 2020-03-22 23:37:40Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -36,6 +36,7 @@
 #include "noneobj.h"
 #include "errorobj.h"
 #include "bytesobj.h"
+#include "dictobj.h"
 
 static Type obj;
 
@@ -58,7 +59,9 @@ static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
 
 static MUST_CHECK Error *hash(Obj *o1, int *hs, linepos_t UNUSED(epoint)) {
     Function *v1 = (Function *)o1;
-    *hs = v1->name_hash;
+    int h = v1->name_hash;
+    if (h < 0) v1->name_hash = h = str_hash(&v1->name);
+    *hs = h;
     return NULL;
 }
 
@@ -101,52 +104,88 @@ static MUST_CHECK Obj *gen_broadcast(Funcargs *vals, linepos_t epoint, func_t f)
     struct values_s *v = vals->val;
     size_t args = vals->len;
     size_t j;
+    struct elements_s {
+        Obj *oval;
+        struct iter_s iters;
+    };
+    struct elements_s elements3[3], *elements;
     for (j = 0; j < args; j++) {
-        const Type *objt = v[j].val->obj;
-        if (objt == TUPLE_OBJ || objt == LIST_OBJ) {
-            List *v1 = (List *)v[j].val, *vv;
-            Obj *oval[3];
-            size_t k;
+        const Type *objt2, *objt = v[j].val->obj;
+        if (objt->iterable) {
+            struct iter_s iter;
+            size_t ln, k;
+            if (args <= lenof(elements3)) {
+                elements = elements3; 
+            } else {
+                if (args > SIZE_MAX / sizeof *elements) goto failed; /* overflow */
+                elements = (struct elements_s *)malloc(args * sizeof *elements);
+                if (elements == NULL) goto failed;
+            }
+            elements[j].oval = v[j].val;
+            iter.data = v[j].val; objt->getiter(&iter);
+            ln = iter.len;
+            if (ln == 1) {
+                v[j].val = iter.next(&iter);
+                iter_destroy(&iter);
+                elements[j].iters.data = NULL;
+            } else elements[j].iters = iter;
             for (k = j + 1; k < args; k++) {
-                oval[k] = v[k].val;
-                if (oval[k]->obj == LIST_OBJ || oval[k]->obj == TUPLE_OBJ) {
-                    List *v2 = (List *)v[k].val;
-                    if (v2->len != 1 && v2->len != v1->len) {
-                        Error *err = new_error(ERROR_CANT_BROADCAS, &v[j].epoint);
-                        err->u.broadcast.v1 = v1->len;
-                        err->u.broadcast.v2 = v2->len;
+                elements[k].oval = v[k].val;
+                objt2 = elements[k].oval->obj;
+                if (objt2->iterable) {
+                    Error *err;
+                    struct iter_s iter2;
+                    iter2.data = elements[k].oval; objt2->getiter(&iter2);
+                    if (iter2.len != 1) {
+                        elements[k].iters = iter2;
+                        if (iter2.len == ln) continue;
+                        if (ln == 1) {
+                            ln = iter2.len;
+                            continue;
+                        }
+                        for (; k > j; k--) {
+                            if (elements[k].iters.data != NULL) iter_destroy(&elements[k].iters);
+                            v[k].val = elements[k].oval;
+                        }
+                        if (elements[j].iters.data != NULL) iter_destroy(&elements[j].iters);
+                        v[j].val = elements[j].oval;
+                        if (elements != elements3) free(elements);
+                        err = new_error(ERROR_CANT_BROADCAS, &v[j].epoint);
+                        err->u.broadcast.v1 = ln;
+                        err->u.broadcast.v2 = iter2.len;
                         return &err->v;
                     }
-                    if (diagnostics.type_mixing && v[k].val->obj != objt) err_msg_type_mixing(&v[j].epoint);
+                    v[k].val = iter2.next(&iter2);
+                    iter_destroy(&iter2);
                 }
+                elements[k].iters.data = NULL;
             }
-            if (v1->len != 0) {
+            if (ln != 0) {
                 size_t i;
-                Obj **vals2, *o1 = v[j].val;
-                vv = (List *)val_alloc(objt);
-                vals2 = list_create_elements(vv, v1->len);
-                for (i = 0; i < v1->len; i++) {
-                    v[j].val = v1->data[i];
-                    for (k = j + 1; k < args; k++) {
-                        if (oval[k]->obj == LIST_OBJ || oval[k]->obj == TUPLE_OBJ) {
-                            List *v2 = (List *)oval[k];
-                            v[k].val = v2->data[(v2->len == 1) ? 0 : i];
-                        }
+                List *vv = (List *)val_alloc(objt == TUPLE_OBJ ? TUPLE_OBJ : LIST_OBJ);
+                Obj **vals2 = vv->data = list_create_elements(vv, ln);
+                for (i = 0; i < ln; i++) {
+                    for (k = j; k < args; k++) {
+                        if (elements[k].iters.data != NULL) v[k].val = elements[k].iters.next(&elements[k].iters);
                     }
                     vals2[i] = gen_broadcast(vals, epoint, f);
                 }
                 vv->len = i;
-                vv->data = vals2;
-                v[j].val = o1;
-                for (i = j + 1; i < args; i++) {
-                    v[i].val = oval[i];
+                for (k = j; k < args; k++) {
+                    if (elements[k].iters.data != NULL) iter_destroy(&elements[k].iters);
+                    v[k].val = elements[k].oval;
                 }
+                if (elements != elements3) free(elements);
                 return &vv->v;
             }
-            return val_reference(&v1->v);
+            if (elements != elements3) free(elements);
+            iter_destroy(&iter);
+            return val_reference(objt == TUPLE_OBJ ? &null_tuple->v : &null_list->v);
         }
     }
     return f(vals, epoint);
+failed:
+    return (Obj *)new_error_mem(epoint);
 }
 
 /* range([start],end,[step]) */
@@ -287,14 +326,13 @@ static MUST_CHECK Obj *function_random(Funcargs *vals, linepos_t epoint) {
 
 static struct oper_s sort_tmp;
 static Obj *sort_error;
-static Obj **sort_vals;
+static Obj *sort_val;
 
-static int sortcomp(const void *a, const void *b) {
+static int sortcomp(void) {
     int ret;
     Obj *result;
-    size_t aa = *(const size_t *)a, bb = *(const size_t *)b;
-    Obj *o1 = sort_tmp.v1 = sort_vals[aa];
-    Obj *o2 = sort_tmp.v2 = sort_vals[bb];
+    Obj *o1 = sort_tmp.v1;
+    Obj *o2 = sort_tmp.v2;
     sort_tmp.inplace = NULL;
     result = sort_tmp.v1->obj->calc2(&sort_tmp);
     if (result->obj == INT_OBJ) ret = (int)((Int *)result)->len;
@@ -318,7 +356,7 @@ static int sortcomp(const void *a, const void *b) {
                     }
                     if (i == v1->len) {
                         val_destroy(result);
-                        return (aa > bb) ? 1 : -1;
+                        return 0;
                     }
                 }
                 sort_tmp.v1 = o1;
@@ -328,16 +366,36 @@ static int sortcomp(const void *a, const void *b) {
         }
     }
     val_destroy(result);
+    return ret;
+}
+
+static int list_sortcomp(const void *a, const void *b) {
+    int ret;
+    size_t aa = *(const size_t *)a, bb = *(const size_t *)b;
+    List *list = (List *)sort_val;
+    sort_tmp.v1 = list->data[aa];
+    sort_tmp.v2 = list->data[bb];
+    ret = sortcomp();
+    if (ret == 0) return (aa > bb) ? 1 : -1;
+    return ret;
+}
+
+static int dict_sortcomp(const void *a, const void *b) {
+    int ret;
+    size_t aa = *(const size_t *)a, bb = *(const size_t *)b;
+    Dict *dict = (Dict *)sort_val;
+    sort_tmp.v1 = dict->data[aa].key;
+    sort_tmp.v2 = dict->data[bb].key;
+    ret = sortcomp();
     if (ret == 0) return (aa > bb) ? 1 : -1;
     return ret;
 }
 
 /* sort() */
 static MUST_CHECK Obj *function_sort(Obj *o1, linepos_t epoint) {
-    if (o1->obj == TUPLE_OBJ || o1->obj == LIST_OBJ) {
-        List *v1 = (List *)o1, *v;
-        size_t ln = v1->len;
-        if (ln > 0) {
+    if (o1->obj == TUPLE_OBJ || o1->obj == LIST_OBJ || o1->obj == DICT_OBJ) {
+        size_t ln = (o1->obj == DICT_OBJ) ? ((Dict *)o1)->len : ((List *)o1)->len;
+        if (ln > 1) {
             size_t i;
             Obj **vals;
             size_t *sort_index;
@@ -345,21 +403,25 @@ static MUST_CHECK Obj *function_sort(Obj *o1, linepos_t epoint) {
             sort_index = (size_t *)malloc(ln * sizeof *sort_index);
             if (sort_index == NULL) goto failed;
             for (i = 0; i < ln; i++) sort_index[i] = i;
-            sort_vals = v1->data;
+            sort_val = o1;
             sort_error = NULL;
             sort_tmp.op = &o_CMP;
             sort_tmp.epoint = sort_tmp.epoint2 = sort_tmp.epoint3 = epoint;
-            qsort(sort_index, ln, sizeof *sort_index, sortcomp);
+            qsort(sort_index, ln, sizeof *sort_index, (o1->obj == DICT_OBJ) ? dict_sortcomp : list_sortcomp);
             if (sort_error != NULL) {
                 free(sort_index);
                 return sort_error;
             }
-            v = (List *)val_alloc(o1->obj);
-            v->data = vals = list_create_elements(v, ln);
-            v->len = ln;
-            for (i = 0; i < ln; i++) vals[i] = val_reference(v1->data[sort_index[i]]);
+            if (o1->obj == DICT_OBJ) o1 = dict_sort((Dict *)o1, sort_index);
+            else {
+                List *v = (List *)val_alloc(o1->obj);
+                v->data = vals = list_create_elements(v, ln);
+                v->len = ln;
+                for (i = 0; i < ln; i++) vals[i] = val_reference(((List *)o1)->data[sort_index[i]]);
+                o1 = &v->v;
+            }
             free(sort_index);
-            return &v->v;
+            return o1;
         }
     }
     return val_reference(o1);
@@ -373,11 +435,13 @@ static MUST_CHECK Obj *function_binary(Funcargs *vals, linepos_t epoint) {
     Error *err;
     ival_t offs = 0;
     uval_t length = (uval_t)-1;
-    char *path = NULL;
+    struct file_s *cfile2 = NULL;
     str_t filename;
 
     if (!tostr(&v[0], &filename)) {
-        path = get_path(&filename, current_file_list->file->realname);
+        char *path = get_path(&filename, current_file_list->file->realname);
+        cfile2 = openfile(path, current_file_list->file->realname, 1, &filename, &v[0].epoint);
+        free(path);
     }
 
     switch (vals->len) {
@@ -392,94 +456,103 @@ static MUST_CHECK Obj *function_binary(Funcargs *vals, linepos_t epoint) {
     default:
         break;
     }
-    if (path != NULL) {
-        struct file_s *cfile2 = openfile(path, current_file_list->file->realname, 1, &filename, epoint);
-        free(path);
-        if (cfile2 != NULL) {
-            size_t offset, ln = cfile2->len;
-            Bytes *b;
-            if (offs < 0) offset = ((uval_t)-offs < ln) ? (ln - (uval_t)-offs) : 0;
-            else offset = (uval_t)offs;
-            if (offset < ln) ln -= offset; else ln = 0;
-            if (length < ln) ln = length;
-            if (ln == 0) return (Obj *)ref_bytes(null_bytes);
-            if (ln > SSIZE_MAX) return (Obj *)new_error_mem(epoint);
-            b = new_bytes(ln);
-            b->len = ln;
-            memcpy(b->data, cfile2->data + offset, ln);
-            return &b->v;
-        }
+    
+    if (cfile2 != NULL) {
+        size_t offset, ln = cfile2->len;
+        Bytes *b;
+        if (offs < 0) offset = ((uval_t)-offs < ln) ? (ln - (uval_t)-offs) : 0;
+        else offset = (uval_t)offs;
+        if (offset < ln) ln -= offset; else ln = 0;
+        if (length < ln) ln = length;
+        if (ln == 0) return (Obj *)ref_bytes(null_bytes);
+        if (ln > SSIZE_MAX) return (Obj *)new_error_mem(epoint);
+        b = new_bytes(ln);
+        b->len = ln;
+        memcpy(b->data, cfile2->data + offset, ln);
+        return &b->v;
     }
     return (Obj *)ref_none();
 }
 
-static MUST_CHECK Obj *apply_func(Obj *o1, Function_types func, linepos_t epoint) {
-    const Type *typ = o1->obj;
-    Obj *err;
+static MUST_CHECK Obj *apply_func(oper_t op) {
+    Obj *o2 = op->v2;
+    const Type *typ = o2->obj;
+    bool inplace = op->inplace == o2;
     double real;
+    Error_types err;
 
-    if (typ == TUPLE_OBJ || typ == LIST_OBJ) {
-        iter_next_t iter_next;
-        Iter *iter = typ->getiter(o1);
-        size_t len = iter->len(iter);
+    if (typ->iterable) {
         List *v;
-        size_t i;
+        size_t i, len;
         Obj **vals;
 
-        if (len == 0) {
-            val_destroy(&iter->v);
-            return val_reference(typ == TUPLE_OBJ ? &null_tuple->v : &null_list->v);
+        if (!inplace || (typ != TUPLE_OBJ && typ != LIST_OBJ)) {
+            struct iter_s iter;
+            iter.data = o2; typ->getiter(&iter);
+            if (iter.len == 0) {
+                iter_destroy(&iter);
+                return val_reference(typ == TUPLE_OBJ ? &null_tuple->v : &null_list->v);
+            }
+            v = (List *)val_alloc(typ == TUPLE_OBJ ? TUPLE_OBJ : LIST_OBJ);
+            v->data = vals = list_create_elements(v, iter.len);
+            for (i = 0; i < iter.len && (o2 = iter.next(&iter)) != NULL; i++) {
+                op->v2 = o2;
+                op->inplace = inplace && o2->refcount == 1 ? o2 : NULL;
+                vals[i] = apply_func(op);
+            }
+            iter_destroy(&iter);
+            v->len = i;
+            return &v->v;
+        } 
+        v = (List *)val_reference(o2);
+        len = v->len;
+        vals = v->data;
+        for (i = 0; i < len; i++) {
+            op->v2 = o2 = vals[i];
+            op->inplace = o2->refcount == 1 ? o2 : NULL;
+            vals[i] = apply_func(op);
+            val_destroy(o2);
         }
-
-        v = (List *)val_alloc(typ == TUPLE_OBJ ? TUPLE_OBJ : LIST_OBJ);
-        v->data = vals = list_create_elements(v, len);
-        iter_next = iter->next;
-        for (i = 0; i < len && (o1 = iter_next(iter)) != NULL; i++) {
-            vals[i] = apply_func(o1, func, epoint);
-        }
-        val_destroy(&iter->v);
-        v->len = i;
         return &v->v;
     }
-    switch (func) {
-    case F_SIZE: return typ->size(o1, epoint);
-    case F_SIGN: return typ->sign(o1, epoint);
-    case F_CEIL: return typ->function(o1, TF_CEIL, epoint);
-    case F_FLOOR: return typ->function(o1, TF_FLOOR, epoint);
-    case F_ROUND: return typ->function(o1, TF_ROUND, epoint);
-    case F_TRUNC: return typ->function(o1, TF_TRUNC, epoint);
-    case F_ABS: return typ->function(o1, TF_ABS, epoint);
+    if (op->v1->obj != FUNCTION_OBJ) return ((Type *)op->v1)->create(op->v2, op->epoint2);
+    switch (((Function *)op->v1)->func) {
+    case F_SIZE: return typ->size(op);
+    case F_SIGN: return typ->sign(o2, op->epoint2);
+    case F_CEIL:
+    case F_FLOOR:
+    case F_ROUND:
+    case F_TRUNC:
+    case F_ABS: return typ->function(op);
     case F_REPR:
         {
-            Obj *v = typ->repr(o1, epoint, SIZE_MAX);
-            return v != NULL ? v : (Obj *)new_error_mem(epoint);
+            Obj *v = typ->repr(o2, op->epoint2, SIZE_MAX);
+            return v != NULL ? v : (Obj *)new_error_mem(op->epoint2);
         }
     default: break;
     }
-    if (typ == FLOAT_OBJ) {
-        real = ((Float *)o1)->real;
-    } else {
-        err = FLOAT_OBJ->create(o1, epoint);
-        if (err->obj != FLOAT_OBJ) return err;
-        real = ((Float *)err)->real;
-        val_destroy(err);
+    if (typ != FLOAT_OBJ) {
+        o2 = FLOAT_OBJ->create(o2, op->epoint2);
+        if (o2->obj != FLOAT_OBJ) return o2;
+        inplace = o2->refcount == 1;
     }
-    switch (func) {
+    real = ((Float *)o2)->real;
+    switch (((Function *)op->v1)->func) {
     case F_SQRT:
         if (real < 0.0) {
-            return (Obj *)new_error_obj(ERROR_SQUARE_ROOT_N, o1, epoint);
+            err = ERROR_SQUARE_ROOT_N; goto failed;
         }
         real = sqrt(real);
         break;
     case F_LOG10:
         if (real <= 0.0) {
-            return (Obj *)new_error_obj(ERROR_LOG_NON_POSIT, o1, epoint);
+            err = ERROR_LOG_NON_POSIT; goto failed;
         }
         real = log10(real);
         break;
     case F_LOG:
         if (real <= 0.0) {
-            return (Obj *)new_error_obj(ERROR_LOG_NON_POSIT, o1, epoint);
+            err = ERROR_LOG_NON_POSIT; goto failed;
         }
         real = log(real);
         break;
@@ -489,13 +562,13 @@ static MUST_CHECK Obj *apply_func(Obj *o1, Function_types func, linepos_t epoint
     case F_TAN: real = tan(real);break;
     case F_ACOS:
         if (real < -1.0 || real > 1.0) {
-            return (Obj *)new_error_obj(ERROR___MATH_DOMAIN, o1, epoint);
+            err = ERROR___MATH_DOMAIN; goto failed;
         }
         real = acos(real);
         break;
     case F_ASIN:
         if (real < -1.0 || real > 1.0) {
-            return (Obj *)new_error_obj(ERROR___MATH_DOMAIN, o1, epoint);
+            err = ERROR___MATH_DOMAIN; goto failed;
         }
         real = asin(real);
         break;
@@ -509,7 +582,15 @@ static MUST_CHECK Obj *apply_func(Obj *o1, Function_types func, linepos_t epoint
     case F_TANH: real = tanh(real);break;
     default: real = HUGE_VAL; break; /* can't happen */
     }
-    return float_from_double(real, epoint);
+    if (!inplace || real == HUGE_VAL || real == -HUGE_VAL) {
+        if (typ != FLOAT_OBJ) val_destroy(o2);
+        return float_from_double(real, op->epoint2);
+    }
+    ((Float *)o2)->real = real;
+    return typ != FLOAT_OBJ ? o2 : val_reference(o2);
+failed:
+    if (typ != FLOAT_OBJ) val_destroy(o2);
+    return (Obj *)new_error_obj(err, op->v2, op->epoint2);
 }
 
 static MUST_CHECK Obj *to_real(struct values_s *v, double *r) {
@@ -611,74 +692,79 @@ static MUST_CHECK Obj *calc2(oper_t op) {
                 switch (func) {
                 case F_HYPOT:
                     if (args != 2) {
-                        err_msg_argnum(args, 2, 2, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 2, 2, op->epoint2);
                     }
                     return gen_broadcast(v2, op->epoint, function_hypot);
                 case F_ATAN2:
                     if (args != 2) {
-                        err_msg_argnum(args, 2, 2, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 2, 2, op->epoint2);
                     }
                     return gen_broadcast(v2, op->epoint, function_atan2);
                 case F_POW:
                     if (args != 2) {
-                        err_msg_argnum(args, 2, 2, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 2, 2, op->epoint2);
                     }
                     return gen_broadcast(v2, op->epoint, function_pow);
                 case F_RANGE:
                     if (args < 1 || args > 3) {
-                        err_msg_argnum(args, 1, 3, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 1, 3, op->epoint2);
                     }
                     return gen_broadcast(v2, op->epoint, function_range);
                 case F_BINARY:
                     if (args < 1 || args > 3) {
-                        err_msg_argnum(args, 1, 3, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 1, 3, op->epoint2);
                     }
                     return gen_broadcast(v2, op->epoint, function_binary);
                 case F_FORMAT:
-                    return isnprintf(v2, op->epoint);
+                    if (args < 1) {
+                        return (Obj *)new_error_argnum(args, 1, 0, op->epoint2);
+                    }
+                    return gen_broadcast(v2, op->epoint, isnprintf);
                 case F_RANDOM:
                     if (args > 3) {
-                        err_msg_argnum(args, 0, 3, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 0, 3, op->epoint2);
                     }
                     return gen_broadcast(v2, op->epoint, function_random);
                 default:
                     if (args != 1) {
-                        err_msg_argnum(args, 1, 1, op->epoint2);
-                        return (Obj *)ref_none();
+                        return (Obj *)new_error_argnum(args, 1, 1, op->epoint2);
                     }
                     switch (func) {
                     case F_ANY: return v[0].val->obj->truth(v[0].val, TRUTH_ANY, &v[0].epoint);
                     case F_ALL: return v[0].val->obj->truth(v[0].val, TRUTH_ALL, &v[0].epoint);
-                    case F_LEN: return v[0].val->obj->len(v[0].val, &v[0].epoint);
+                    case F_LEN: 
+                        op->v2 = v[0].val;
+                        return v[0].val->obj->len(op);
                     case F_SORT: return function_sort(v[0].val, &v[0].epoint);
-                    default: return apply_func(v[0].val, func, &v[0].epoint);
+                    default: 
+                        op->v2 = v[0].val;
+                        op->inplace = v[0].val->refcount == 1 ? v[0].val : NULL;
+                        return apply_func(op);
                     }
                 }
             default: break;
             }
             break;
         }
-    case T_LIST:
-    case T_TUPLE:
-    case T_DICT:
-        if (op->op != &o_MEMBER && op->op != &o_X) {
-            return o2->obj->rcalc2(op);
-        }
-        break;
     case T_NONE:
     case T_ERROR:
         return val_reference(o2);
-    default: break;
+    default:
+        if (o2->obj->iterable && op->op != &o_MEMBER && op->op != &o_X) {
+            return o2->obj->rcalc2(op);
+        }
+        break;
     }
     return obj_oper_error(op);
 }
 
+MUST_CHECK Obj *apply_convert(oper_t op) {
+    struct values_s *v = ((Funcargs *)op->v2)->val;
+    op->v2 = v[0].val;
+    op->inplace = v[0].val->refcount == 1 ? v[0].val : NULL;
+    return apply_func(op);
+}
+
 void functionobj_init(void) {
     new_type(&obj, T_FUNCTION, "function", sizeof(Function));
     obj.create = create;
@@ -742,7 +828,7 @@ void functionobj_names(void) {
         func->name.data = (const uint8_t *)builtin_functions[i].name;
         func->name.len = strlen(builtin_functions[i].name);
         func->func = builtin_functions[i].func;
-        func->name_hash = str_hash(&func->name);
+        func->name_hash = -1;
         new_builtin(builtin_functions[i].name, &func->v);
     }
 }
