@@ -1,5 +1,5 @@
 /*
-    $Id: labelobj.h 1710 2018-12-15 23:08:17Z soci $
+    $Id: labelobj.h 2134 2020-01-01 19:10:10Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,9 +28,10 @@ typedef struct Label {
     Obj v;
     str_t name;
     str_t cfname;
+    int hash;
 
     Obj *value;
-    struct file_list_s *file_list;
+    const struct file_list_s *file_list;
     struct linepos_s epoint;
     bool ref;
     bool update_after;
@@ -39,6 +40,7 @@ typedef struct Label {
     uint8_t usepass;
     uint8_t defpass;
     uint8_t strength;
+    uint8_t fwpass;
 } Label;
 
 extern void labelobj_init(void);
