@@ -1,5 +1,5 @@
 /*
-    $Id: listing.c 1899 2019-02-17 21:42:52Z soci $
+    $Id: listing.c 2098 2019-11-24 05:07:52Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -136,7 +136,6 @@ MUST_CHECK Listing *listing_open(const char *filename, int argc, char *argv[]) {
     Listing *ls;
     struct linepos_s nopoint = {0, 0};
     time_t t;
-    const char *prgname;
     int i;
     FILE *flist;
 
@@ -167,18 +166,9 @@ MUST_CHECK Listing *listing_open(const char *filename, int argc, char *argv[]) {
     ls->i = 0;
 
     fputs("\n; 64tass Turbo Assembler Macro V" VERSION " listing file\n;", flist);
-    prgname = *argv;
-    if (prgname != NULL) {
-        const char *newp = strrchr(prgname, '/');
-        if (newp != NULL) prgname = newp + 1;
-#if defined _WIN32 || defined __WIN32__ || defined __EMX__ || defined __MSDOS__ || defined __DOS__
-        newp = strrchr(prgname, '\\');
-        if (newp != NULL) prgname = newp + 1;
-#endif
-    }
     for (i = 0; i < argc; i++) {
         putc(' ', flist);
-        argv_print((i != 0) ? argv[i] : prgname, flist);
+        argv_print(argv[i], flist);
     }
     fputs("\n; ", flist);
     time(&t); fputs(ctime(&t), flist);
@@ -366,8 +356,7 @@ static void printmon(Listing *ls, unsigned int cod, int ln, uint32_t adr) {
         out_pb(ls, (address_t)((int8_t)(adr >> 8) + (int)current_address->l_address.address));
         return;
     case ADR_REL_L: if (ln > 0) out_pb(ls, adr + (((cod & 0x0F) == 3) ? -1u : 0) + current_address->l_address.address); else ls->i--; return;
-    case ADR_MOVE: out_byte(ls, adr >> 8); ls->s[ls->i++] = ','; out_byte(ls, adr); return;
-    case ADR_LEN: return;/* not an addressing mode */
+    case ADR_MOVE: out_byte(ls, adr >> 8); ls->s[ls->i++] = ','; ls->s[ls->i++] = '#'; out_byte(ls, adr); return;
     }
     while (*mode != 0) ls->s[ls->i++] = *mode++;
 }
@@ -380,7 +369,26 @@ static void printsource(Listing *ls, linecpos_t pos) {
     newline(ls);
 }
 
-void FAST_CALL listing_line(Listing *ls, linecpos_t pos) {
+FAST_CALL void listing_equal2(Listing *ls, Obj *val, linecpos_t pos) {
+    if (ls == NULL) return;
+    if (nolisting != 0 || !ls->source || temporary_label_branch != 0) return;
+    if (ls->linenum) {
+        printline(ls);
+        padding2(ls, ls->columns.addr);
+    }
+    putc('=', ls->flist);
+    ls->c += val_print(val, ls->flist) + 1;
+    if (ls->verbose) {
+        printllist(ls);
+        newline(ls);
+    } else {
+        printsource(ls, pos);
+    }
+    llist = NULL;
+}
+
+
+FAST_CALL void listing_line(Listing *ls, linecpos_t pos) {
     size_t i;
     if (nolisting != 0  || temporary_label_branch != 0 || llist == NULL) return;
     if (ls == NULL) {
@@ -418,7 +426,7 @@ void FAST_CALL listing_line(Listing *ls, linecpos_t pos) {
     llist = NULL;
 }
 
-void FAST_CALL listing_line_cut(Listing *ls, linecpos_t pos) {
+FAST_CALL void listing_line_cut(Listing *ls, linecpos_t pos) {
     size_t i;
     if (nolisting != 0 || temporary_label_branch != 0 || llist == NULL) return;
     if (ls == NULL) {
@@ -444,7 +452,7 @@ void FAST_CALL listing_line_cut(Listing *ls, linecpos_t pos) {
     llist = NULL;
 }
 
-void FAST_CALL listing_line_cut2(Listing *ls, linecpos_t pos) {
+FAST_CALL void listing_line_cut2(Listing *ls, linecpos_t pos) {
     if (ls == NULL || !ls->verbose || llist == NULL) return;
     if (nolisting == 0 && ls->source && temporary_label_branch == 0) {
         if (ls->linenum) printline(ls);
