@@ -1,5 +1,5 @@
 /*
-    $Id: listobj.c 1898 2019-02-17 21:17:18Z soci $
+    $Id: listobj.c 2137 2020-01-02 00:52:17Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -46,13 +46,17 @@ Tuple *null_tuple;
 List *null_list;
 Addrlist *null_addrlist;
 
+static FAST_CALL NO_INLINE void list_destroy(List *v1) {
+    free(v1->data);
+}
+
 static FAST_CALL void destroy(Obj *o1) {
     List *v1 = (List *)o1;
     size_t i;
     for (i = 0; i < v1->len; i++) {
         val_destroy(v1->data[i]);
     }
-    if (v1->u.val != v1->data) free(v1->data);
+    if (v1->u.val != v1->data) list_destroy(v1);
 }
 
 static FAST_CALL void garbage(Obj *o1, int j) {
@@ -90,7 +94,8 @@ static Obj **lnew(List *v, size_t len) {
         Obj **n = (Obj **)malloc(len * sizeof *v->data);
         if (n != NULL) {
             v->data = n;
-            v->u.max = len;
+            v->u.s.max = len;
+            v->u.s.hash = -1;
             return n;
         }
     }
@@ -108,13 +113,14 @@ static Obj **lextend(List *v, size_t len) {
     if (len > SIZE_MAX / sizeof *v->data) return NULL; /* overflow */
     if (v->u.val != v->data) {
         size_t len2;
-        if (len <= v->u.max) return v->data;
+        if (len <= v->u.s.max) return v->data;
         len2 = len + (len < 1024 ? len : 1024);
         if (len2 > len) len = len2;
         tmp = (Obj **)realloc(v->data, len * sizeof *v->data);
         if (tmp != NULL) {
             v->data = tmp;
-            v->u.max = len;
+            v->u.s.max = len;
+            v->u.s.hash = -1;
         }
         return tmp;
     }
@@ -122,7 +128,8 @@ static Obj **lextend(List *v, size_t len) {
     if (tmp != NULL) {
         memcpy(tmp, v->u.val, v->len * sizeof *v->data);
         v->data = tmp;
-        v->u.max = len;
+        v->u.s.max = len;
+        v->u.s.hash = -1;
     }
     return tmp;
 }
@@ -148,7 +155,7 @@ static MUST_CHECK Obj *list_create(Obj *o1, linepos_t epoint) {
     case T_ERROR:
     case T_LIST: return val_reference(o1);
     case T_TUPLE: return tuple_from_list((List *)o1, LIST_OBJ, epoint);
-    case T_CODE: return tuple_from_code((Code *)o1, LIST_OBJ, epoint);
+    case T_CODE: return tuple_from_code((Code *)o1, LIST_OBJ);
     default: break;
     }
     return (Obj *)new_error_conv(o1, LIST_OBJ, epoint);
@@ -160,7 +167,7 @@ static MUST_CHECK Obj *tuple_create(Obj *o1, linepos_t epoint) {
     case T_ERROR:
     case T_TUPLE: return val_reference(o1);
     case T_LIST: return tuple_from_list((List *)o1, TUPLE_OBJ, epoint);
-    case T_CODE: return tuple_from_code((Code *)o1, TUPLE_OBJ, epoint);
+    case T_CODE: return tuple_from_code((Code *)o1, TUPLE_OBJ);
     default: break;
     }
     return (Obj *)new_error_conv(o1, TUPLE_OBJ, epoint);
@@ -180,27 +187,65 @@ static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
 static MUST_CHECK Obj *truth(Obj *o1, Truth_types type, linepos_t epoint) {
     List *v1 = (List *)o1;
     size_t i;
-    Obj *val;
+    Obj *val, *val2;
     switch (type) {
     case TRUTH_ALL:
+        val2 = NULL;
         for (i = 0; i < v1->len; i++) {
             val = v1->data[i]->obj->truth(v1->data[i], type, epoint);
-            if ((Bool *)val != true_value) return val;
-            val_destroy(val);
+            if ((Bool *)val == true_value) {
+                val_destroy(val);
+                continue;
+            }
+            if (val2 != NULL) val_destroy(val2);
+            if ((Bool *)val == false_value) {
+                return val;
+            }
+            val2 = val;
         }
-        return (Obj *)ref_bool(true_value);
+        return (val2 != NULL) ? val2 : (Obj *)ref_bool(true_value);
     case TRUTH_ANY:
+        val2 = NULL;
         for (i = 0; i < v1->len; i++) {
             val = v1->data[i]->obj->truth(v1->data[i], type, epoint);
-            if ((Bool *)val != false_value) return val;
-            val_destroy(val);
+            if ((Bool *)val == false_value) {
+                val_destroy(val);
+                continue;
+            }
+            if (val2 != NULL) val_destroy(val2);
+            if ((Bool *)val == true_value) return val;
+            val2 = val;
         }
-        return (Obj *)ref_bool(false_value);
+        return (val2 != NULL) ? val2 : (Obj *)ref_bool(false_value);
     default:
         return DEFAULT_OBJ->truth(o1, type, epoint);
     }
 }
 
+static MUST_CHECK Error *hash(Obj *o1, int *hs, linepos_t epoint) {
+    List *v1 = (List *)o1;
+    size_t i, l = v1->len;
+    Obj **vals = v1->data;
+    unsigned int h;
+    if (vals != v1->u.val && v1->u.s.hash >= 0) {
+        *hs = v1->u.s.hash;
+        return NULL;
+    }
+    h = 0;
+    for (i = 0; i < l; i++) {
+        int h2;
+        Obj *o2 = vals[i];
+        Error *err = o2->obj->hash(o2, &h2, epoint);
+        if (err != NULL) return err;
+        h += h2;
+    }
+    h ^= i;
+    h &= ((~0U) >> 1);
+    if (vals != v1->u.val) v1->u.s.hash = h;
+    *hs = h;
+    return NULL;
+}
+
 static MUST_CHECK Obj *repr_listtuple(Obj *o1, linepos_t epoint, size_t maxsize) {
     Tuple *v1 = (List *)o1;
     bool tupleorlist = (o1->obj != ADDRLIST_OBJ && o1->obj != COLONLIST_OBJ);
@@ -266,38 +311,75 @@ static MUST_CHECK Obj *repr_listtuple(Obj *o1, linepos_t epoint, size_t maxsize)
     return &v->v;
 }
 
-static MUST_CHECK Obj *len(Obj *o1, linepos_t UNUSED(epoint)) {
-    List *v1 = (List *)o1;
+static MUST_CHECK Obj *len(oper_t op) {
+    List *v1 = (List *)op->v2;
     return (Obj *)int_from_size(v1->len);
 }
 
-static FAST_CALL MUST_CHECK Obj *next(Iter *v1) {
-    const List *vv1 = (List *)v1->data;
-    if (v1->val >= vv1->len) return NULL;
-    return vv1->data[v1->val++];
+static FAST_CALL MUST_CHECK Obj *next(struct iter_s *v1) {
+    if (v1->val >= v1->len) return NULL;
+    return ((List *)v1->data)->data[v1->val++];
 }
 
-static size_t iter_len(Iter *v1) {
-    return ((List *)v1->data)->len - v1->val;
-}
-
-static MUST_CHECK Iter *getiter(Obj *v1) {
-    Iter *v = (Iter *)val_alloc(ITER_OBJ);
+static void getiter(struct iter_s *v) {
     v->iter = NULL;
     v->val = 0;
-    v->data = val_reference(v1);
+    v->data = val_reference(v->data);
     v->next = next;
-    v->len = iter_len;
-    return v;
+    v->len = ((List *)v->data)->len;
 }
 
 Obj **list_create_elements(List *v, size_t n) {
     if (n <= lenof(v->u.val)) return v->u.val;
     if (n > SIZE_MAX / sizeof *v->data) err_msg_out_of_memory(); /* overflow */
-    v->u.max = n;
+    v->u.s.max = n;
+    v->u.s.hash = -1;
     return (Obj **)mallocx(n * sizeof *v->data);
 }
 
+MUST_CHECK bool list_extend(List *lst) {
+    Obj **vals;
+    size_t o = lst->len, n;
+    if (lst->data == lst->u.val) {
+        n = 16;
+        vals = (Obj **)malloc(n * sizeof *lst->data);
+        if (vals == NULL) return true;
+        memcpy(vals, lst->u.val, o * sizeof *lst->data);
+    } else {
+        if (o < 256) n = o * 2;
+        else {
+            n = o + 256;
+            if (/*n < 256 ||*/ n > SIZE_MAX / sizeof *lst->data) return true; /* overflow */
+        }
+        vals = (Obj **)realloc(lst->data, n * sizeof *lst->data);
+        if (vals == NULL) return true;
+    }
+    lst->data = vals;
+    lst->u.s.max = n;
+    lst->u.s.hash = -1;
+    return false;
+}
+
+void list_shrink(List *lst, size_t i) {
+    size_t j = i;
+    while (j < lst->len) val_destroy(lst->data[j++]);
+    lst->len = i;
+    if (lst->data != lst->u.val) {
+        if (lst->len <= lenof(lst->u.val)) {
+            memcpy(lst->u.val, lst->data, lst->len * sizeof *lst->data);
+            free(lst->data);
+            lst->data = lst->u.val;
+        } else {
+            Obj **v = (Obj **)realloc(lst->data, lst->len * sizeof *lst->data);
+            if (v != NULL) {
+                lst->data = v;
+                lst->u.s.max = lst->len;
+                lst->u.s.hash = -1;
+            }
+        }
+    }
+}
+
 MUST_CHECK Tuple *new_tuple(size_t n) {
      Tuple *v = (Tuple *)val_alloc(TUPLE_OBJ);
      v->len = n;
@@ -306,7 +388,8 @@ MUST_CHECK Tuple *new_tuple(size_t n) {
          return v;
      }
      if (n > SIZE_MAX / sizeof *v->data) err_msg_out_of_memory(); /* overflow */
-     v->u.max = n;
+     v->u.s.max = n;
+     v->u.s.hash = -1;
      v->data = (Obj **)mallocx(n * sizeof *v->data);
      return v;
 }
@@ -321,6 +404,7 @@ static MUST_CHECK Obj *calc1(oper_t op) {
         if (inplace) {
             v = ref_list(v1);
             vals = v1->data;
+            if (vals != v->u.val) v->u.s.hash = -1;
         } else {
             v = (List *)val_alloc(o1->obj);
             vals = lnew(v, v1->len);
@@ -379,10 +463,12 @@ static MUST_CHECK Obj *calc2_list(oper_t op) {
                         v = ref_list(v1);
                         vals = v1->data;
                         inplace = v1;
-                    } else if (op->inplace == &v2->v) {
+                        if (vals != v->u.val) v->u.s.hash = -1;
+                    } else if (o1->obj == o2->obj && op->inplace == &v2->v) {
                         v = ref_list(v2);
                         vals = v2->data;
                         inplace = v2;
+                        if (vals != v->u.val) v->u.s.hash = -1;
                     } else {
                         v = (List *)val_alloc(o1->obj);
                         vals = lnew(v, v1->len);
@@ -441,7 +527,7 @@ static MUST_CHECK Obj *calc2_list(oper_t op) {
                 i = v1->len;
                 v1->len = ln;
                 v = (List *)val_reference(o1);
-            } else if (op->inplace == &v2->v) {
+            } else if (o1->obj == o2->obj && op->inplace == &v2->v) {
                 vals = lextend(v2, ln);
                 if (vals == NULL) goto failed;
                 memmove(vals + v1->len, v2->data, v2->len * sizeof *v2->data);
@@ -500,11 +586,11 @@ static inline MUST_CHECK Obj *repeat(oper_t op) {
     return (Obj *)new_error_mem(op->epoint3);
 }
 
-static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
+static MUST_CHECK Obj *slice(oper_t op, size_t indx) {
     Obj **vals;
     Obj *o2 = op->v2;
     size_t offs2;
-    List *v, *v1 = (List *)o1;
+    List *v, *v1 = (List *)op->v1;
     Funcargs *args = (Funcargs *)o2;
     size_t i, ln;
     Error *err;
@@ -512,8 +598,7 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
     linepos_t epoint2;
 
     if (args->len < 1) {
-        err_msg_argnum(args->len, 1, 0, op->epoint2);
-        return (Obj *)ref_none();
+        return (Obj *)new_error_argnum(args->len, 1, 0, op->epoint2);
     }
 
     o2 = args->val[indx].val;
@@ -521,36 +606,34 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
 
     ln = v1->len;
 
-    if (o2->obj == LIST_OBJ) {
-        iter_next_t iter_next;
-        Iter *iter = o2->obj->getiter(o2);
-        size_t len = iter->len(iter);
+    if (o2->obj->iterable) {
+        struct iter_s iter;
+        iter.data = o2; o2->obj->getiter(&iter);
 
-        if (len == 0) {
-            val_destroy(&iter->v);
-            return val_reference((o1->obj == TUPLE_OBJ) ? &null_tuple->v : &null_list->v);
+        if (iter.len == 0) {
+            iter_destroy(&iter);
+            return val_reference((v1->v.obj == TUPLE_OBJ) ? &null_tuple->v : &null_list->v);
         }
-        v = (List *)val_alloc(o1->obj);
-        vals = lnew(v, len);
+        v = (List *)val_alloc(v1->v.obj);
+        vals = lnew(v, iter.len);
         if (vals == NULL) {
-            val_destroy(&iter->v);
+            iter_destroy(&iter);
             goto failed;
         }
-        iter_next = iter->next;
-        for (i = 0; i < len && (o2 = iter_next(iter)) != NULL; i++) {
+        for (i = 0; i < iter.len && (o2 = iter.next(&iter)) != NULL; i++) {
             err = indexoffs(o2, ln, &offs2, epoint2);
             if (err != NULL) {
                 vals[i] = &err->v;
                 continue;
             }
             if (more) {
-                Obj *vv = v1->data[offs2];
-                vals[i] = vv->obj->slice(vv, op, indx + 1);
+                op->v1 = v1->data[offs2];
+                vals[i] = op->v1->obj->slice(op, indx + 1);
             } else {
                 vals[i] = val_reference(v1->data[offs2]);
             }
         }
-        val_destroy(&iter->v);
+        iter_destroy(&iter);
         v->len = i;
         return &v->v;
     }
@@ -562,19 +645,19 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
         if (err != NULL) return &err->v;
 
         if (length == 0) {
-            return val_reference((o1->obj == TUPLE_OBJ) ? &null_tuple->v : &null_list->v);
+            return val_reference((v1->v.obj == TUPLE_OBJ) ? &null_tuple->v : &null_list->v);
         }
 
         if (step == 1 && length == v1->len && !more) {
-            return val_reference(o1); /* original tuple */
+            return val_reference(&v1->v); /* original tuple */
         }
-        v = (List *)val_alloc(o1->obj);
+        v = (List *)val_alloc(v1->v.obj);
         vals = lnew(v, length);
         if (vals == NULL) goto failed;
         for (i = 0; i < length; i++) {
             if (more) {
-                Obj *vv = v1->data[offs];
-                vals[i] = vv->obj->slice(vv, op, indx + 1);
+                op->v1 = v1->data[offs];
+                vals[i] = op->v1->obj->slice(op, indx + 1);
             } else {
                 vals[i] = val_reference(v1->data[offs]);
             }
@@ -585,8 +668,8 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
     err = indexoffs(o2, ln, &offs2, epoint2);
     if (err != NULL) return &err->v;
     if (more) {
-        Obj *vv = v1->data[offs2];
-        return vv->obj->slice(vv, op, indx + 1);
+        op->v1 = v1->data[offs2];
+        return op->v1->obj->slice(op, indx + 1);
     }
     return val_reference(v1->data[offs2]);
 failed:
@@ -607,7 +690,6 @@ static MUST_CHECK Obj *calc2(oper_t op) {
         return o2->obj->rcalc2(op);
     }
     if (o2->obj == TUPLE_OBJ || o2->obj == LIST_OBJ) {
-        if (diagnostics.type_mixing && o1->obj != o2->obj) err_msg_type_mixing(op->epoint3);
         return calc2_list(op);
     }
     if (o2 == &none_value->v || o2->obj == ERROR_OBJ) {
@@ -641,6 +723,7 @@ static MUST_CHECK Obj *calc2(oper_t op) {
         if (inplace) {
             list = (List *)val_reference(o1);
             vals = list->data;
+            if (vals != list->u.val) list->u.s.hash = -1;
         } else {
             list = (List *)val_alloc(o1->obj);
             vals = lnew(list, v1->len);
@@ -689,7 +772,6 @@ static MUST_CHECK Obj *rcalc2(oper_t op) {
         return (Obj *)ref_bool(false_value);
     }
     if (o1->obj == TUPLE_OBJ || o1->obj == LIST_OBJ) {
-        if (diagnostics.type_mixing && o1->obj != o2->obj) err_msg_type_mixing(op->epoint3);
         return calc2_list(op);
     }
     if (o1 == &none_value->v || o1->obj == ERROR_OBJ) {
@@ -723,6 +805,7 @@ static MUST_CHECK Obj *rcalc2(oper_t op) {
         if (inplace) {
             v = (List *)val_reference(o2);
             vals = v->data;
+            if (vals != v->u.val) v->u.s.hash = -1;
         } else {
             v = (List *)val_alloc(o2->obj);
             vals = lnew(v, v2->len);
@@ -748,10 +831,12 @@ static MUST_CHECK Obj *rcalc2(oper_t op) {
 }
 
 static void init(Type *obj) {
+    obj->iterable = true;
     obj->destroy = destroy;
     obj->garbage = garbage;
     obj->same = same;
     obj->truth = truth;
+    obj->hash = hash;
     obj->len = len;
     obj->getiter = getiter;
     obj->calc1 = calc1;
