@@ -1,5 +1,5 @@
 /*
-    $Id: macro.c 1891 2019-02-11 20:25:43Z soci $
+    $Id: macro.c 2138 2020-01-02 01:15:03Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -40,15 +40,29 @@
 
 static int functionrecursion;
 
+struct macro_rpos_s {
+    size_t opos, olen, pos, len, param;
+};
+
 struct macro_pline_s {
-    size_t len;
     uint8_t *data;
+    size_t len;
+    struct macro_rpos_s *rpositions;
+    size_t rp, rlen;
+};
+
+struct macro_value_s {
+    const uint8_t *data;
+    size_t len;
+    size_t pos;
+    bool init;
 };
 
 struct macro_params_s {
     size_t len, size;
-    str_t *param, all;
+    struct macro_value_s *param, all;
     struct macro_pline_s pline;
+    bool used;
     Obj *macro;
 };
 
@@ -59,10 +73,60 @@ static struct {
 
 bool in_macro;
 
+const struct file_list_s *macro_error_translate(struct linepos_s *opoint, size_t pos) {
+    const struct file_list_s *ret = NULL;
+    if (pline == macro_parameters.current->pline.data) {
+        const struct file_list_s *flist = current_file_list;
+        size_t p = macro_parameters.p;
+        while (p != 0) {
+            p--;
+            const struct macro_pline_s *mline = &macro_parameters.params[p].pline;
+            size_t i;
+            for (i = 0; i < mline->rp; i++) {
+                size_t c = pos - mline->rpositions[i].pos;
+                if (c < mline->rpositions[i].len) {
+                    size_t param = mline->rpositions[i].param;
+                    if (param != SIZE_MAX) {
+                        if (macro_parameters.params[p].param[param].init) return ret;
+                        pos = macro_parameters.params[p].param[param].pos + c;
+                    } else {
+                        pos = macro_parameters.params[p].all.pos + c;
+                    }
+                    opoint->pos = pos;
+                    opoint->line = flist->epoint.line;
+                    ret = flist->parent;
+                    flist = ret;
+                    break;
+                }
+            }
+            if (i == mline->rp) break;
+            if (p > 0 && !macro_parameters.params[p - 1].used) break;
+        }
+    }
+    return ret;
+}
+
+size_t macro_error_translate2(size_t pos) {
+    const struct macro_pline_s *mline = &macro_parameters.current->pline;
+    if (macro_parameters.p != 0 && pline == mline->data) {
+        size_t i, pos2 = pos;
+        for (i = 0; i < mline->rp; i++) {
+            const struct macro_rpos_s *rpositions = &mline->rpositions[i];
+            if (rpositions->pos > pos2) break;
+            if (rpositions->pos + rpositions->len > pos2) {
+                pos = rpositions->opos;
+                break;
+            }
+            pos = pos + rpositions->olen - rpositions->len;
+        }
+    }
+    return pos;
+}
+
 /* ------------------------------------------------------------------------------ */
 bool mtranslate(void) {
     unsigned int q;
-    size_t j;
+    size_t j, n, op;
     size_t p, p2;
     size_t last, last2;
     struct macro_pline_s *mline;
@@ -76,7 +140,7 @@ bool mtranslate(void) {
     if (changed) return false;
     mline = &macro_parameters.current->pline;
 
-    q = p = p2 = 0; last = last2 = 0; fault = false;
+    q = p = p2 = n = 0; last = last2 = 0; fault = false;
     while (pline[p2] != 0) {
         str_t param;
         switch (pline[p2]) {
@@ -104,35 +168,43 @@ bool mtranslate(void) {
             j = (uint8_t)(pline[p2 + 1] - '1');
             if (j < 9) {   /* \1..\9 */
                 p += p2 - last2;
+                op = p2;
                 p2 += 2;
                 break;
             }
             if (j == ('@' - '1')) { /* \@ gives complete parameter list */
                 j = SIZE_MAX;
                 p += p2 - last2;
+                op = p2;
                 p2 += 2;
                 break;
             }
             if (j == ('{' - '1')) {
                 lpoint.pos = p2 + 2;
                 param.data = pline + lpoint.pos;
-                param.len = get_label();
+                param.len = get_label(param.data);
+                lpoint.pos += param.len;
                 if (pline[lpoint.pos] == '}') lpoint.pos++;
                 else param.len = 0;
             } else {
                 lpoint.pos = p2 + 1;
                 param.data = pline + lpoint.pos;
-                param.len = get_label();
+                param.len = get_label(param.data);
+                lpoint.pos += param.len;
             }
             if (param.len != 0) {
                 Macro *macro = (Macro *)macro_parameters.current->macro;
                 str_t cf;
                 p += p2 - last2;
+                op = p2;
                 p2 = lpoint.pos;
                 str_cfcpy(&cf, &param);
                 for (j = 0; j < macro->argc; j++) {
-                    if (macro->param[j].cfname.data == NULL) continue;
-                    if (str_cmp(&macro->param[j].cfname, &cf) == 0) break;
+                    const uint8_t *data;
+                    if (macro->param[j].cfname.len != cf.len) continue;
+                    data = macro->param[j].cfname.data;
+                    if (data[0] != cf.data[0]) continue;
+                    if (memcmp(data, cf.data, cf.len) == 0) break;
                 }
                 if (j < macro->argc) break;
                 lpoint.pos -= param.len + 1;
@@ -149,6 +221,7 @@ bool mtranslate(void) {
                 j = (uint8_t)(pline[p2 + 1] - '1');
                 if (j < 9) { /* @1..@9 */
                     p += p2 - last2;
+                    op = p2;
                     p2 += 2;
                     break;
                 }
@@ -158,9 +231,11 @@ bool mtranslate(void) {
         }
 
         if (j < macro_parameters.current->len) {
-            param = macro_parameters.current->param[j];
+            param.data = macro_parameters.current->param[j].data;
+            param.len = macro_parameters.current->param[j].len;
         } else if (j == SIZE_MAX) {
-            param = macro_parameters.current->all;
+            param.data = macro_parameters.current->all.data;
+            param.len = macro_parameters.current->all.len;
         } else {
             switch (macro_parameters.current->macro->obj->type) {
             case T_STRUCT:
@@ -183,11 +258,21 @@ bool mtranslate(void) {
             if (p < last) err_msg_out_of_memory(); /* overflow */
             memcpy(mline->data + last, pline + last2, p - last);
         }
+        if (n >= mline->rlen) {
+            mline->rlen += 8;
+            if (mline->rlen < 8) err_msg_out_of_memory(); /* overflow */
+            mline->rpositions = (struct macro_rpos_s *)reallocx(mline->rpositions, mline->rlen * sizeof *mline->rpositions);
+        }
+        mline->rpositions[n].opos = op;
+        mline->rpositions[n].olen = p2 - op;
+        mline->rpositions[n].pos = p;
+        mline->rpositions[n].param = j;
+        mline->rpositions[n++].len = param.len;
         switch (param.len) {
         case 0: 
             if (param.data == NULL) {
                 lpoint.pos = last2 + p - last;
-                err_msg_missing_argument(current_file_list, &lpoint, j);
+                err_msg_missing_argument(&lpoint, j);
                 fault = true;
             }
             break;
@@ -200,6 +285,7 @@ bool mtranslate(void) {
         }
         last = p; last2 = p2;
     }
+    mline->rp = n;
     if (last2 != 0) {
         while (p2 != 0 && (pline[p2 - 1] == 0x20 || pline[p2 - 1] == 0x09)) p2--;
         p += p2 - last2;
@@ -278,6 +364,12 @@ Obj *macro_recurse(Wait_types t, Obj *tmp2, Namespace *context, linepos_t epoint
         macro_parameters.current->size = 0;
         macro_parameters.current->pline.len = 0;
         macro_parameters.current->pline.data = NULL;
+        macro_parameters.current->pline.rpositions = NULL;
+        macro_parameters.current->pline.rp = 0;
+        macro_parameters.current->pline.rlen = 0;
+    }
+    if (macro_parameters.p > 0) {
+        macro_parameters.params[macro_parameters.p - 1].used = (pline == macro_parameters.params[macro_parameters.p - 1].pline.data);
     }
     macro_parameters.current = &macro_parameters.params[macro_parameters.p];
     macro_parameters.current->macro = val_reference(&macro->v);
@@ -290,7 +382,7 @@ Obj *macro_recurse(Wait_types t, Obj *tmp2, Namespace *context, linepos_t epoint
 
         ignore(); opoint = lpoint;
         for (;;) {
-            str_t *param, *params = macro_parameters.current->param;
+            struct macro_value_s *param, *params = macro_parameters.current->param;
             if ((here() == 0 || here() == ';') && p >= macro->argc) break;
             if (p >= macro_parameters.current->size) {
                 if (macro_parameters.current->size < macro->argc) macro_parameters.current->size = macro->argc;
@@ -299,15 +391,18 @@ Obj *macro_recurse(Wait_types t, Obj *tmp2, Namespace *context, linepos_t epoint
                     /*if (macro_parameters.current->size < 4) err_msg_out_of_memory();*/ /* overflow */
                 }
                 if (macro_parameters.current->size > SIZE_MAX / sizeof *params) err_msg_out_of_memory();
-                params = (str_t *)reallocx(params, macro_parameters.current->size * sizeof *params);
+                params = (struct macro_value_s *)reallocx(params, macro_parameters.current->size * sizeof *params);
                 macro_parameters.current->param = params;
             }
             param = params + p;
+            param->pos = lpoint.pos;
             param->data = pline + lpoint.pos;
             param->len = macro_param_find();
-            if (param->len == 0) {
+            param->init = (param->len == 0);
+            if (param->init) {
                 if (p < macro->argc) {
-                    *param = macro->param[p].init;
+                    param->data = macro->param[p].init.data;
+                    param->len = macro->param[p].init.len;
                 } else param->data = NULL;
                 if (param->data == NULL) {
                     if (macro->v.obj->type == T_STRUCT || macro->v.obj->type == T_UNION) {
@@ -325,6 +420,7 @@ Obj *macro_recurse(Wait_types t, Obj *tmp2, Namespace *context, linepos_t epoint
             ignore();
         }
         macro_parameters.current->len = p;
+        macro_parameters.current->all.pos = opoint.pos;
         macro_parameters.current->all.data = pline + opoint.pos;
         npoint = lpoint;
         while (npoint.pos > opoint.pos && (pline[npoint.pos-1] == 0x20 || pline[npoint.pos-1] == 0x09)) npoint.pos--;
@@ -336,13 +432,13 @@ Obj *macro_recurse(Wait_types t, Obj *tmp2, Namespace *context, linepos_t epoint
         if (context != NULL) pop_context();
     } else {
         line_t lin = lpoint.line;
-        bool labelexists;
-        struct star_s *s = new_star(vline, &labelexists);
+        bool starexists;
+        struct star_s *s = new_star(vline, &starexists);
         struct avltree *stree_old = star_tree;
         line_t ovline = vline;
 
         if (diagnostics.optimize) cpu_opt_invalidate();
-        if (labelexists && s->addr != star) {
+        if (starexists && s->addr != star) {
             if (fixeddig && pass > max_pass) err_msg_cant_calculate(NULL, &lpoint);
             fixeddig = false;
         }
@@ -381,8 +477,9 @@ Obj *mfunc_recurse(Mfunc *mfunc, Namespace *context, uint8_t strength, linepos_t
         return NULL;
     }
     for (i = 0; i < mfunc->argc; i++) {
+        const struct mfunc_param_s *param = &mfunc->param[i];
         bool labelexists;
-        if (mfunc->param[i].init == &default_value->v) {
+        if (param->init == &default_value->v) {
             size_t j, len = get_val_remaining();
             tuple = new_tuple(len);
             for (j = 0; j < len; j++) {
@@ -393,15 +490,15 @@ Obj *mfunc_recurse(Mfunc *mfunc, Namespace *context, uint8_t strength, linepos_t
             struct values_s *vs;
             vs = get_val();
             if (vs == NULL) {
-                val = mfunc->param[i].init;
+                val = param->init;
                 if (val == NULL) { max = i + 1; val = (Obj *)none_value; }
             } else {
                 val = vs->val;
             }
         }
-        label = new_label(&mfunc->param[i].name, context, strength, &labelexists, mfunc->file_list);
+        label = new_label(&param->name, context, strength, &labelexists, mfunc->file_list);
         if (labelexists) {
-            if (label->constant) err_msg_double_defined(label, &mfunc->param[i].name, &mfunc->param[i].epoint); /* not possible in theory */
+            if (label->constant) err_msg_double_defined(label, &param->name, &param->epoint); /* not possible in theory */
             else {
                 if (label->defpass != pass) {
                     label->ref = false;
@@ -411,9 +508,9 @@ Obj *mfunc_recurse(Mfunc *mfunc, Namespace *context, uint8_t strength, linepos_t
                 }
                 label->owner = false;
                 if (label->file_list != mfunc->file_list) {
-                    label_move(label, &mfunc->param[i].name, mfunc->file_list);
+                    label_move(label, &param->name, mfunc->file_list);
                 }
-                label->epoint = mfunc->param[i].epoint;
+                label->epoint = param->epoint;
                 val_replace(&label->value, val);
                 label->usepass = 0;
             }
@@ -421,7 +518,7 @@ Obj *mfunc_recurse(Mfunc *mfunc, Namespace *context, uint8_t strength, linepos_t
             label->constant = false;
             label->owner = false;
             label->value = val_reference(val);
-            label->epoint = mfunc->param[i].epoint;
+            label->epoint = param->epoint;
         }
     }
     if (tuple != NULL) val_destroy(&tuple->v);
@@ -429,8 +526,8 @@ Obj *mfunc_recurse(Mfunc *mfunc, Namespace *context, uint8_t strength, linepos_t
     if (max != 0) err_msg_argnum(args, max, mfunc->argc, epoint);
     {
         line_t lin = lpoint.line;
-        bool labelexists;
-        struct star_s *s = new_star(vline, &labelexists);
+        bool starexists;
+        struct star_s *s = new_star(vline, &starexists);
         struct avltree *stree_old = star_tree;
         line_t ovline = vline;
         size_t oldbottom;
@@ -438,7 +535,7 @@ Obj *mfunc_recurse(Mfunc *mfunc, Namespace *context, uint8_t strength, linepos_t
         in_macro = false;
 
         if (diagnostics.optimize) cpu_opt_invalidate();
-        if (labelexists && s->addr != star) {
+        if (starexists && s->addr != star) {
             if (fixeddig && pass > max_pass) err_msg_cant_calculate(NULL, &lpoint);
             fixeddig = false;
         }
@@ -490,8 +587,9 @@ void get_func_params(Mfunc *v) {
         }
         new_mfunc.param[i].epoint = lpoint;
         label.data = pline + lpoint.pos;
-        label.len = get_label();
+        label.len = get_label(label.data);
         if (label.len != 0) {
+            lpoint.pos += label.len;
             if (label.len > 1 && label.data[0] == '_' && label.data[1] == '_') {err_msg2(ERROR_RESERVED_LABL, &label, &new_mfunc.param[i].epoint);break;}
             if ((size_t)(label.data - v->file_list->file->data) < v->file_list->file->len) new_mfunc.param[i].name = label;
             else str_cpy(&new_mfunc.param[i].name, &label);
@@ -562,9 +660,10 @@ void get_macro_params(Obj *v) {
         }
         epoints[i] = lpoint;
         label.data = pline + lpoint.pos;
-        label.len = get_label();
+        label.len = get_label(label.data);
         if (label.len != 0) {
             str_t cf;
+            lpoint.pos += label.len;
             if (label.len > 1 && label.data[0] == '_' && label.data[1] == '_') {err_msg2(ERROR_RESERVED_LABL, &label, &epoints[i]);new_macro.param[i].cfname.len = 0; new_macro.param[i].cfname.data = NULL;}
             str_cfcpy(&cf, &label);
             if (cf.data == label.data) {
@@ -612,6 +711,27 @@ void get_macro_params(Obj *v) {
     free(epoints);
 }
 
+static bool clean_namespace(Namespace *v1) {
+    size_t n, n2;
+    if (v1->len == 0) return true;
+    for (n = n2 = 0; n <= v1->mask && n2 < v1->len; n++) {
+        const Label *p = v1->data[n];
+        if (p == NULL) continue;
+        if (p->constant) {
+            return false;
+        }
+        n2++;
+    }
+    for (n2 = 0; n2 < n; n2++) {
+        Label *p = v1->data[n2];
+        if (p == NULL) continue;
+        val_destroy(&p->v);
+        v1->data[n2] = NULL;
+    }
+    v1->len = 0;
+    return true;
+}
+
 Obj *mfunc2_recurse(Mfunc *mfunc, struct values_s *vals, size_t args, linepos_t epoint) {
     size_t i;
     Label *label;
@@ -619,7 +739,6 @@ Obj *mfunc2_recurse(Mfunc *mfunc, struct values_s *vals, size_t args, linepos_t
     Tuple *tuple;
     Obj *retval = NULL;
     Namespace *context;
-    struct linepos_s xpoint;
 
     if (mfunc->recursion_pass == pass) return NULL;
     if (functionrecursion>100) {
@@ -628,15 +747,40 @@ Obj *mfunc2_recurse(Mfunc *mfunc, struct values_s *vals, size_t args, linepos_t
         return NULL;
     }
 
-    xpoint.line = mfunc->line;
-    xpoint.pos = 0;
-    context = new_namespace(mfunc->file_list, &xpoint);
+    if (mfunc->inamespaces == null_tuple) {
+        val_destroy(&mfunc->inamespaces->v);
+        mfunc->inamespaces = new_tuple(1);
+        context = NULL;
+    } else {
+        List *lst = mfunc->inamespaces;
+        if (mfunc->ipoint >= lst->len) {
+            if ((lst->data == lst->u.val ? mfunc->ipoint >= lenof(lst->u.val) : mfunc->ipoint >= lst->u.s.max)) {
+                if (list_extend(lst)) {
+                    err_msg2(ERROR_OUT_OF_MEMORY, NULL, epoint);
+                    return (Obj *)ref_none();
+                }
+            }
+            lst->len++;
+            context = NULL;
+        } else context = (Namespace *)lst->data[mfunc->ipoint];
+    }
+    if (context == NULL) {
+        struct linepos_s xpoint;
+        xpoint.line = mfunc->line;
+        xpoint.pos = 0;
+        context = new_namespace(mfunc->file_list, &xpoint);
+        mfunc->inamespaces->data[mfunc->ipoint] = &context->v;
+    } else {
+        context->backr = context->forwr = 0;
+    }
+    mfunc->ipoint++;
 
     enterfile(mfunc->file_list->file, epoint);
     tuple = NULL;
     for (i = 0; i < mfunc->argc; i++) {
+        const struct mfunc_param_s *param = &mfunc->param[i];
         bool labelexists;
-        if (mfunc->param[i].init == &default_value->v) {
+        if (param->init == &default_value->v) {
             if (i < args) {
                 size_t j = i;
                 tuple = new_tuple(args - i);
@@ -651,23 +795,41 @@ Obj *mfunc2_recurse(Mfunc *mfunc, struct values_s *vals, size_t args, linepos_t
             }
             val = &tuple->v;
         } else {
-            val = (i < args) ? vals[i].val : (mfunc->param[i].init != NULL) ? mfunc->param[i].init : (Obj *)none_value;
+            val = (i < args) ? vals[i].val : (param->init != NULL) ? param->init : (Obj *)none_value;
         }
-        label = new_label(&mfunc->param[i].name, context, 0, &labelexists, current_file_list);
-        if (!labelexists) {
+        label = new_label(&param->name, context, 0, &labelexists, mfunc->file_list);
+        if (labelexists) {
+            if (label->constant) {
+                err_msg_double_defined(label, &param->name, &param->epoint);
+            } else {
+                if (label->defpass != pass) {
+                    label->ref = false;
+                    label->defpass = pass;
+                } else {
+                    if (diagnostics.unused.variable && label->usepass != pass) err_msg_unused_variable(label);
+                }
+                label->owner = false;
+                if (label->file_list != mfunc->file_list) {
+                    label_move(label, &param->name, mfunc->file_list);
+                }
+                label->epoint = param->epoint;
+                val_destroy(label->value);
+                label->value = val_reference(val);
+                label->usepass = 0;
+            }
+        } else {
             label->constant = false;
             label->owner = false;
             label->value = val_reference(val);
-            label->file_list = current_file_list;
-            label->epoint = mfunc->param[i].epoint;
+            label->epoint = param->epoint;
         }
     }
     if (tuple != NULL) val_destroy(&tuple->v);
     else if (i < args) err_msg_argnum(args, i, i, &vals[i].epoint);
     {
         line_t lin = lpoint.line;
-        bool labelexists;
-        struct star_s *s = new_star(vline, &labelexists);
+        bool starexists;
+        struct star_s *s = new_star(vline, &starexists);
         struct avltree *stree_old = star_tree;
         line_t ovline = vline;
         struct linepos_s opoint = lpoint;
@@ -679,7 +841,7 @@ Obj *mfunc2_recurse(Mfunc *mfunc, struct values_s *vals, size_t args, linepos_t
         in_macro = false;
 
         if (diagnostics.optimize) cpu_opt_invalidate();
-        if (labelexists && s->addr != star) {
+        if (starexists && s->addr != star) {
             if (fixeddig && pass > max_pass) err_msg_cant_calculate(NULL, &lpoint);
             fixeddig = false;
         }
@@ -733,8 +895,9 @@ Obj *mfunc2_recurse(Mfunc *mfunc, struct values_s *vals, size_t args, linepos_t
         in_macro = in_macro_old;
     }
     exitfile();
-    if (diagnostics.unused.macro || diagnostics.unused.consts || diagnostics.unused.label || diagnostics.unused.variable) unused_check(context);
-    val_destroy(&context->v);
+    if (context->v.refcount == 1 && clean_namespace(context)) {
+        mfunc->ipoint--;
+    }
     return retval;
 }
 
@@ -747,6 +910,7 @@ void init_macro(void) {
 void free_macro(void) {
     size_t i;
     for (i = 0; i < macro_parameters.len; i++) {
+        free(macro_parameters.params[i].pline.rpositions);
         free(macro_parameters.params[i].pline.data);
         free(macro_parameters.params[i].param);
     }
