@@ -1,5 +1,5 @@
 /*
-    $Id: mem.c 1863 2019-02-03 19:51:36Z soci $
+    $Id: mem.c 2197 2020-04-06 19:00:55Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -21,6 +21,10 @@
 #ifdef _WIN32
 #include <fcntl.h>
 #endif
+#ifdef __DJGPP__
+#include <io.h>
+#include <fcntl.h>
+#endif
 #include <errno.h>
 #include "error.h"
 #include "file.h"
@@ -39,47 +43,46 @@ static int memblockcomp(const void *a, const void *b) {
     return (ad > bd) ? 1 : 0;
 }
 
-static void memcomp(Memblocks *memblocks) {
+static void memcomp(Memblocks *memblocks, bool nomerge) {
     unsigned int i, j, k;
-    if (memblocks->compressed) return;
-    memblocks->compressed = true;
-    memjmp(memblocks, 0);
+    if (!memblocks->flattened) {
+        memblocks->flattened = true;
+        memjmp(memblocks, 0);
 
-    for (j = 0; j < memblocks->p; j++) { /* replace references with real copies */
-        Memblocks *b = memblocks->data[j].ref;
-        if (b != NULL) {
-            size_t rest;
-            memcomp(b);
-            rest = memblocks->p - j - 1;
-            memblocks->p += b->p - 1;
-            if (memblocks->p >= memblocks->len) {
-                memblocks->len = memblocks->p + 64;
-                if (/*memblocks->len < 64 ||*/ memblocks->len > SIZE_MAX / sizeof *memblocks->data) err_msg_out_of_memory(); /* overflow */
-                memblocks->data = (struct memblock_s *)reallocx(memblocks->data, memblocks->len * sizeof *memblocks->data);
-            }
-            memmove(&memblocks->data[j + b->p], &memblocks->data[j + 1], rest * sizeof *memblocks->data);
-            for (k = 0; k < b->p; k++) {
-                struct memblock_s *b2 = &memblocks->data[j + k];
-                b2->p = memblocks->mem.p;
-                b2->len = b->data[k].len;
-                b2->ref = NULL;
-                b2->addr = b->data[k].addr;
-                memblocks->mem.p += b2->len;
-                if (memblocks->mem.p >= memblocks->mem.len) {
-                    memblocks->mem.len = memblocks->mem.p + 0x1000;
-                    if (memblocks->mem.len < 0x1000) err_msg_out_of_memory(); /* overflow */
-                    memblocks->mem.data = (uint8_t *)reallocx(memblocks->mem.data, memblocks->mem.len);
+        for (j = 0; j < memblocks->p; j++) { /* replace references with real copies */
+            Memblocks *b = memblocks->data[j].ref;
+            if (b != NULL) {
+                size_t rest;
+                memcomp(b, nomerge);
+                rest = memblocks->p - j - 1;
+                memblocks->p += b->p - 1;
+                if (memblocks->p >= memblocks->len) {
+                    memblocks->len = memblocks->p + 64;
+                    if (/*memblocks->len < 64 ||*/ memblocks->len > SIZE_MAX / sizeof *memblocks->data) err_msg_out_of_memory(); /* overflow */
+                    memblocks->data = (struct memblock_s *)reallocx(memblocks->data, memblocks->len * sizeof *memblocks->data);
                 }
-                memcpy(&memblocks->mem.data[b2->p], &b->mem.data[b->data[k].p], b2->len);
+                memmove(&memblocks->data[j + b->p], &memblocks->data[j + 1], rest * sizeof *memblocks->data);
+                for (k = 0; k < b->p; k++) {
+                    struct memblock_s *b2 = &memblocks->data[j + k];
+                    b2->p = memblocks->mem.p;
+                    b2->len = b->data[k].len;
+                    b2->ref = NULL;
+                    b2->addr = b->data[k].addr;
+                    memblocks->mem.p += b2->len;
+                    if (memblocks->mem.p >= memblocks->mem.len) {
+                        memblocks->mem.len = memblocks->mem.p + 0x1000;
+                        if (memblocks->mem.len < 0x1000) err_msg_out_of_memory(); /* overflow */
+                        memblocks->mem.data = (uint8_t *)reallocx(memblocks->mem.data, memblocks->mem.len);
+                    }
+                    memcpy(&memblocks->mem.data[b2->p], &b->mem.data[b->data[k].p], b2->len);
+                }
+                j--;
+                val_destroy(&b->v);
             }
-            j--;
-            val_destroy(&b->v);
         }
     }
-    if (memblocks->p < 2) return;
-    if (arguments.output.mode == OUTPUT_XEX ||
-            arguments.output.mode == OUTPUT_IHEX ||
-            arguments.output.mode == OUTPUT_SREC) return;
+    if (memblocks->p < 2 || nomerge || memblocks->merged) return;
+    memblocks->merged = true;
 
     for (k = j = 0; j < memblocks->p; j++) {
         struct memblock_s *bj = &memblocks->data[j];
@@ -148,44 +151,30 @@ void memref(Memblocks *memblocks, Memblocks *ref) {
     block->addr = memblocks->lastaddr;
 }
 
-static void printrange(address_t start, address_t end) {
-    char temp[10], temp2[10];
-    sprintf(temp, "$%04" PRIaddress, start);
-    sprintf(temp2, "$%04" PRIaddress, end-1);
-    printf("Memory range: %10s-%-7s $%04" PRIaddress "\n", temp, temp2, end-start);
-}
-
 void memprint(Memblocks *memblocks) {
-    unsigned int i;
-    bool over;
-    address_t start, end;
+    size_t i;
+    char temp[10], temp2[10];
 
-    memcomp(memblocks);
+    memcomp(memblocks, false);
 
-    if (memblocks->p == 0) return;
-
-    i = 0;
-    start = memblocks->data[i].addr;
-    end = memblocks->data[i].addr + memblocks->data[i].len;
-    over = end < start;
-    for (i++; i < memblocks->p; i++) {
+    for (i = 0; i < memblocks->p;) {
         const struct memblock_s *block = &memblocks->data[i];
-        if (block->addr != end || over) {
-            printrange(start, end);
-            start = block->addr;
+        address_t start = block->addr;
+        address_t size = block->len;
+        for (i++; i < memblocks->p; i++) {
+            const struct memblock_s *b = &memblocks->data[i];
+            address_t addr = start + size;
+            if (b->addr != addr || addr < start) break;
+            size += b->len;
         }
-        end = block->addr + block->len;
-        over = end < block->addr;
+        sprintf(temp, "$%04" PRIaddress, start);
+        sprintf(temp2, "$%04" PRIaddress, (address_t)(start + size - 1));
+        printf("Memory range: %10s-%-7s $%04" PRIaddress "\n", temp, temp2, size);
     }
-    printrange(start, end);
-}
-
-static void putlw(unsigned int w, FILE *f) {
-    putc(w, f);
-    putc(w >> 8, f);
 }
 
 static void padding(size_t size, FILE *f) {
+    unsigned char nuls[256];
     while (size >= 0x80000000) {
         if (fseek(f, 0x40000000, SEEK_CUR) != 0) goto err;
         size -= 0x40000000;
@@ -193,75 +182,91 @@ static void padding(size_t size, FILE *f) {
     if ((long)size > 256 && fseek(f, (long)size, SEEK_CUR) == 0) {
         return;
     }
-err:while ((size--) != 0) if (putc(0, f) == EOF) break;
+err:
+    nuls[0] = 1;
+    while (size != 0) {
+        size_t db = size < sizeof nuls ? size : sizeof nuls;
+        size -= db;
+        if (nuls[0] != 0) memset(nuls, 0, db);
+        if (fwrite(nuls, db, 1, f) == 0) return;
+    }
 }
 
 static void output_mem_c64(FILE *fout, const Memblocks *memblocks, const struct output_s *output) {
     address_t pos, end;
     unsigned int i;
+    unsigned char header[4];
 
-    if (memblocks->p != 0) {
-        pos = memblocks->data[0].addr;
-        if (output->mode == OUTPUT_CBM || output->mode == OUTPUT_APPLE) {
-            putlw(pos, fout);
-        }
-        if (output->mode == OUTPUT_APPLE) {
-            end = memblocks->data[memblocks->p - 1].addr + memblocks->data[memblocks->p - 1].len;
-            end -= pos;
-            putlw(end, fout);
-        } else if (output->mode == OUTPUT_CBM) {
-            if (output->longaddr) putc(pos >> 16,fout);
-        }
-        for (i = 0; i < memblocks->p; i++) {
-            const struct memblock_s *block = &memblocks->data[i];
-            padding(block->addr - pos, fout);
-            if (fwrite(memblocks->mem.data + block->p, block->len, 1, fout) == 0) return;
-            pos = block->addr + block->len;
-        }
+    if (memblocks->p == 0) return;
+    pos = memblocks->data[0].addr;
+    switch (output->mode) {
+    case OUTPUT_CBM:
+        header[0] = pos;
+        header[1] = pos >> 8;
+        if (output->longaddr) {
+            header[2] = pos >> 16;
+            i = 3;
+        } else i = 2;
+        break;
+    case OUTPUT_APPLE:
+        header[0] = pos;
+        header[1] = pos >> 8;
+        end = memblocks->data[memblocks->p - 1].addr + memblocks->data[memblocks->p - 1].len;
+        end -= pos;
+        header[2] = end;
+        header[3] = end >> 8;
+        i = 4;
+        break;
+    default:
+        i = 0;
+        break;
+    }
+    if (i != 0 && fwrite(header, i, 1, fout) == 0) return;
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *block = &memblocks->data[i];
+        padding(block->addr - pos, fout);
+        if (fwrite(memblocks->mem.data + block->p, block->len, 1, fout) == 0) return;
+        pos = block->addr + block->len;
     }
 }
 
 static void output_mem_nonlinear(FILE *fout, const Memblocks *memblocks, bool longaddr) {
-    address_t start, size;
-    unsigned int i, last;
-
-    if (memblocks->p != 0) {
-        i = 0;
-        start = memblocks->data[i].addr;
-        size = memblocks->data[i].len;
-        last = i;
-        for (i++; i < memblocks->p; i++) {
-            const struct memblock_s *block = &memblocks->data[i];
-            if (block->addr != start + size) {
-                putlw(size,fout);
-                if (longaddr) putc(size >> 16,fout);
-                putlw(start,fout);
-                if (longaddr) putc(start >> 16,fout);
-                while (last < i) {
-                    const struct memblock_s *b = &memblocks->data[last++];
-                    if (fwrite(memblocks->mem.data + b->p, b->len, 1, fout) == 0) return;
-                }
-                start = block->addr;
-                size = 0;
-            }
-            size += block->len;
+    size_t i, j;
+    unsigned char header[6];
+    for (i = 0; i < memblocks->p;) {
+        const struct memblock_s *block = &memblocks->data[i];
+        address_t start = block->addr;
+        address_t size = block->len;
+        for (j = i + 1; j < memblocks->p; j++) {
+            const struct memblock_s *b = &memblocks->data[j];
+            address_t addr = start + size;
+            if (b->addr != addr || addr < start) break;
+            size += b->len;
         }
-        putlw(size,fout);
-        if (longaddr) putc(size >> 16, fout);
-        putlw(start, fout);
-        if (longaddr) putc(start >> 16,fout);
-        while (last < i) {
-            const struct memblock_s *b = &memblocks->data[last++];
+        header[0] = size;
+        header[1] = size >> 8;
+        if (longaddr) {
+            header[2] = size >> 16;
+            header[3] = start;
+            header[4] = start >> 8;
+            header[5] = start >> 16;
+        } else {
+            header[2] = start;
+            header[3] = start >> 8;
+        }
+        if (fwrite(header, longaddr ? 6 : 4, 1, fout) == 0) return;
+        for (;i < j; i++) {
+            const struct memblock_s *b = &memblocks->data[i];
             if (fwrite(memblocks->mem.data + b->p, b->len, 1, fout) == 0) return;
         }
     }
-    putlw(0, fout);
-    if (longaddr) putc(0, fout);
+    memset(header, 0, 4);
+    fwrite(header, longaddr ? 3 : 2, 1, fout);
 }
 
 static void output_mem_flat(FILE *fout, const Memblocks *memblocks) {
     address_t pos;
-    unsigned int i;
+    size_t i;
 
     for (pos = i = 0; i < memblocks->p; i++) {
         const struct memblock_s *block = &memblocks->data[i];
@@ -272,43 +277,45 @@ static void output_mem_flat(FILE *fout, const Memblocks *memblocks) {
 }
 
 static void output_mem_atari_xex(FILE *fout, const Memblocks *memblocks) {
-    address_t start, size;
-    unsigned int i, last;
-
-    if (memblocks->p != 0) {
-        i = 0;
-        start = memblocks->data[i].addr;
-        size = memblocks->data[i].len;
-        last = i;
-        for (i++; i < memblocks->p; i++) {
-            const struct memblock_s *block = &memblocks->data[i];
-            if (block->addr != start + size) {
-                if (last == 0 || start == 0xffff) putlw(0xffff, fout);
-                putlw(start, fout);
-                putlw(start + size - 1, fout);
-                while (last < i) {
-                    const struct memblock_s *b = &memblocks->data[last++];
-                    if (fwrite(memblocks->mem.data + b->p, b->len, 1, fout) == 0) return;
-                }
-                start = block->addr;
-                size = 0;
-            }
-            size += block->len;
+    size_t i, j;
+    unsigned char header[6];
+    header[0] = 0xff;
+    header[1] = 0xff;
+    for (i = 0; i < memblocks->p;) {
+        const struct memblock_s *block = &memblocks->data[i];
+        bool special;
+        address_t end, start = block->addr;
+        address_t size = block->len;
+        for (j = i + 1; j < memblocks->p; j++) {
+            const struct memblock_s *b = &memblocks->data[j];
+            address_t addr = start + size;
+            if (b->addr != addr || addr < start) break;
+            size += b->len;
         }
-        if (last == 0 || start == 0xffff) putlw(0xffff, fout);
-        putlw(start, fout);
-        putlw(start + size - 1, fout);
-        while (last < i) {
-            const struct memblock_s *b = &memblocks->data[last++];
+        special = (i == 0 || start == 0xffff);
+        header[2] = start;
+        header[3] = start >> 8;
+        end = start + size - 1;
+        header[4] = end;
+        header[5] = end >> 8;
+        if (fwrite(special ? header : &header[2], special ? 6 : 4, 1, fout) == 0) return;
+        for (;i < j; i++) {
+            const struct memblock_s *b = &memblocks->data[i];
             if (fwrite(memblocks->mem.data + b->p, b->len, 1, fout) == 0) return;
         }
     }
 }
 
-static void hexput(FILE *fout, unsigned int b) {
+struct hexput_s {
+    char *line;
+    unsigned int sum;
+};
+
+static void hexput(struct hexput_s *h, unsigned int b) {
     const char *hex = "0123456789ABCDEF";
-    putc(hex[(b >> 4) & 0xf], fout);
-    putc(hex[b & 0xf], fout);
+    *h->line++ = hex[b >> 4];
+    *h->line++ = hex[b & 0xf];
+    h->sum += b;
 }
 
 struct ihex_s {
@@ -318,23 +325,25 @@ struct ihex_s {
     unsigned int length;
 };
 
-static void output_mem_ihex_line(FILE *fout, unsigned int length, address_t address, unsigned int type, const uint8_t *data) {
-    unsigned int i, sum = length + address + (address >> 8) + type;
-    putc(':', fout);
-    hexput(fout, length);
-    hexput(fout, address >> 8);
-    hexput(fout, address);
-    hexput(fout, type);
+static void output_mem_ihex_line(struct ihex_s *ihex, unsigned int length, address_t address, unsigned int type, const uint8_t *data) {
+    unsigned int i;
+    char line[1+(1+2+1+32+1)*2+1];
+    struct hexput_s h = { line + 1, 0 };
+    line[0] = ':';
+    hexput(&h, length);
+    hexput(&h, (address >> 8) & 0xff);
+    hexput(&h, address & 0xff);
+    hexput(&h, type);
     for (i = 0; i < length; i++) {
-        hexput(fout, data[i]);
-        sum += data[i];
+        hexput(&h, data[i]);
     }
-    hexput(fout, -sum);
-    putc('\n', fout);
+    hexput(&h, (-h.sum) & 0xff);
+    *h.line++ = '\n';
+    fwrite(line, h.line - line, 1, ihex->file);
 }
 
 static void output_mem_ihex_data(struct ihex_s *ihex) {
-    uint8_t *data = ihex->data;
+    const uint8_t *data = ihex->data;
     if ((ihex->address & 0xffff) + ihex->length > 0x10000) {
         uint16_t length = (uint16_t)-ihex->address;
         unsigned int remains = ihex->length - length;
@@ -347,10 +356,10 @@ static void output_mem_ihex_data(struct ihex_s *ihex) {
         uint8_t ez[2];
         ez[0] = (uint8_t)(ihex->address >> 24);
         ez[1] = (uint8_t)(ihex->address >> 16);
-        output_mem_ihex_line(ihex->file, sizeof ez, 0, 4, ez);
+        output_mem_ihex_line(ihex, sizeof ez, 0, 4, ez);
         ihex->segment = ihex->address;
     }
-    output_mem_ihex_line(ihex->file, ihex->length, ihex->address, 0, data);
+    output_mem_ihex_line(ihex, ihex->length, ihex->address, 0, data);
     ihex->address += ihex->length;
     ihex->length = 0;
 }
@@ -359,35 +368,33 @@ static void output_mem_ihex(FILE *fout, const Memblocks *memblocks) {
     struct ihex_s ihex;
     size_t i;
 
-    if (memblocks->p != 0) {
-        ihex.file = fout;
-        ihex.address = 0;
-        ihex.segment = 0;
-        ihex.length = 0;
-        for (i = 0; i < memblocks->p; i++) {
-            const struct memblock_s *b = &memblocks->data[i];
-            const uint8_t *d = memblocks->mem.data + b->p;
-            address_t addr = b->addr;
-            size_t blen = b->len;
-            if (blen != 0 && ihex.address + ihex.length != addr) {
-                if (ihex.length != 0) output_mem_ihex_data(&ihex);
-                ihex.address = addr;
-            }
-            while (blen != 0) {
-                size_t left = (sizeof ihex.data) - ihex.length;
-                size_t copy = blen > left ? left : blen;
-                memcpy(ihex.data + ihex.length, d, copy);
-                ihex.length += copy;
-                d += copy;
-                blen -= copy;
-                if (ihex.length == sizeof ihex.data) {
-                    output_mem_ihex_data(&ihex);
-                }
+    ihex.file = fout;
+    ihex.address = 0;
+    ihex.segment = 0;
+    ihex.length = 0;
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *b = &memblocks->data[i];
+        const uint8_t *d = memblocks->mem.data + b->p;
+        address_t addr = b->addr;
+        size_t blen = b->len;
+        if (blen != 0 && ihex.address + ihex.length != addr) {
+            if (ihex.length != 0) output_mem_ihex_data(&ihex);
+            ihex.address = addr;
+        }
+        while (blen != 0) {
+            size_t left = (sizeof ihex.data) - ihex.length;
+            size_t copy = blen > left ? left : blen;
+            memcpy(ihex.data + ihex.length, d, copy);
+            ihex.length += copy;
+            d += copy;
+            blen -= copy;
+            if (ihex.length == sizeof ihex.data) {
+                output_mem_ihex_data(&ihex);
             }
         }
-        if (ihex.length != 0) output_mem_ihex_data(&ihex);
-        output_mem_ihex_line(fout, 0, 0, 1, NULL);
     }
+    if (ihex.length != 0) output_mem_ihex_data(&ihex);
+    output_mem_ihex_line(&ihex, 0, 0, 1, NULL);
 }
 
 struct srecord_s {
@@ -400,68 +407,66 @@ struct srecord_s {
 
 static void output_mem_srec_line(struct srecord_s *srec) {
     unsigned int i;
-    unsigned int sum = srec->length + srec->address + (srec->address >> 8) + (srec->address >> 16) + (srec->address >> 24);
-    putc('S', srec->file);
-    putc(srec->length ? ('1' + srec->type) : ('9' - srec->type), srec->file);
-    sum += srec->type + 3;
-    hexput(srec->file, srec->length + srec->type + 3);
-    if (srec->type > 1) hexput(srec->file, srec->address >> 24);
-    if (srec->type > 0) hexput(srec->file, srec->address >> 16);
-    hexput(srec->file, srec->address >> 8);
-    hexput(srec->file, srec->address);
+    char line[1+1+(1+4+32+1)*2+1];
+    struct hexput_s h = { line + 2, 0 };
+    line[0] = 'S';
+    line[1] = srec->length != 0 ? ('1' + srec->type) : ('9' - srec->type);
+    hexput(&h, srec->length + srec->type + 3);
+    if (srec->type > 1) hexput(&h, (srec->address >> 24) & 0xff);
+    if (srec->type > 0) hexput(&h, (srec->address >> 16) & 0xff);
+    hexput(&h, (srec->address >> 8) & 0xff);
+    hexput(&h, srec->address & 0xff);
     for (i = 0; i < srec->length; i++) {
-        hexput(srec->file, srec->data[i]);
-        sum += srec->data[i];
+        hexput(&h, srec->data[i]);
     }
-    hexput(srec->file, ~sum);
-    putc('\n', srec->file);
+    hexput(&h, (~h.sum) & 0xff);
+    *h.line++ = '\n';
     srec->address += srec->length;
     srec->length = 0;
+    fwrite(line, h.line - line, 1, srec->file);
 }
 
 static void output_mem_srec(FILE *fout, const Memblocks *memblocks) {
     struct srecord_s srec;
     size_t i;
 
-    if (memblocks->p != 0) {
-        srec.file = fout;
-        srec.type = 0;
-        srec.address = 0;
-        srec.length = 0;
-        for (i = 0; i < memblocks->p; i++) {
-            const struct memblock_s *b = &memblocks->data[i];
-            size_t end = b->addr + b->len - 1;
-            if (end >= 0x10000 && srec.type < 1) srec.type = 1;
-            if (end >= 0x1000000 && srec.type < 2) {
-                srec.type = 2;
-                break;
-            }
+    srec.file = fout;
+    srec.type = 0;
+    srec.address = 0;
+    srec.length = 0;
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *b = &memblocks->data[i];
+        size_t end = b->addr + b->len - 1;
+        if (end >= 0x10000 && srec.type < 1) srec.type = 1;
+        if (end >= 0x1000000 && srec.type < 2) {
+            srec.type = 2;
+            break;
         }
-        for (i = 0; i < memblocks->p; i++) {
-            const struct memblock_s *b = &memblocks->data[i];
-            const uint8_t *d = memblocks->mem.data + b->p;
-            address_t addr = b->addr;
-            size_t blen = b->len;
-            if (blen != 0 && srec.address + srec.length != addr) {
-                if (srec.length != 0) output_mem_srec_line(&srec);
-                srec.address = addr;
-            }
-            while (blen != 0) {
-                size_t left = (sizeof srec.data) - srec.length;
-                size_t copy = blen > left ? left : blen;
-                memcpy(srec.data + srec.length, d, copy);
-                srec.length += copy;
-                d += copy;
-                blen -= copy;
-                if (srec.length == sizeof srec.data) {
-                    output_mem_srec_line(&srec);
-                }
-            }
-        }
-        if (srec.length != 0) output_mem_srec_line(&srec);
-        srec.address = memblocks->data[0].addr;
-        output_mem_srec_line(&srec);
     }
+    for (i = 0; i < memblocks->p; i++) {
+        const struct memblock_s *b = &memblocks->data[i];
+        const uint8_t *d = memblocks->mem.data + b->p;
+        address_t addr = b->addr;
+        size_t blen = b->len;
+        if (blen != 0 && srec.address + srec.length != addr) {
+            if (srec.length != 0) output_mem_srec_line(&srec);
+            srec.address = addr;
+        }
+        while (blen != 0) {
+            size_t left = (sizeof srec.data) - srec.length;
+            size_t copy = blen > left ? left : blen;
+            memcpy(srec.data + srec.length, d, copy);
+            srec.length += copy;
+            d += copy;
+            blen -= copy;
+            if (srec.length == sizeof srec.data) {
+                output_mem_srec_line(&srec);
+            }
+        }
+    }
+    if (srec.length != 0) output_mem_srec_line(&srec);
+    srec.address = memblocks->data[0].addr;
+    output_mem_srec_line(&srec);
 }
 
 void output_mem(Memblocks *memblocks, const struct output_s *output) {
@@ -469,14 +474,17 @@ void output_mem(Memblocks *memblocks, const struct output_s *output) {
     struct linepos_s nopoint = {0, 0};
     bool binary = (output->mode != OUTPUT_IHEX) && (output->mode != OUTPUT_SREC);
     int err;
+#if defined _WIN32 || defined __DJGPP__
+    int oldmode = -1;
+#endif
 
-    memcomp(memblocks);
+    memcomp(memblocks, output->mode == OUTPUT_XEX || output->mode == OUTPUT_IHEX || output->mode == OUTPUT_SREC);
 
     if (memblocks->mem.p == 0) return;
 
     if (dash_name(output->name)) {
-#ifdef _WIN32
-        if (binary) setmode(fileno(stdout), O_BINARY);
+#if defined _WIN32 || defined __DJGPP__
+        if (binary) oldmode = setmode(STDOUT_FILENO, O_BINARY);
 #endif
         fout = stdout;
     } else {
@@ -500,8 +508,8 @@ void output_mem(Memblocks *memblocks, const struct output_s *output) {
     err = ferror(fout);
     err |= (fout != stdout) ? fclose(fout) : fflush(fout);
     if (err != 0 && errno != 0) err_msg_file(ERROR_CANT_WRTE_OBJ, output->name, &nopoint);
-#ifdef _WIN32
-    setmode(fileno(stdout), O_TEXT);
+#if defined _WIN32 || defined __DJGPP__
+    if (oldmode >= 0) setmode(STDOUT_FILENO, oldmode);
 #endif
 }
 
@@ -528,11 +536,12 @@ FAST_CALL uint8_t *alloc_mem(Memblocks *memblocks, size_t len) {
 
 static size_t omemp;
 static size_t ptextaddr;
-static address_t oaddr, oaddr2;
+static address_t oaddr, oaddr2, olastaddr;
 
 void mark_mem(const Memblocks *memblocks, address_t adr, address_t adr2) {
     ptextaddr = memblocks->mem.p;
     omemp = memblocks->p;
+    olastaddr = memblocks->lastaddr;
     oaddr = adr;
     oaddr2 = adr2;
 }
@@ -596,18 +605,17 @@ void list_mem(const Memblocks *memblocks) {
         }
 
         if (first) {
-            if (addr2 < addr) {
-                addr = addr2;
+            first = false;
+            if (addr != olastaddr) {
                 len = 0;
                 o--;
             } else {
-                address_t diff = addr2 - addr;
+                address_t diff = (addr2 - addr) & all_mem2;
                 if (diff > len) continue;
-                addr = addr2;
                 p += diff;
                 len -= diff;
             }
-            first = false;
+            addr = addr2;
         } else {
             if (len == 0) continue;
         }
