@@ -1,5 +1,5 @@
 /*
-    $Id: mfuncobj.h 1794 2019-01-12 15:36:49Z soci $
+    $Id: mfuncobj.h 2137 2020-01-02 00:52:17Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@
 #define MFUNCOBJ_H
 #include "obj.h"
 #include "str.h"
+#include "stdbool.h"
 
 extern struct Type *const MFUNC_OBJ;
 
@@ -34,11 +35,14 @@ typedef struct Mfunc {
     Obj v;
     size_t argc;
     struct mfunc_param_s *param;
-    struct file_list_s *file_list;
+    const struct file_list_s *file_list;
     line_t line;
+    bool retval;
     uint8_t recursion_pass;
     size_t nslen;
-    struct Namespace **namespaces;
+    struct Namespace **namespaces, *names;
+    size_t ipoint;
+    struct List *inamespaces;
 } Mfunc;
 
 extern void mfuncobj_init(void);
