@@ -1,5 +1,5 @@
 /*
-    $Id: strobj.c 1858 2019-01-31 18:50:31Z soci $
+    $Id: strobj.c 2122 2019-12-21 06:27:50Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -52,15 +52,27 @@ static MUST_CHECK Obj *create(Obj *v1, linepos_t epoint) {
     }
 }
 
+static FAST_CALL NO_INLINE void str_destroy(Str *v1) {
+    free(v1->data);
+}
+
 static FAST_CALL void destroy(Obj *o1) {
     Str *v1 = (Str *)o1;
-    if (v1->u.val != v1->data) free(v1->data);
+    if (v1->u.val != v1->data) str_destroy(v1);
+}
+
+static FAST_CALL NO_INLINE bool str_same(const Str *v1, const Str *v2) {
+    return memcmp(v1->data, v2->data, v2->len) == 0;
 }
 
 static FAST_CALL bool same(const Obj *o1, const Obj *o2) {
     const Str *v1 = (const Str *)o1, *v2 = (const Str *)o2;
-    return o2->obj == STR_OBJ && v1->len == v2->len && (
-            v1->data == v2->data || memcmp(v1->data, v2->data, v2->len) == 0);
+    if (o1->obj != o2->obj || v1->len != v2->len) return false;
+    switch (v1->len) {
+    case 0: return true;
+    case 1: return v1->data[0] == v2->data[0];
+    default: return v1->data == v2->data || str_same(v1, v2);
+    }
 }
 
 static MUST_CHECK Obj *truth(Obj *o1, Truth_types type, linepos_t epoint) {
@@ -248,52 +260,43 @@ static MUST_CHECK Obj *sign(Obj *o1, linepos_t epoint) {
     return ret;
 }
 
-static MUST_CHECK Obj *function(Obj *o1, Func_types UNUSED(f), linepos_t epoint) {
-    Str *v1 = (Str *)o1;
-    return int_from_str(v1, epoint);
+static MUST_CHECK Obj *function(oper_t op) {
+    Str *v1 = (Str *)op->v2;
+    return int_from_str(v1, op->epoint2);
 }
 
-static MUST_CHECK Obj *len(Obj *o1, linepos_t UNUSED(epoint)) {
-    Str *v1 = (Str *)o1;
+static MUST_CHECK Obj *len(oper_t op) {
+    Str *v1 = (Str *)op->v2;
     return (Obj *)int_from_size(v1->chars);
 }
 
-static size_t iter_len(Iter *v1) {
-    return ((Str *)v1->data)->chars - v1->val;
-}
-
-static FAST_CALL MUST_CHECK Obj *next(Iter *v1) {
+static FAST_CALL MUST_CHECK Obj *next(struct iter_s *v1) {
     Str *iter;
-    const Str *str = (Str *)v1->data;
+    const Str *string = (Str *)v1->data;
     unsigned int ln;
     const uint8_t *s;
-    if (v1->val >= str->len) return NULL;
-    s = str->data + v1->val;
+    if (v1->val >= string->len) return NULL;
+    s = string->data + v1->val;
     ln = utf8len(*s);
     v1->val += ln;
     iter = (Str *)v1->iter;
-    if (iter == NULL) {
-    renew:
+    if (iter->v.refcount != 1) {
+        iter->v.refcount--;
         iter = new_str(6);
         iter->chars = 1;
         v1->iter = &iter->v;
-    } else if (iter->v.refcount != 1) {
-        iter->v.refcount--;
-        goto renew;
     }
     iter->len = ln;
     memcpy(iter->data, s, ln);
     return &iter->v;
 }
 
-static MUST_CHECK Iter *getiter(Obj *v1) {
-    Iter *v = (Iter *)val_alloc(ITER_OBJ);
-    v->iter = NULL;
+static void getiter(struct iter_s *v) {
+    v->iter = val_reference(v->data);
     v->val = 0;
-    v->data = val_reference(v1);
+    v->data = val_reference(v->data);
     v->next = next;
-    v->len = iter_len;
-    return v;
+    v->len = ((Str *)v->data)->chars;
 }
 
 MUST_CHECK Obj *str_from_str(const uint8_t *s, size_t *ln, linepos_t epoint) {
@@ -355,7 +358,8 @@ MALLOC Str *new_str(size_t ln) {
     return v;
 }
 
-static int icmp(Str *v1, Str *v2) {
+static int icmp(oper_t op) {
+    const Str *v1 = (Str *)op->v1, *v2 = (Str *)op->v2;
     int h = memcmp(v1->data, v2->data, (v1->len < v2->len) ? v1->len : v2->len);
     if (h != 0) return h;
     if (v1->len < v2->len) return -1;
@@ -382,7 +386,7 @@ static MUST_CHECK Obj *calc1(oper_t op) {
     default: return obj_oper_error(op);
     }
     op->v1 = tmp;
-    op->inplace = NULL;
+    op->inplace = (tmp->refcount == 1) ? tmp : NULL;
     v = tmp->obj->calc1(op);
     val_destroy(tmp);
     return v;
@@ -404,7 +408,7 @@ static MUST_CHECK Obj *calc2_str(oper_t op) {
             tmp2 = int_from_str(v2, op->epoint2);
             op->v1 = tmp;
             op->v2 = tmp2;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp2);
             val_destroy(tmp);
@@ -419,7 +423,7 @@ static MUST_CHECK Obj *calc2_str(oper_t op) {
             tmp2 = bytes_from_str(v2, op->epoint2, BYTES_MODE_TEXT);
             op->v1 = tmp;
             op->v2 = tmp2;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp2);
             val_destroy(tmp);
@@ -433,24 +437,24 @@ static MUST_CHECK Obj *calc2_str(oper_t op) {
             tmp2 = bits_from_str(v2, op->epoint2);
             op->v1 = tmp;
             op->v2 = tmp2;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp2);
             val_destroy(tmp);
             return result;
         }
     case O_CMP:
-        val = icmp(v1, v2);
+        val = icmp(op);
         if (val < 0) return (Obj *)ref_int(minus1_value);
         return (Obj *)ref_int(int_value[(val > 0) ? 1 : 0]);
-    case O_EQ: return truth_reference(icmp(v1, v2) == 0);
-    case O_NE: return truth_reference(icmp(v1, v2) != 0);
+    case O_EQ: return truth_reference(icmp(op) == 0);
+    case O_NE: return truth_reference(icmp(op) != 0);
     case O_MIN:
-    case O_LT: return truth_reference(icmp(v1, v2) < 0);
-    case O_LE: return truth_reference(icmp(v1, v2) <= 0);
+    case O_LT: return truth_reference(icmp(op) < 0);
+    case O_LE: return truth_reference(icmp(op) <= 0);
     case O_MAX:
-    case O_GT: return truth_reference(icmp(v1, v2) > 0);
-    case O_GE: return truth_reference(icmp(v1, v2) >= 0);
+    case O_GT: return truth_reference(icmp(op) > 0);
+    case O_GE: return truth_reference(icmp(op) >= 0);
     case O_CONCAT:
         if (v1->len == 0) {
             return (Obj *)ref_str(v2);
@@ -538,19 +542,18 @@ static inline MUST_CHECK Obj *repeat(oper_t op) {
     return (Obj *)new_error_mem(op->epoint3);
 }
 
-static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
+static MUST_CHECK Obj *slice(oper_t op, size_t indx) {
     uint8_t *p, *p2;
     size_t offs2;
     size_t len1, len2;
     Obj *o2 = op->v2;
-    Str *v, *v1 = (Str *)o1;
+    Str *v, *v1 = (Str *)op->v1;
     Funcargs *args = (Funcargs *)o2;
     Error *err;
     linepos_t epoint2;
 
     if (args->len < 1 || args->len > indx + 1) {
-        err_msg_argnum(args->len, 1, indx + 1, op->epoint2);
-        return (Obj *)ref_none();
+        return (Obj *)new_error_argnum(args->len, 1, indx + 1, op->epoint2);
     }
 
     o2 = args->val[indx].val;
@@ -558,35 +561,33 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
 
     len1 = v1->chars;
 
-    if (o2->obj == LIST_OBJ) {
-        iter_next_t iter_next;
-        Iter *iter = o2->obj->getiter(o2);
+    if (o2->obj->iterable) {
+        struct iter_s iter;
         size_t i;
+        iter.data = o2; o2->obj->getiter(&iter);
 
-        len2 = iter->len(iter);
-        if (len2 == 0) {
-            val_destroy(&iter->v);
+        if (iter.len == 0) {
+            iter_destroy(&iter);
             return (Obj *)ref_str(null_str);
         }
 
-        iter_next = iter->next;
         if (v1->len == v1->chars) {
-            v = new_str2(len2);
+            v = new_str2(iter.len);
             if (v == NULL) {
-                val_destroy(&iter->v);
+                iter_destroy(&iter);
                 goto failed;
             }
             p2 = v->data;
-            for (i = 0; i < len2 && (o2 = iter_next(iter)) != NULL; i++) {
+            for (i = 0; i < iter.len && (o2 = iter.next(&iter)) != NULL; i++) {
                 err = indexoffs(o2, len1, &offs2, epoint2);
                 if (err != NULL) {
                     val_destroy(&v->v);
-                    val_destroy(&iter->v);
+                    iter_destroy(&iter);
                     return &err->v;
                 }
                 *p2++ = v1->data[offs2];
             }
-            val_destroy(&iter->v);
+            iter_destroy(&iter);
             len1 = i;
         } else {
             size_t m = v1->len;
@@ -595,18 +596,18 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
 
             v = new_str2(m);
             if (v == NULL) {
-                val_destroy(&iter->v);
+                iter_destroy(&iter);
                 goto failed;
             }
             o = p2 = v->data;
             p = v1->data;
 
-            for (i = 0; i < len2 && (o2 = iter_next(iter)) != NULL; i++) {
+            for (i = 0; i < iter.len && (o2 = iter.next(&iter)) != NULL; i++) {
                 unsigned int k;
                 err = indexoffs(o2, len1, &offs2, epoint2);
                 if (err != NULL) {
                     val_destroy(&v->v);
-                    val_destroy(&iter->v);
+                    iter_destroy(&iter);
                     return &err->v;
                 }
                 while (offs2 != j) {
@@ -623,13 +624,13 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
                     const uint8_t *r = o;
                     m += 4096;
                     if (m < 4096) {
-                        val_destroy(&iter->v);
+                        iter_destroy(&iter);
                         goto failed2; /* overflow */
                     }
                     if (o == v->u.val) {
                         o = (uint8_t *)malloc(m);
                         if (o == NULL) {
-                            val_destroy(&iter->v);
+                            iter_destroy(&iter);
                             goto failed2;
                         }
                         v->data = o;
@@ -638,7 +639,7 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
                     } else {
                         o = (uint8_t *)realloc(o, m);
                         if (o == NULL) {
-                            val_destroy(&iter->v);
+                            iter_destroy(&iter);
                             goto failed2;
                         }
                         v->data = o;
@@ -648,7 +649,7 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
                 }
                 memcpy(p2, p, k);p2 += k;
             }
-            val_destroy(&iter->v);
+            iter_destroy(&iter);
             len1 = i;
             len2 = (size_t)(p2 - o);
             if (o != v->u.val) {
@@ -696,27 +697,59 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
                 }
                 len2 = (size_t)(p - v1->data) - len2;
             }
-            v = new_str2(len2);
-            if (v == NULL) goto failed;
+            if (op->inplace == &v1->v) {
+                v = ref_str(v1);
+                v->len = len2;
+                if (v->data != v->u.val && len2 <= sizeof v->u.val) {
+                    memcpy(v->u.val, v1->data + offs, len2);
+                    free(v->data);
+                    v->data = v->u.val;
+                } else {
+                    if (offs != 0) memmove(v->data, v1->data + offs, len2);
+                    if (v->data != v->u.val) v->u.s.hash = -1;
+                }
+            } else {
+                v = new_str2(len2);
+                if (v == NULL) goto failed;
+                memcpy(v->data, v1->data + offs, len2);
+            }
             v->chars = length;
-            memcpy(v->data, v1->data + offs, len2);
             return &v->v;
         }
         if (v1->len == v1->chars) {
             size_t i;
-            v = new_str2(length);
-            if (v == NULL) goto failed;
-            p2 = v->data;
+            if (step > 0 && op->inplace == &v1->v) {
+                v = ref_str(v1);
+                if (v->data != v->u.val && length <= sizeof v->u.val) {
+                    p2 = v->u.val;
+                } else {
+                    p2 = v->data;
+                    if (v->data != v->u.val) v->u.s.hash = -1;
+                }
+                v->len = length;
+            } else {
+                v = new_str2(length);
+                if (v == NULL) goto failed;
+                p2 = v->data;
+            }
             for (i = 0; i < length; i++) {
                 p2[i] = v1->data[offs];
                 offs += step;
             }
+            if (p2 != v->data) {
+                free(v->data);
+                v->data = p2;
+            }
         } else {
             ival_t i, k;
             size_t j;
             uint8_t *o;
-            v = new_str2(v1->len);
-            if (v == NULL) goto failed;
+            if (step > 0 && op->inplace == &v1->v) {
+                v = ref_str(v1);
+            } else {
+                v = new_str2(v1->len);
+                if (v == NULL) goto failed;
+            }
             o = p2 = v->data;
             p = v1->data;
             for (i = 0; i < offs; i++) {
@@ -725,7 +758,10 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
             if (step > 0) {
                 for (k = i; i < end; i++) {
                     j = utf8len(*p);
-                    if (i == k) {memcpy(p2, p, j);p2 += j; k += step;}
+                    if (i != k) {
+                        for (offs2 = 0; offs2 < j; offs2++) p2[offs2] = p[offs2];
+                        p2 += j; k += step;
+                    }
                     p += j;
                 }
             } else {
@@ -748,6 +784,7 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
                     uint8_t *oo = (uint8_t *)realloc(o, len2);
                     v->data = (oo != NULL) ? oo : o;
                     v->u.s.max = len2;
+                    v->u.s.hash = -1;
                 }
             }
             v->len = len2;
@@ -759,19 +796,36 @@ static MUST_CHECK Obj *slice(Obj *o1, oper_t op, size_t indx) {
     if (err != NULL) return &err->v;
 
     if (v1->len == v1->chars) {
-        v = new_str2(1);
+        p = v1->data + offs2;
+        len1 = 1;
+    } else {
+        p = v1->data;
+        for (; offs2 != 0; offs2--) p += utf8len(*p);
+        len1 = utf8len(*p);
+    }
+
+    if (op->inplace == &v1->v) {
+        v = ref_str(v1);
+        if (v->data != v->u.val) {
+            p2 = v->u.val;
+        } else {
+            p2 = v->data;
+        }
+        v->len = len1;
+    } else {
+        v = new_str2(len1);
         if (v == NULL) goto failed;
-        v->chars = 1;
-        v->data[0] = v1->data[offs2];
-        return &v->v;
+        p2 = v->data;
     }
-    p = v1->data;
-    while ((offs2--) != 0) p += utf8len(*p);
-    len1 = utf8len(*p);
-    v = new_str2(len1);
-    if (v == NULL) goto failed;
     v->chars = 1;
-    memcpy(v->data, p, len1);
+    p2[0] = p[0];
+    if (len1 > 1) {
+        for (offs2 = 1; offs2 < len1; offs2++) p2[offs2] = p[offs2];
+    }
+    if (p2 != v->data) {
+        free(v->data);
+        v->data = p2;
+    }
     return &v->v;
 failed2:
     val_destroy(&v->v);
@@ -797,6 +851,11 @@ static MUST_CHECK Obj *calc2(oper_t op) {
         if (diagnostics.strict_bool) err_msg_bool_oper(op);
         return val_reference(i ? v2 : &v1->v);
     }
+    if (v2->obj->iterable) {
+        if (op->op != &o_MEMBER) {
+            return v2->obj->rcalc2(op);
+        }
+    }
     switch (v2->obj->type) {
     case T_STR: return calc2_str(op);
     case T_BOOL:
@@ -819,7 +878,7 @@ static MUST_CHECK Obj *calc2(oper_t op) {
             default: tmp = int_from_str(v1, op->epoint);
             }
             op->v1 = tmp;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp);
             return result;
@@ -829,15 +888,12 @@ static MUST_CHECK Obj *calc2(oper_t op) {
             Obj *result;
             tmp = bytes_from_str(v1, op->epoint, BYTES_MODE_TEXT);
             op->v1 = tmp;
-            op->inplace = NULL;
+            op->inplace = (tmp->refcount == 1) ? tmp : NULL;
             result = tmp->obj->calc2(op);
             val_destroy(tmp);
             return result;
         }
-    case T_TUPLE:
-    case T_LIST:
     case T_GAP:
-    case T_REGISTER:
     case T_DICT:
         if (op->op != &o_MEMBER) {
             return v2->obj->rcalc2(op);
@@ -913,6 +969,9 @@ void strobj_init(void) {
     obj.str = str;
     obj.ival = ival;
     obj.uval = uval;
+    obj.uval2 = uval;
+    obj.iaddress = ival;
+    obj.uaddress = uval;
     obj.sign = sign;
     obj.function = function;
     obj.len = len;
