@@ -1,5 +1,5 @@
 /*
-    $Id: variables.c 1786 2019-01-07 03:49:17Z soci $
+    $Id: variables.c 2137 2020-01-02 00:52:17Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -48,7 +48,6 @@
 #include "errorobj.h"
 #include "mfuncobj.h"
 
-static struct namespacekey_s *lastlb2 = NULL;
 static Label *lastlb = NULL;
 
 #define EQUAL_COLUMN 16
@@ -57,6 +56,7 @@ Namespace *root_namespace;
 static Namespace *builtin_namespace;
 Namespace *current_context;
 Namespace *cheap_context;
+size_t fwcount;
 
 struct cstack_s {
     Namespace *normal;
@@ -100,6 +100,29 @@ bool pop_context(void) {
     return true;
 }
 
+void push_context2(Namespace *name) {
+    if (context_stack.p >= context_stack.len) {
+        context_stack.len += 8;
+        if (/*context_stack.len < 8 ||*/ context_stack.len > SIZE_MAX / sizeof *context_stack.stack) err_msg_out_of_memory(); /* overflow */
+        context_stack.stack = (struct cstack_s *)reallocx(context_stack.stack, context_stack.len * sizeof *context_stack.stack);
+    }
+    context_stack.stack[context_stack.p].normal = context_stack.stack[context_stack.p - 1].normal;
+    context_stack.stack[context_stack.p - 1].normal = ref_namespace(name);
+    context_stack.stack[context_stack.p].cheap = ref_namespace(name);
+    context_stack.p++;
+}
+
+bool pop_context2(void) {
+    if (context_stack.p > 1 + context_stack.bottom) {
+        struct cstack_s *c = &context_stack.stack[--context_stack.p];
+        val_destroy(&context_stack.stack[context_stack.p - 1].normal->v);
+        context_stack.stack[context_stack.p - 1].normal = c->normal;
+        val_destroy(&c->cheap->v);
+        return false;
+    }
+    return true;
+}
+
 void reset_context(void) {
     context_stack.bottom = 0;
     while (context_stack.p != 0) {
@@ -156,213 +179,242 @@ size_t context_get_bottom(void) {
 
 /* --------------------------------------------------------------------------- */
 
-static int label_compare(const struct avltree_node *aa, const struct avltree_node *bb)
-{
-    const struct namespacekey_s *a = cavltree_container_of(aa, struct namespacekey_s, node);
-    const struct namespacekey_s *b = cavltree_container_of(bb, struct namespacekey_s, node);
-    const str_t *s1, *s2;
-    int h = a->hash - b->hash;
-    if (h != 0) return h;
-    s1 = &a->key->cfname;
-    s2 = &b->key->cfname;
-    if (s1->len != s2->len) return s1->len > s2->len ? 1 : -1;
-    if (s1->data == s2->data) return 0;
-    return memcmp(s1->data, s2->data, s1->len);
+static Label *namespace_update(Namespace *ns, Label *p) {
+    size_t mask, hash, offs;
+    if (ns->len * 3 / 2 >= ns->mask) {
+        size_t i, max = (ns->data == NULL) ? 8 : (ns->mask + 1) << 1;
+        Label **n = (Label **)calloc(max, sizeof *n);
+        if (n == NULL) err_msg_out_of_memory();
+        mask = max - 1;
+        if (ns->data != NULL) {
+            for (i = 0; i <= ns->mask; i++) if (ns->data[i] != NULL) {
+                hash = (size_t)ns->data[i]->hash;
+                offs = hash & mask;
+                while (n[offs] != NULL) {
+                    hash >>= 5;
+                    offs = (5 * offs + hash + 1) & mask;
+                }
+                n[offs] = ns->data[i];
+            }
+            free(ns->data);
+        }
+        ns->mask = mask;
+        ns->data = n;
+    }
+    mask = ns->mask;
+    hash = (size_t)p->hash;
+    offs = hash & mask;
+    while (ns->data[offs] != NULL) {
+        Label *d = ns->data[offs];
+        if (p->hash == d->hash && p->strength == d->strength) {
+            const str_t *s1 = &p->cfname;
+            const str_t *s2 = &d->cfname;
+            if (s1->len == s2->len && (s1->data == s2->data || memcmp(s1->data, s2->data, s1->len) == 0)) {
+                return d;
+            }
+        }
+        hash >>= 5;
+        offs = (5 * offs + hash + 1) & mask;
+    } 
+    ns->data[offs] = p;
+    ns->len++;
+    return NULL;
 }
 
-static int label_compare2(const struct avltree_node *aa, const struct avltree_node *bb)
-{
-    const struct namespacekey_s *a = cavltree_container_of(aa, struct namespacekey_s, node);
-    const struct namespacekey_s *b = cavltree_container_of(bb, struct namespacekey_s, node);
-    int h = a->hash - b->hash;
-    if (h != 0) return h;
-    h = str_cmp(&a->key->cfname, &b->key->cfname);
-    if (h != 0) return h;
-    return b->key->strength - a->key->strength;
+static Label *namespace_lookup(const Namespace *ns, const Label *p) {
+    Label *ret = NULL;
+    size_t mask = ns->mask;
+    size_t hash = (size_t)p->hash;
+    size_t offs = hash & mask;
+    if (ns->data == NULL) return ret;
+    while (ns->data[offs] != NULL) {
+        Label *d = ns->data[offs];
+        if (p->hash == d->hash) {
+            if (d->defpass == pass || (d->constant && (!fixeddig || d->defpass == pass - 1))) {
+                const str_t *s1 = &p->cfname;
+                const str_t *s2 = &d->cfname;
+                if (s1->len == s2->len && (s1->data == s2->data || memcmp(s1->data, s2->data, s1->len) == 0)) {
+                    if (d->strength == 0) { ret = d; break; }
+                    if (ret == NULL || d->strength < ret->strength) ret = d;
+                }
+            }
+        }
+        hash >>= 5;
+        offs = (5 * offs + hash + 1) & mask;
+    }
+    if (ret != NULL && ret->constant && ret->defpass == pass - 1 && ret->fwpass != pass) {
+        ret->fwpass = pass;
+        fwcount++;
+    }
+    return ret;
 }
 
-static struct namespacekey_s *strongest_label(struct avltree_node *b) {
-    struct namespacekey_s *a = NULL, *c;
-    struct avltree_node *n = b;
+static Label *namespace_lookup2(const Label *p) {
+    const Namespace *ns = builtin_namespace;
+    size_t mask = ns->mask;
+    size_t hash = (size_t)p->hash;
+    size_t offs = hash & mask;
+    while (ns->data[offs] != NULL) {
+        Label *d = ns->data[offs];
+        if (p->hash == d->hash) {
+            const str_t *s1 = &p->cfname;
+            const str_t *s2 = &d->cfname;
+            if (s1->len == s2->len && memcmp(s1->data, s2->data, s1->len) == 0) {
+                return d;
+            }
+        }
+        hash >>= 5;
+        offs = (5 * offs + hash + 1) & mask;
+    }
+    return NULL;
+}
 
-    do {
-        c = avltree_container_of(n, struct namespacekey_s, node);
-        if (c->key->defpass == pass || (c->key->constant && (!fixeddig || c->key->defpass == pass - 1))) {
-            if (c->key->strength == 0) return c;
-            a = c;
+static Label *namespace_lookup3(const Namespace *ns, const Label *p) {
+    size_t mask = ns->mask;
+    size_t hash = (size_t)p->hash;
+    size_t offs = hash & mask;
+    if (ns->data == NULL) return NULL;
+    while (ns->data[offs] != NULL) {
+        Label *d = ns->data[offs];
+        if (p->hash == d->hash && p->strength == d->strength) {
+            const str_t *s1 = &p->cfname;
+            const str_t *s2 = &d->cfname;
+            if (s1->len == s2->len && (s1->data == s2->data || memcmp(s1->data, s2->data, s1->len) == 0)) {
+                return d;
+            }
         }
-        n = avltree_next(n);
-    } while (n != NULL && label_compare(n, b) == 0);
-    if (a != NULL) return a;
-    n = avltree_prev(b);
-    while (n != NULL && label_compare(n, b) == 0) {
-        c = avltree_container_of(n, struct namespacekey_s, node);
-        if (c->key->defpass == pass || (c->key->constant && (!fixeddig || c->key->defpass == pass - 1))) {
-            return c;
-        }
-        n = avltree_prev(n);
+        hash >>= 5;
+        offs = (5 * offs + hash + 1) & mask;
     }
     return NULL;
 }
 
 Label *find_label(const str_t *name, Namespace **here) {
-    struct avltree_node *b;
-    struct namespacekey_s tmp, *c;
     size_t p = context_stack.p;
-    Label label;
+    Label label, *c;
 
     str_cfcpy(&label.cfname, name);
-    tmp.hash = str_hash(&label.cfname);
-    tmp.key = &label;
+    label.hash = str_hash(&label.cfname);
 
     while (context_stack.bottom < p) {
         Namespace *context = context_stack.stack[--p].normal;
-        b = avltree_lookup(&tmp.node, &context->members, label_compare);
-        if (b != NULL) {
-            c = strongest_label(b);
-            if (c != NULL) {
-                Label *key2 = c->key;
-                if (!diagnostics.shadow || !fixeddig || constcreated || (here != NULL && *here == context)) {
-                    if (here != NULL) *here = context;
-                    return key2;
-                }
-                if (here != NULL) *here = context;
-                while (context_stack.bottom < p) {
-                    b = avltree_lookup(&tmp.node, &context_stack.stack[--p].normal->members, label_compare);
-                    if (b != NULL) {
-                        const struct namespacekey_s *l2 = strongest_label(b);
-                        if (l2 != NULL) {
-                            Label *key1 = l2->key;
-                            Obj *o1 = key1->value;
-                            Obj *o2 = key2->value;
-                            if (o1 != o2 && !o1->obj->same(o1, o2)) {
-                                err_msg_shadow_defined(key1, key2);
-                                return key2;
-                            }
-                        }
-                    }
-                }
-                b = avltree_lookup(&tmp.node, &builtin_namespace->members, label_compare);
-                if (b != NULL) {
-                    const struct namespacekey_s *l2 = cavltree_container_of(b, struct namespacekey_s, node);
-                    Label *key1 = l2->key;
+        Label *key2 = namespace_lookup(context, &label);
+        if (key2 != NULL) {
+            if (here != NULL) *here = context;
+            if (!diagnostics.shadow || !fixeddig || constcreated || (here != NULL && *here == context)) {
+                return key2;
+            }
+            while (context_stack.bottom < p) {
+                Label *key1 = namespace_lookup(context_stack.stack[--p].normal, &label);
+                if (key1 != NULL) {
                     Obj *o1 = key1->value;
                     Obj *o2 = key2->value;
                     if (o1 != o2 && !o1->obj->same(o1, o2)) {
-                        err_msg_shadow_defined2(key2);
+                        err_msg_shadow_defined(key1, key2);
+                        return key2;
                     }
                 }
-                return key2;
             }
+            c = namespace_lookup2(&label);
+            if (c != NULL) {
+                Obj *o1 = c->value;
+                Obj *o2 = key2->value;
+                if (o1 != o2 && !o1->obj->same(o1, o2)) {
+                    err_msg_shadow_defined2(key2);
+                }
+            }
+            return key2;
         }
     }
-    b = avltree_lookup(&tmp.node, &builtin_namespace->members, label_compare);
-    if (b != NULL) {
-        if (here != NULL) *here = builtin_namespace;
-        return avltree_container_of(b, struct namespacekey_s, node)->key;
-    }
-    if (here != NULL) *here = NULL;
-    return NULL;
+    c = namespace_lookup2(&label);
+    if (here != NULL) *here = (c != NULL) ? builtin_namespace : NULL;
+    return c;
 }
+
 Label *find_label2(const str_t *name, Namespace *context) {
-    struct avltree_node *b;
-    struct namespacekey_s tmp, *c;
     Label label;
 
     str_cfcpy(&label.cfname, name);
-    tmp.hash = str_hash(&label.cfname);
-    tmp.key = &label;
+    label.hash = str_hash(&label.cfname);
 
-    b = avltree_lookup(&tmp.node, &context->members, label_compare);
-    if (b == NULL) return NULL;
-    c = strongest_label(b);
-    return (c != NULL) ? c->key : NULL;
+    return namespace_lookup(context, &label);
 }
 
-static struct {
-    uint8_t dir;
-    uint8_t padding[3];
-    int32_t count;
-} anon_idents;
+struct anonident_s {
+    uint8_t dir, pad;
+    uint8_t count[sizeof(uint32_t)];
+};
 
 Label *find_label3(const str_t *name, Namespace *context, uint8_t strength) {
-    struct avltree_node *b;
-    struct namespacekey_s tmp, *c;
     Label label;
 
     label.strength = strength;
-    if (name->len == sizeof anon_idents && name->data[1] == 0) label.cfname = *name;
+    if (name->len > 1 && name->data[1] == 0) label.cfname = *name;
     else str_cfcpy(&label.cfname, name);
-    tmp.hash = str_hash(&label.cfname);
-    tmp.key = &label;
+    label.hash = str_hash(&label.cfname);
 
-    b = avltree_lookup(&tmp.node, &context->members, label_compare2);
-    if (b == NULL) return NULL;
-    c = avltree_container_of(b, struct namespacekey_s, node);
-    return (c != NULL) ? c->key : NULL;
+    return namespace_lookup3(context, &label);
 }
 
 Label *find_anonlabel(int32_t count) {
-    struct avltree_node *b;
-    struct namespacekey_s tmp, *c;
     size_t p = context_stack.p;
     Namespace *context;
-    Label label;
+    Label label, *c;
+    struct anonident_s anonident;
 
-    anon_idents.dir = (count >= 0) ? '+' : '-';
+    anonident.dir = (count >= 0) ? '+' : '-';
+    anonident.pad = 0;
 
-    label.cfname.data = (const uint8_t *)&anon_idents;
-    label.cfname.len = sizeof anon_idents;
-    tmp.key = &label;
+    label.cfname.data = (const uint8_t *)&anonident;
 
     while (context_stack.bottom < p) {
         context = context_stack.stack[--p].normal;
-        anon_idents.count = (int32_t)((count >= 0) ? context->forwr : context->backr) + count;
-        tmp.hash = str_hash(&label.cfname);
-        b = avltree_lookup(&tmp.node, &context->members, label_compare);
-        if (b != NULL) {
-            c = strongest_label(b);
-            if (c != NULL) return c->key;
-        }
+        uint32_t count2 = (uint32_t)((int32_t)((count >= 0) ? context->forwr : context->backr) + count);
+        label.cfname.len = 2;
+        while (count2 != 0) {
+            anonident.count[label.cfname.len - 2] = count2;
+            label.cfname.len++;
+            count2 >>= 8;
+        } 
+
+        label.hash = str_hash(&label.cfname);
+        c = namespace_lookup(context, &label);
+        if (c != NULL) return c;
     }
     return NULL;
 }
 
 Label *find_anonlabel2(int32_t count, Namespace *context) {
-    struct avltree_node *b;
-    struct namespacekey_s tmp, *c;
     Label label;
+    uint32_t count2 = (uint32_t)((int32_t)((count >= 0) ? 0 : context->backr) + count);
+    struct anonident_s anonident;
 
-    anon_idents.dir = (count >= 0) ? '+' : '-';
-    anon_idents.count = (int32_t)((count >= 0) ? 0 : context->backr) + count;
+    anonident.dir = (count >= 0) ? '+' : '-';
+    anonident.pad = 0;
 
-    label.cfname.data = (const uint8_t *)&anon_idents;
-    label.cfname.len = sizeof anon_idents;
-    tmp.hash = str_hash(&label.cfname);
-    tmp.key = &label;
+    label.cfname.data = (const uint8_t *)&anonident;
+    label.cfname.len = 2;
+    while (count2 != 0) {
+        anonident.count[label.cfname.len - 2] = count2;
+        label.cfname.len++;
+        count2 >>= 8;
+    } 
+    label.hash = str_hash(&label.cfname);
 
-    b = avltree_lookup(&tmp.node, &context->members, label_compare);
-    if (b == NULL) return NULL;
-    c = strongest_label(b);
-    return (c != NULL) ? c->key : NULL;
+    return namespace_lookup(context, &label);
 }
 
 /* --------------------------------------------------------------------------- */
-Label *new_label(const str_t *name, Namespace *context, uint8_t strength, bool *exists, struct file_list_s *cflist) {
-    struct avltree_node *b;
-    Label *tmp;
-    if (lastlb2 == NULL) {
-        lastlb2 = namespacekey_alloc();
-    }
+Label *new_label(const str_t *name, Namespace *context, uint8_t strength, bool *exists, const struct file_list_s *cflist) {
+    Label *b;
     if (lastlb == NULL) lastlb = (Label *)val_alloc(LABEL_OBJ);
 
     if (name->len > 1 && name->data[1] == 0) lastlb->cfname = *name;
     else str_cfcpy(&lastlb->cfname, name);
-    lastlb2->hash = str_hash(&lastlb->cfname);
-    lastlb2->key = lastlb;
+    lastlb->hash = str_hash(&lastlb->cfname);
     lastlb->strength = strength;
 
-    b = avltree_insert(&lastlb2->node, &context->members, label_compare2);
+    b = namespace_update(context, lastlb);
 
     if (b == NULL) { /* new label */
         if ((size_t)(name->data - cflist->file->data) < cflist->file->len) lastlb->name = *name;
@@ -373,19 +425,18 @@ Label *new_label(const str_t *name, Namespace *context, uint8_t strength, bool *
         lastlb->ref = false;
         lastlb->update_after = false;
         lastlb->usepass = 0;
+        lastlb->fwpass = 0;
         lastlb->defpass = pass;
         *exists = false;
-        tmp = lastlb;
+        b = lastlb;
         lastlb = NULL;
-        lastlb2 = NULL;
-        context->len++;
-        return tmp;
+        return b;
     }
     *exists = true;
-    return avltree_container_of(b, struct namespacekey_s, node)->key;            /* already exists */
+    return b;
 }
 
-void label_move(Label *label, const str_t *name, struct file_list_s *cflist) {
+void label_move(Label *label, const str_t *name, const struct file_list_s *cflist) {
     bool cfsame = (label->cfname.data == label->name.data);
     if ((size_t)(label->name.data - label->file_list->file->data) < label->file_list->file->len) {
         if ((size_t)(name->data - cflist->file->data) < cflist->file->len) label->name = *name;
@@ -397,29 +448,19 @@ void label_move(Label *label, const str_t *name, struct file_list_s *cflist) {
     label->file_list = cflist;
 }
 
-void unused_check(Namespace *members) {
-    const struct avltree_node *n;
+void unused_check(Namespace *names) {
+    size_t n, ln;
 
-    for (n = avltree_first(&members->members); n != NULL; n = avltree_next(n)) {
-        const struct namespacekey_s *l = cavltree_container_of(n, struct namespacekey_s, node);
-        Label *key2 = l->key;
+    if (names->len == 0) return;
+    ln = names->len; names->len = 0;
+    for (n = 0; n <= names->mask; n++) {
+        Label *key2 = names->data[n];
         Obj *o;
         Namespace *ns;
 
-        if (key2->defpass != pass) continue;
+        if (key2 == NULL || key2->defpass != pass) continue;
 
         o  = key2->value;
-        switch (o->obj->type) {
-        case T_CODE:
-            ns = ((Code *)o)->names;
-            break;
-        case T_NAMESPACE:
-            ns = (Namespace *)o;
-            break;
-        default:
-            ns = NULL;
-            break;
-        }
         if (key2->usepass != pass && (key2->name.data[0] != '.' && key2->name.data[0] != '#')) {
             if (!key2->constant) {
                 if (diagnostics.unused.variable) err_msg_unused_variable(key2);
@@ -438,34 +479,37 @@ void unused_check(Namespace *members) {
                 continue;
             }
         }
-        if (ns != NULL && ns->len != 0 && key2->owner) {
-            size_t ln = ns->len;
-            ns->len = 0;
+        if (!key2->owner) continue;
+        switch (o->obj->type) {
+        case T_CODE:
+            ns = ((Code *)o)->names;
+            break;
+        case T_NAMESPACE:
+            ns = (Namespace *)o;
+            break;
+        case T_MFUNC:
+            {
+                Mfunc *mfunc = (Mfunc *)o;
+                List *lst = mfunc->inamespaces;
+                size_t i;
+                for (i = 0; i < lst->len; i++) {
+                    ns = (Namespace *)lst->data[i];
+                    if (ns->len != 0) unused_check(ns);
+                }
+                ns = mfunc->names;
+            }
+            break;
+        default:
+            ns = NULL;
+            break;
+        }
+        if (ns != NULL && ns->len != 0) {
             push_context(ns);
             unused_check(ns);
             pop_context();
-            ns->len = ln;
         }
     }
-}
-
-static Label *find_strongest_label(struct avltree_node **x, avltree_cmp_fn_t cmp) {
-    struct namespacekey_s *a = NULL, *c;
-    struct avltree_node *b = *x, *n = b;
-    do {
-        c = avltree_container_of(n, struct namespacekey_s, node);
-        if (c->key->defpass == pass) a = c;
-        n = avltree_next(n);
-    } while (n != NULL && cmp(n, b) == 0);
-    *x = n;
-    if (a != NULL) return a->key;
-    n = avltree_prev(b);
-    while (n != NULL && cmp(n, b) == 0) {
-        c = avltree_container_of(n, struct namespacekey_s, node);
-        if (c->key->defpass == pass) return c->key;
-        n = avltree_prev(n);
-    }
-    return NULL;
+    names->len = ln;
 }
 
 static inline void padding(size_t l, size_t t, FILE *f) {
@@ -476,7 +520,7 @@ static inline void padding(size_t l, size_t t, FILE *f) {
     while (l < t) { l++; putc(' ', f);}
 }
 
-static void labelname_print(Label *l, FILE *flab, char d) {
+static void labelname_print(const Label *l, FILE *flab, char d) {
     size_t p;
     for (p = 0; p < label_stack.p; p++) {
         printable_print2(label_stack.stack[p]->name.data, flab, label_stack.stack[p]->name.len);
@@ -485,13 +529,34 @@ static void labelname_print(Label *l, FILE *flab, char d) {
     printable_print2(l->name.data, flab, l->name.len);
 }
 
-static void labelprint2(const struct avltree *members, FILE *flab, int labelmode) {
-    struct avltree_node *n;
-    Label *l;
+struct vice_address_s {
+    address_t addr;
+    struct avltree_node node;
+};
 
-    n = avltree_first(members);
-    while (n != NULL) {
-        l = find_strongest_label(&n, label_compare);            /* already exists */
+static struct vice_addresses_s {
+    struct vice_address_s list[256];
+    struct vice_addresses_s *next;
+} *vice_addresses = NULL;
+
+static struct avltree vice_filter;
+static struct vice_address_s *lastvc;
+static int vice_addrp;
+
+static FAST_CALL int duplicate_compare(const struct avltree_node *aa, const struct avltree_node *bb)
+{
+    const struct vice_address_s *a = cavltree_container_of(aa, struct vice_address_s, node);
+    const struct vice_address_s *b = cavltree_container_of(bb, struct vice_address_s, node);
+    return (int)a->addr - (int)b->addr;
+}
+
+static void labelprint2(Namespace *names, FILE *flab, int labelmode) {
+    size_t n, ln;
+
+    if (names->len == 0) return;
+    ln = names->len; names->len = 0;
+    for (n = 0; n <= names->mask; n++) {
+        Label *l = names->data[n];
         if (l == NULL || l->name.data == NULL) continue;
         if (l->name.len > 1 && l->name.data[1] == 0) continue;
         switch (l->value->obj->type) {
@@ -500,13 +565,18 @@ static void labelprint2(const struct avltree *members, FILE *flab, int labelmode
         case T_SEGMENT:
         case T_UNION:
         case T_STRUCT: continue;
+        case T_CODE:
+            if (((Code *)l->value)->pass != ((Code *)l->value)->apass) continue;
+            break;
         default:break;
         }
-        if (labelmode == LABEL_VICE) {
+        if (l != namespace_lookup(names, l)) continue;
+        if (labelmode == LABEL_VICE || labelmode == LABEL_VICE_NUMERIC) {
             Obj *val;
             size_t i, j = l->name.len;
             const uint8_t *d = l->name.data;
 
+            if (!l->constant) continue;
             for (i = 0; i < j; i++) {
                 uint8_t c = d[i];
                 if (c < '0') break;
@@ -520,11 +590,20 @@ static void labelprint2(const struct avltree *members, FILE *flab, int labelmode
             if (i != j) continue;
 
             val = l->value;
-            if (val->obj == ADDRESS_OBJ || val->obj == CODE_OBJ) {
+            if (val->obj == ADDRESS_OBJ || val->obj == CODE_OBJ || (labelmode == LABEL_VICE_NUMERIC && (val->obj == BITS_OBJ || val->obj == INT_OBJ))) {
                 struct linepos_s epoint;
                 uval_t uv;
                 Error *err = val->obj->uval(val, &uv, 24, &epoint);
                 if (err == NULL) {
+                    lastvc->addr = uv;
+                    if (avltree_insert(&lastvc->node, &vice_filter, duplicate_compare) != NULL) continue;
+                    if (vice_addrp == 255) {
+                        struct vice_addresses_s *old = vice_addresses;
+                        vice_addresses = (struct vice_addresses_s *)mallocx(sizeof *vice_addresses);
+                        vice_addresses->next = old;
+                        vice_addrp = 0;
+                    } else vice_addrp++;
+                    lastvc = &vice_addresses->list[vice_addrp];
                     fprintf(flab, "al %" PRIx32 " .", uv & 0xffffff);
                     labelname_print(l, flab, ':');
                     putc('\n', flab);
@@ -533,12 +612,9 @@ static void labelprint2(const struct avltree *members, FILE *flab, int labelmode
             if (l->owner) {
                 Namespace *ns = get_namespace(val);
                 if (ns != NULL && ns->len != 0) {
-                    size_t ln = ns->len;
-                    ns->len = 0;
                     push_label(l);
-                    labelprint2(&ns->members, flab, labelmode);
+                    labelprint2(ns, flab, labelmode);
                     pop_label();
-                    ns->len = ln;
                 }
             }
         } else {
@@ -556,20 +632,23 @@ static void labelprint2(const struct avltree *members, FILE *flab, int labelmode
             val_destroy(&str->v);
         }
     }
+    names->len = ln;
 }
 
 static inline const uint8_t *get_line(const struct file_s *file, size_t line) {
     return &file->data[file->line[line - 1]];
 }
 
-static void labeldump(Namespace *members, FILE *flab) {
-    const struct avltree_node *n;
+static void labeldump(Namespace *names, FILE *flab) {
+    size_t n, ln;
 
-    for (n = avltree_first(&members->members); n != NULL; n = avltree_next(n)) {
-        const struct namespacekey_s *l = cavltree_container_of(n, struct namespacekey_s, node);
-        Label *l2 = l->key;
+    if (names->len == 0) return;
+    ln = names->len; names->len = 0;
+    for (n = 0; n <= names->mask; n++) {
+        Label *l2 = names->data[n];
         Namespace *ns;
 
+        if (l2 == NULL) continue;
         if (l2->name.len < 2 || l2->name.data[1] != 0) {
             Str *val = (Str *)l2->value->obj->repr(l2->value, NULL, SIZE_MAX);
             if (val != NULL) {
@@ -592,15 +671,13 @@ static void labeldump(Namespace *members, FILE *flab) {
 
         if (ns != NULL && ns->len != 0) {
             if (l2->name.len < 2 || l2->name.data[1] != 0) {
-                size_t ln = ns->len;
-                ns->len = 0;
                 push_label(l2);
                 labeldump(ns, flab);
                 pop_label();
-                ns->len = ln;
             }
         }
     }
+    names->len = ln;
 }
 
 static Namespace *find_space(const char *here, bool use) {
@@ -614,8 +691,9 @@ static Namespace *find_space(const char *here, bool use) {
     pline = (const uint8_t *)here;
     lpoint.pos = 0;
     do {
-        labelname.data = pline + lpoint.pos; labelname.len = get_label();
+        labelname.data = pline + lpoint.pos; labelname.len = get_label(labelname.data);
         if (labelname.len == 0) return NULL;
+        lpoint.pos += labelname.len;
         l = find_label2(&labelname, space);
         if (l == NULL) return NULL;
         space = get_namespace(l->value);
@@ -623,6 +701,8 @@ static Namespace *find_space(const char *here, bool use) {
         lpoint.pos++;
     } while (labelname.data[labelname.len] == '.');
 
+    if (labelname.data[labelname.len] != 0) return NULL;
+
     if (use) {
         l->usepass = pass;
         l->ref = true;
@@ -631,7 +711,6 @@ static Namespace *find_space(const char *here, bool use) {
 }
 
 bool labelprint(const struct symbol_output_s *output, bool append) {
-    bool oldreferenceit = referenceit;
     FILE *flab;
     struct linepos_s nopoint = {0, 0};
     int err;
@@ -643,7 +722,6 @@ bool labelprint(const struct symbol_output_s *output, bool append) {
         return true;
     }
     clearerr(flab); errno = 0;
-    referenceit = false;
     label_stack.stack = NULL;
     label_stack.p = label_stack.len = 0;
     space = find_space(output->space, false);
@@ -654,11 +732,22 @@ bool labelprint(const struct symbol_output_s *output, bool append) {
         err_msg2(ERROR____LABEL_ROOT, &labelname, &nopoint);
     } else if (output->mode == LABEL_DUMP) {
         labeldump(space, flab);
+    } else if (output->mode == LABEL_VICE || output->mode == LABEL_VICE_NUMERIC) {
+        vice_addresses = (struct vice_addresses_s *)mallocx(sizeof *vice_addresses);
+        vice_addresses->next = NULL;
+        vice_addrp = 0;
+        avltree_init(&vice_filter);
+        lastvc = &vice_addresses->list[vice_addrp];
+        labelprint2(space, flab, output->mode);
+        while (vice_addresses != NULL) {
+            struct vice_addresses_s *old = vice_addresses;
+            vice_addresses = vice_addresses->next;
+            free(old);
+        }
     } else {
-        labelprint2(&space->members, flab, output->mode);
+        labelprint2(space, flab, output->mode);
     }
     free(label_stack.stack);
-    referenceit = oldreferenceit;
     err = ferror(flab);
     err |= (flab != stdout) ? fclose(flab) : fflush(flab);
     if (err != 0 && errno != 0) {
@@ -673,15 +762,14 @@ void ref_labels(void) {
     for (j = 0; j < arguments.symbol_output_len; j++) {
         const struct symbol_output_s *output = &arguments.symbol_output[j];
         Namespace *space;
-        struct avltree_node *n;
+        size_t n;
 
-        if (output->mode != LABEL_64TASS) continue;
+        if (output->mode != LABEL_EXPORT) continue;
         space = find_space(output->space, true);
-        if (space == NULL) continue;
+        if (space == NULL || space->len == 0) continue;
 
-        n = avltree_first(&space->members);
-        while (n != NULL) {
-            Label *l = find_strongest_label(&n, label_compare);            /* already exists */
+        for (n = 0; n <= space->mask; n++) {
+            Label *l = space->data[n];
             if (l == NULL || l->name.data == NULL) continue;
             if (l->name.len > 1 && l->name.data[1] == 0) continue;
             switch (l->value->obj->type) {
@@ -692,6 +780,8 @@ void ref_labels(void) {
             case T_STRUCT: continue;
             default:break;
             }
+            if (l != namespace_lookup(space, l)) continue;
+            if (l->value->obj == ERROR_OBJ) err_msg_output((Error *)l->value);
             l->ref = true;
             l->usepass = pass;
         }
@@ -761,7 +851,6 @@ void destroy_variables(void)
     val_destroy(&root_namespace->v);
     val_destroy(&cheap_context->v);
     destroy_lastlb();
-    if (lastlb2 != NULL) namespacekey_free(lastlb2);
     while (context_stack.p != 0) {
         struct cstack_s *c = &context_stack.stack[--context_stack.p];
         val_destroy(&c->normal->v);
