@@ -1,5 +1,5 @@
 /*
-    $Id: wait_e.h 1798 2019-01-12 21:02:31Z soci $
+    $Id: wait_e.h 2131 2019-12-30 19:02:20Z soci $
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@ typedef enum Wait_types {
     W_NONE, W_ENDM, W_ENDM3, W_BEND, W_BEND2, W_HERE, W_HERE2, W_ENDU, W_ENDU2,
     W_ENDU3, W_ENDS, W_ENDS2, W_ENDS3, W_ENDC, W_ENDP, W_ENDP2, W_NEXT,
     W_NEXT2, W_NEXT3, W_SEND, W_SEND2, W_PEND, W_FI, W_FI2, W_ENDF, W_ENDF3,
-    W_SWITCH, W_SWITCH2, W_WEAK, W_WEAK2, W_ENDN, W_ENDN2, W_ENDV, W_ENDV2
+    W_SWITCH, W_SWITCH2, W_WEAK, W_WEAK2, W_ENDN, W_ENDN2, W_ENDV, W_ENDV2,
+    W_ENDWITH, W_ENDWITH2
 } Wait_types;
 #endif
